Package: ENmix
Type: Package
Title: Quality control and analysis tools for Illumina DNA methylation
        BeadChip
Version: 1.47.3
Authors@R: 
  c(person("Zongli","Xu",email="xuz@niehs.nih.gov",role=c("cre","aut")),
    person("Liang", "Niu", role="aut",email="niulg@ucmail.uc.edu"),
    person("Jack", "Taylor",role="ctb",email="taylor@niehs.nih.gov"))
Description: Tools for quanlity control, analysis and visulization
             of Illumina DNA methylation array data. 
License: Artistic-2.0
biocViews: DNAMethylation, Preprocessing, QualityControl, TwoChannel,
        Microarray, OneChannel, MethylationArray, BatchEffect,
        Normalization, DataImport, Regression,
        PrincipalComponent,Epigenetics, MultiChannel,
        DifferentialMethylation, ImmunoOncology
Depends: parallel,doParallel,foreach,SummarizedExperiment,stats,R (>=
        3.5.0)
Imports: grDevices,graphics,matrixStats,methods,utils,irlba,
        geneplotter,impute,minfi,RPMM,illuminaio,dynamicTreeCut,IRanges,gtools,
        Biobase,ExperimentHub,AnnotationHub,genefilter,gplots,quadprog,S4Vectors
Suggests: minfiData, RUnit, BiocGenerics, BiocStyle, knitr, rmarkdown
VignetteBuilder: knitr
Encoding: UTF-8
URL: https://github.com/Bioconductor/ENmix
BugReports: https://github.com/Bioconductor/ENmix/issues
NeedsCompilation: no
git_url: https://git.bioconductor.org/packages/ENmix
git_branch: devel
git_last_commit: dbede0f
git_last_commit_date: 2026-04-07
Repository: Bioconductor 3.23
Date/Publication: 2026-04-07
Packaged: 2026-04-07 21:31:31 UTC; biocbuild
Author: Zongli Xu [cre, aut],
  Liang Niu [aut],
  Jack Taylor [ctb]
Maintainer: Zongli Xu <xuz@niehs.nih.gov>

import(parallel)
import(doParallel)
import(foreach)
import(stats)
import(SummarizedExperiment)
importFrom(IRanges,IRanges,reduce,start,end,width)
importFrom(S4Vectors,metadata,SimpleList)
importFrom(illuminaio,readIDAT)
importFrom(matrixStats, colMedians,colMads)
importFrom(grDevices,dev.off,jpeg,colorRampPalette, pdf)
importFrom("grDevices", "postscript")
importFrom(graphics,abline,axis,boxplot,hist,legend,
    lines,par,plot,polygon,text,points,mtext)
importFrom(methods, as, callNextMethod, is, new)
importFrom(utils, read.table,write.table,head,tail,write.csv)
importFrom(geneplotter,multidensity)
importFrom(impute,impute.knn)
importFrom(RPMM,blc,betaObjf)
importFrom(dynamicTreeCut,printFlush)
importFrom(minfi,getProbeInfo,getManifestInfo,getControlAddress,
    detectionP,preprocessRaw,getAnnotation,featureNames,
    getProbeType)
importFrom(Biobase,featureNames,rowMedians)
importFrom(ExperimentHub,ExperimentHub)
importFrom(AnnotationHub,query)
importFrom(genefilter,rowFtests,rowttests)
importFrom(gplots,heatmap.2)
importFrom(quadprog,solve.QP)
importFrom(gtools,mixedsort)
importFrom(irlba,prcomp_irlba)


exportClasses(rgDataSet,methDataSet)

export(preprocessENmix,
    nmode,norm.quantile,plotCtrl,
    QCinfo,pcrplot,freqpoly,multifreqpoly,
    qcfilter,rcp,oxBS.MLE,ctrlsva,relic,mpreprocess,
    getB,B2M,M2B,rgDataSet,methDataSet,repicc,
    calcdetP,getCGinfo,getmeth,readidat,dupicc,
    readmanifest,estimateCellProp,predSex,methyAge,
    combp,ipdmr,mhtplot,p.qqplot,simubed,methscore,rm.cgsuffix,rcp2)

##### a function to get a table of p-values for estimating acf
#####loc should be increasing; 
acf.table<-function(x,loc,dist.cutoff){
  flag=TRUE; lag=1; result=NULL
  while(flag){
    x1=head(x,-lag); x2=tail(x,-lag); dist=diff(loc,lag=lag)
    index=(dist<dist.cutoff)  
    if(all(!index)){flag=FALSE}else{
      result=rbind(result,data.frame(x1=x1[index],x2=x2[index],dist=dist[index]))
    lag=lag+1
    }
  }
  return(result)  
}

##### a function to estimate acf
get.acf<-function(data,dist.cutoff,bin.size){
  temp<-NULL
  for (chr in unique(as.vector(data$chr))){
    y<-data[as.vector(data$chr)==chr,]; y<-y[order(y$end),]
    temp<-rbind(temp,acf.table(y$p,y$end,dist.cutoff))
  }
  bin.label<-findInterval(temp$dist,seq(bin.size,dist.cutoff,bin.size))
  temp.stouffer<-by(temp,bin.label,FUN=function(x){cor.test(qnorm(x$x1),
               qnorm(x$x2),alternative="greater")},simplify=FALSE)

  cor.stouffer<-sapply(temp.stouffer,function(x){x$estimate})
  p.stouffer<-sapply(temp.stouffer,function(x){x$p.value})

  if (any(p.stouffer>0.05)){
    index=min(which(p.stouffer>0.05))
    cor.stouffer[index:length(cor.stouffer)]=0
  }
  return(cor.stouffer)
}

#regional P value plot
regplot<-function(ref,sig,extend=2000,outf="region_plot.pdf"){
  sig=sig[order(sig[,"chr"],sig[,"start"]),]
  ref=ref[order(ref[,"chr"],ref[,"start"]),]

  pdf(outf)
  for(i in 1:nrow(sig)){
    chr=as.vector(sig$chr[i])
    pos1=sig$start[i]
    pos2=sig$end[i]
    subset=ref[as.vector(ref$chr)==chr & ref$end>=(pos1-extend) & ref$end<=(pos2+extend),]
    subset$col="black"
    subset$col[subset$end>=pos1 & subset$end<=pos2]="red"

    ylab=bquote('-log'['10']*'(P) value')
    pos1=pos1-1
    plot(subset$start,-log10(subset$p),col=subset$col,pch=20,xlim=c(pos1-extend,
          pos2+extend),xlab="Chromosome position",ylab=ylab,main=paste0("Chromosome ",chr,": ",pos1,"-",pos2))
  }
  dev.off()
}

##### interval method
##### 1. get interval p-values; 2. select all intervals with fdr<seed; 
###3. combine nearby sig. intervals and sig. probes (probes with fdr<seed)
### into regions; 4. find region p-values
ipdmr<-function(data,include.all.sig.sites=TRUE,dist.cutoff=1000,bin.size=50,
                seed=0.05,region_plot=TRUE,mht_plot=TRUE,verbose=TRUE){
  data=as.data.frame(data)
  data$start=as.numeric(as.vector(data$start))
  data$end=as.numeric(as.vector(data$end))
  data=data[!is.na(data$start) & !is.na(data$end),]
  data$p=as.numeric(as.vector(data$p))

  if(sum(data$p<=0 | data$p>=1)>0){
    cat("P values <=0 were re-assigned with a value of 1E-300\n
P values >=1 were re-assigned with a value of 0.99999\n")
    data$p[data$p<=0]=1E-300
    data$p[data$p>=1]=0.99999
  }


  tmp=as.vector(data$chr); tmp=toupper(tmp)
  tmp=sub("CHR","",tmp);data$chr=tmp

  acf<-get.acf(data,dist.cutoff,bin.size)
  if(verbose){
    cat("P value correlations:\n")
    bin=seq(bin.size,dist.cutoff,bin.size)
    if(!(dist.cutoff%%bin.size==0)){bin=c(bin,dist.cutoff)}
    print(data.frame(bin=bin,acf=acf))
  }
  result<-NULL
  for (chr in unique(as.vector(data$chr))){
    y=data[as.vector(data$chr)==chr,]; y=y[order(y$end),]
    pos=y$end; p=qnorm(y$p)
    index=which(diff(pos)<dist.cutoff)
    if(length(index)>0){ 
      int<-findInterval(diff(pos)[index],seq(bin.size,dist.cutoff,bin.size))
      sd<-sqrt(acf[int+1]*2+2)
      int.p<-pnorm(p[index]+p[index+1],mean=0,sd=sd)
      start=pos[index]; end=pos[index+1];
      result=rbind(result,data.frame(chr,start,end,int.p))
    }
  }

  if (include.all.sig.sites){
    temp=data[p.adjust(data$p,method="fdr")<seed,]
  }else{
    int.sites=unique(c(paste(result$chr,result$start),
                  paste(result$chr,result$end)))
    data.1=data[!(paste(as.vector(data$chr),data$end) %in% int.sites),]
    temp=data.1[p.adjust(data.1$p,method="fdr")<seed,]
  }

  result=result[p.adjust(result$int.p,method="fdr")<seed,]
  result=rbind(result,data.frame(chr=temp$chr,start=temp$end,end=temp$end,int.p=temp$p))

  result.fdr=NULL
  if (nrow(result)>0){
    for (chr in unique(result$"chr")){
      y=data[as.vector(data$chr)==chr,]; y=y[order(y$end),]
      pos=y$end; p=qnorm(y$p)

      result.chr=result[result$"chr"==chr,]
      a=IRanges::IRanges(start=result.chr$start,end=result.chr$end)
      b=IRanges::reduce(a,min.gapwidth=dist.cutoff)

      start=IRanges::start(b); end=IRanges::end(b)
      region.max<-max(IRanges::width(b))
      temp=sapply(1:length(b),function(i){
         index.i=(pos>=start[i] & pos<=end[i]);
         if (sum(index.i)>1){  
           int<-findInterval(c(dist(pos[index.i])),
                  seq(bin.size,region.max+bin.size,bin.size))
           sd<-sqrt(sum(ifelse(int<length(acf),acf[int+1],0))*2+sum(index.i))
           return(pnorm(sum(p[index.i]),mean=0,sd=sd))
         }else{
            return(y$p[index.i])
         }
      })
      result.fdr=rbind(result.fdr,data.frame(chr,start,end,p=temp))
    }
  
    ##### BH correction
    result.fdr$fdr=p.adjust(result.fdr$p,method="fdr")
    result.fdr<-result.fdr[order(result.fdr$p),]

    ##### use 0-coordinate
#    result.fdr$start=(result.fdr$start-1)
  }

  if(is.null(result.fdr)){cat("Number of identified DMR:  0\n")}else{
  result.fdr$start=as.numeric(as.vector(result.fdr$start))
  result.fdr$end=as.numeric(as.vector(result.fdr$end))
  result.fdr$chr=factor(result.fdr$chr)

    ndmr=nrow(result.fdr)
    cat("Number of DMRs identified:  ",ndmr, "\n")
    if(region_plot){
      cat("Drawing regional plot: region_plot.pdf ...\n")
      sig=result.fdr
      regplot(ref=data,sig)
    }
    if(mht_plot){
      cat("Drawing manhattan plot: mht.jpg ...\n")
      set2=NULL
      for(i in 1:ndmr){
        set2=c(set2,as.vector(data$probe[as.vector(data$chr)==as.vector(result.fdr$chr[i]) 
              & data$end>=result.fdr$start[i] & data$end<=result.fdr$end[i]]))
      }
      mhtplot(probe=as.vector(data$probe),chr=as.vector(data$chr),pos=data$start,p=data$p,color="gray",markprobe=set2)
  }
  #number of probes within eath DMR
  result.fdr$nprobe=NA
  for(i in 1:nrow(result.fdr)){
result.fdr$nprobe[i]=nrow(data[as.vector(data$chr)==as.vector(result.fdr$chr[i]) 
& data$end>=result.fdr$start[i] & data$end<=result.fdr$end[i],])
  }
#assume start is 1 base less than end
  result.fdr$start=(result.fdr$start-1)

#probes in significant DMR
  result.fdr$probe=NA
  for(i in 1:nrow(result.fdr)){
  tmp=data[data$chr==as.vector(result.fdr$chr[i]) & data$start >= result.fdr$start[i] & data$end <= result.fdr$end[i],]
  result.fdr$probe[i]=paste0(tmp$probe,collapse=";")}

  write.table(result.fdr,"resu_ipdmr.csv",row.names=FALSE,sep=",")
  } 
}

##### comb_p-like method
##### 1. get smoothed p-values; 2. select all probes with smoothed p-values with fdr<seed; 3. combine nearby sig. probes into regions; 4. find region p-values
combp<-function(data,dist.cutoff=1000,bin.size=310,seed=0.01,
               region_plot=TRUE,mht_plot=TRUE,nCores=10,verbose=TRUE){
  if(nCores>detectCores()){nCores=detectCores()}
  data=as.data.frame(data)
  data$start=as.numeric(as.vector(data$start))
  data$end=as.numeric(as.vector(data$end))
  data=data[!is.na(data$start) & !is.na(data$end),]
  data$p=as.numeric(as.vector(data$p))

  if(sum(data$p<=0 | data$p>=1)>0){
    cat("P values <=0 were re-assigned with a value of 1E-300\n
P values >=1 were re-assigned with a value of 0.99999\n")
    data$p[data$p<=0]=1E-300
    data$p[data$p>=1]=0.99999
  }

  acf<-get.acf(data,dist.cutoff,bin.size)
  if(verbose){
    cat("P value correlations:\n")
    bin=seq(bin.size,dist.cutoff,bin.size)
    if(!(dist.cutoff%%bin.size==0)){bin=c(bin,dist.cutoff)}
    print(data.frame(bin=bin,acf=acf))
  }

  result<-mclapply(unique(as.vector(data$chr)), function(chr){
    y=data[as.vector(data$chr)==chr,]; y=y[order(y$end),]
    pos=y$end; p=qnorm(y$p)

    temp=sapply(pos,function(i){
      index.i=(abs(pos-i)<bin.size);
      if (sum(index.i)>1){  
        int<-findInterval(c(dist(pos[index.i])),c(bin.size,2*bin.size))
        sd<-sqrt(sum(acf[int+1])*2+sum(index.i))
        return(pnorm(sum(p[index.i]),mean=0,sd=sd))
      }else{return(y$p[index.i])}
    })

    return(data.frame(chr,start=pos,end=pos,s.p=temp))
  },mc.cores=nCores)

  result <- do.call("rbind", result)
  names(result)=c("chr","start","end","s.p")

  result=result[p.adjust(result$s.p,method="fdr")<seed,]

  result.fdr=NULL
  if (nrow(result)>0){
    for (chr in unique(result$"chr")){
      y=data[as.vector(data$chr)==chr,]; y=y[order(y$end),]
      pos=y$end; p=qnorm(y$p)

      result.chr=result[result$"chr"==chr,]
      a=IRanges::IRanges(start=result.chr$start,end=result.chr$end)
      b=IRanges::reduce(a,min.gapwidth=dist.cutoff)

      start=IRanges::start(b); end=IRanges::end(b)
      region.max<-max(IRanges::width(b))
      temp=sapply(1:length(b),function(i){
        index.i=(pos>=start[i] & pos<=end[i]);
        if (sum(index.i)>1){  
          int<-findInterval(c(dist(pos[index.i])),
              seq(bin.size,region.max+bin.size,bin.size))
          sd<-sqrt(sum(ifelse(int<length(acf),
              acf[int+1],0))*2+sum(index.i))
          return(pnorm(sum(p[index.i]),mean=0,sd=sd))
        }else{return(y$p[index.i])}
      })
      result.fdr=rbind(result.fdr,data.frame(chr,start,end,p=temp))
    }

    ##### BH FDR correction and Sidak correction
    result.fdr$fdr=p.adjust(result.fdr$p,method="fdr")
    result.fdr$sidak=(1-(1-result.fdr$p)^(nrow(data)/(result.fdr$end-result.fdr$start+1)))
    result.fdr<-result.fdr[order(result.fdr$p),]

    ##### use 0-coordinate
#    result.fdr$start=(result.fdr$start-1)
  }

  if(is.null(result.fdr)){cat("Number of identified DMR:  0\n")}else{
    ndmr=nrow(result.fdr)
  result.fdr$start=as.numeric(as.vector(result.fdr$start))
  result.fdr$end=as.numeric(as.vector(result.fdr$end))
  result.fdr$chr=factor(result.fdr$chr)

    cat("Number of DMRs identified:  ",ndmr, "\n")
    if(region_plot){
      cat("Drawing regional plot: region_plot.pdf ...\n")
      sig=result.fdr
      regplot(ref=data,sig)
    }
  if(mht_plot){
    cat("Drawing manhattan plot: mht.jpg ...\n")
    set2=NULL
    for(i in 1:ndmr){
        set2=c(set2,as.vector(data$probe[as.vector(data$chr)==as.vector(result.fdr$chr[i])
           & data$end>=result.fdr$start[i] & data$end<=result.fdr$end[i]]))
    }
  mhtplot(probe=data$probe,chr=as.vector(data$chr),pos=data$start,p=data$p,color="gray",markprobe=set2)
  }
  #number of probes within eath DMR

  result.fdr$nprobe=NA
  for(i in 1:nrow(result.fdr)){
result.fdr$nprobe[i]=nrow(data[as.vector(data$chr)==as.vector(result.fdr$chr[i])
& data$end>=result.fdr$start[i] & data$end<=result.fdr$end[i],])
}
#assume start is 1 base less than end
    result.fdr$start=(result.fdr$start-1)
#probes in significant DMR
  result.fdr$probe=NA
  for(i in 1:nrow(result.fdr)){
  tmp=data[data$chr==as.vector(result.fdr$chr[i]) & data$start >= result.fdr$start[i] & data$end <= result.fdr$end[i],]
  result.fdr$probe[i]=paste0(tmp$probe,collapse=";")}

  write.table(result.fdr,"resu_combp.csv",row.names=FALSE,sep=",")
  }
}

## Adopted from R package DunedinPACE. 
##Daniel W Belsky, Avshalom Caspi, David L Corcoran,et al. DunedinPACE, a DNA methylation biomarker of the pace of aging, eLIfe, 2022

DunedinPACE <-function (betas, proportionOfProbesRequired = 0.8) 
{
#    require(preprocessCore)
    mPOA_Models=list()
    load(system.file("mPOA_Models.RData",package="ENmix"))
        if (!is.numeric(as.matrix(betas))) {
            stop("betas matrix/data.frame is not numeric!")
        }
        probeOverlap <- length(which(rownames(betas) %in% mPOA_Models$model_probes))/length(mPOA_Models$model_probes)
        probeOverlap_background <- length(which(rownames(betas) %in% 
            mPOA_Models$gold_standard_probes))/length(mPOA_Models$gold_standard_probes)
        if (probeOverlap < proportionOfProbesRequired | probeOverlap_background < 
            proportionOfProbesRequired) {
            result <- rep(NA, ncol(betas))
            names(result) <- colnames(betas)
            model_results=result
            model_results
        }
        else {
            betas.mat <- as.matrix(betas[which(rownames(betas) %in% 
                mPOA_Models$gold_standard_probes), 
                ])
            probesNotInMatrix <- mPOA_Models$gold_standard_probes[which(mPOA_Models$gold_standard_probes %in% 
                rownames(betas.mat) == F)]
            if (length(probesNotInMatrix) > 0) {
                for (probe in probesNotInMatrix) {
                  tmp.mat <- matrix(0, nrow = 1, ncol = ncol(betas.mat))
                  rownames(tmp.mat) <- probe
                  colnames(tmp.mat) <- colnames(betas.mat)
                  tmp.mat[probe, ] <- rep(mPOA_Models$gold_standard_means[probe], 
                    ncol(tmp.mat))
                  betas.mat <- rbind(betas.mat, tmp.mat)
                }
            }
            samplesToRemove <- colnames(betas.mat)[which(apply(betas.mat, 
                2, function(x) {
                  1 - (length(which(is.na(x)))/length(x)) < proportionOfProbesRequired
                }))]
            if (length(samplesToRemove) > 0) {
                betas.mat <- betas.mat[, -which(colnames(betas.mat) %in% 
                  samplesToRemove)]
            }
            if (ncol(betas.mat) > 0){
                pctValuesPresent <- apply(betas.mat, 1, function(x) {
                  1 - (length(which(is.na(x)))/length(x))
                })
                probesToAdjust <- which(pctValuesPresent < 1 & 
                  pctValuesPresent >= proportionOfProbesRequired)
                if (length(probesToAdjust) > 0) {
                  if (length(probesToAdjust) > 1) {
                    betas.mat[probesToAdjust, ] <- t(apply(betas.mat[probesToAdjust, 
                      ], 1, function(x) {
                      x[is.na(x)] = mean(x, na.rm = TRUE)
                      x
                    }))
                  }
                  else {
                    betas.mat[probesToAdjust, which(is.na(betas.mat[probesToAdjust, 
                      ]))] <- mean(betas.mat[probesToAdjust, 
                      ], na.rm = T)
                  }
                }
                if (length(which(pctValuesPresent < proportionOfProbesRequired)) > 
                  0) {
                  probesToReplaceWithMean <- rownames(betas.mat)[which(pctValuesPresent < 
                    proportionOfProbesRequired)]
                  for (probe in probesToReplaceWithMean) {
                    betas.mat[probe, ] <- rep(mPOA_Models$model_means[probe], 
                      ncol(betas.mat))
                  }
                }
                betas.norm <- normalize.quantiles.use.target2(betas.mat, 
                  target = mPOA_Models$gold_standard_means)
                rownames(betas.norm) <- rownames(betas.mat)
                colnames(betas.norm) <- colnames(betas.mat)
                score = mPOA_Models$model_intercept + 
                  rowSums(t(betas.norm[mPOA_Models$model_probes, 
                    ]) %*% diag(mPOA_Models$model_weights))
                names(score) <- colnames(betas.norm)
                if (length(samplesToRemove) > 0) {
                  score.tmp <- rep(NA, length(samplesToRemove))
                  names(score.tmp) <- samplesToRemove
                  score <- c(score, score.tmp)
                }
                model_results <- score[colnames(betas)]
                model_results
            }
            else {
                model_results <- rep(NA, ncol(betas.mat))
                names(model_results) <- colnames(betas.mat)
                model_results
            }
        }
}

#Adopted from Hovath modified BMIQ method code

#require(RPMM);
betaEst2=function (y, w, weights) 
{
    yobs = !is.na(y)
    if (sum(yobs) <= 1) 
        return(c(1, 1))
    y = y[yobs]
    w = w[yobs]
    weights = weights[yobs]
    N = sum(weights * w)
    p = sum(weights * w * y)/N
    v = sum(weights * w * y * y)/N - p * p
    logab = log(c(p, 1 - p)) + log(pmax(1e-06, p * (1 - p)/v - 
        1))
    if (sum(yobs) == 2) 
        return(exp(logab))
    opt = try(optim(logab, RPMM::betaObjf, ydata = y, wdata = w, weights = weights, 
        method = "Nelder-Mead",control=list(maxit=50) ), silent = TRUE)
    if (inherits(opt, "try-error")) 
        return(c(1, 1))
    exp(opt$par)
} # end of function betaEst

blc2=function (Y, w, maxiter = 25, tol = 1e-06, weights = NULL, verbose = TRUE) 
{
    Ymn = min(Y[Y > 0], na.rm = TRUE)
    Ymx = max(Y[Y < 1], na.rm = TRUE)
    Y = pmax(Y, Ymn/2)
    Y = pmin(Y, 1 - (1 - Ymx)/2)
    Yobs = !is.na(Y)
    J = dim(Y)[2]
    K = dim(w)[2]
    n = dim(w)[1]
    if (n != dim(Y)[1]) 
        stop("Dimensions of w and Y do not agree")
    if (is.null(weights)) 
        weights = rep(1, n)
    mu = a = b = matrix(Inf, K, J)
    crit = Inf
    for (i in 1:maxiter) {
        warn0 = options()$warn
        options(warn = -1)
        eta = apply(weights * w, 2, sum)/sum(weights)
        mu0 = mu
        for (k in 1:K) {
            for (j in 1:J) {
                ab = betaEst2(Y[, j], w[, k], weights)
                a[k, j] = ab[1]
                b[k, j] = ab[2]
                mu[k, j] = ab[1]/sum(ab)
            }
        }
        ww = array(0, dim = c(n, J, K))
        for (k in 1:K) {
            for (j in 1:J) {
                ww[Yobs[, j], j, k] = dbeta(Y[Yobs[, j], j], 
                  a[k, j], b[k, j], log = TRUE)
            }
        }
        options(warn = warn0)
        w = apply(ww, c(1, 3), sum, na.rm = TRUE)
        wmax = apply(w, 1, max)
        for (k in 1:K) w[, k] = w[, k] - wmax
        w = t(eta * t(exp(w)))
        like = apply(w, 1, sum)
        w = (1/like) * w
        llike = weights * (log(like) + wmax)
        crit = max(abs(mu - mu0))
        if (verbose) 
            print(crit)
        if (crit < tol) 
            break
    }
    return(list(a = a, b = b, eta = eta, mu = mu, w = w, llike = sum(llike)))
}

BMIQcalibration=function(datM,goldstandard.beta,nL=3,doH=TRUE,nfit=20000,th1.v=c(0.2,0.75),th2.v=NULL,niter=5,tol=0.001,plots=FALSE,calibrateUnitInterval=TRUE){
if (length(goldstandard.beta) !=dim(datM)[[2]] ) {stop("Error in function arguments length(goldstandard.beta) !=dim(datM)[[2]]. Consider transposing datM.")}
if (plots ) {par(mfrow=c(2,2))}
beta1.v = goldstandard.beta

if (calibrateUnitInterval ) {datM=CalibrateUnitInterval(datM)}

### estimate initial weight matrix from type1 distribution
w0.m = matrix(0,nrow=length(beta1.v),ncol=nL);
w0.m[which(beta1.v <= th1.v[1]),1] = 1;
w0.m[intersect(which(beta1.v > th1.v[1]),which(beta1.v <= th1.v[2])),2] = 1;
w0.m[which(beta1.v > th1.v[2]),3] = 1;
### fit type1
print("Fitting EM beta mixture to goldstandard probes");
set.seed(1)
rand.idx = sample(1:length(beta1.v),min(c(nfit, length(beta1.v))  )   ,replace=FALSE)
em1.o =RPMM::blc(matrix(beta1.v[rand.idx],ncol=1),w=w0.m[rand.idx,],maxiter=niter,tol=tol);
subsetclass1.v = apply(em1.o$w,1,which.max);
subsetth1.v = c(mean(max(beta1.v[rand.idx[subsetclass1.v==1]]),min(beta1.v[rand.idx[subsetclass1.v==2]])),mean(max(beta1.v[rand.idx[subsetclass1.v==2]]),min(beta1.v[rand.idx[subsetclass1.v==3]])));
class1.v = rep(2,length(beta1.v));
class1.v[which(beta1.v < subsetth1.v[1])] = 1;
class1.v[which(beta1.v > subsetth1.v[2])] = 3;
nth1.v = subsetth1.v;
print("Done");

### generate plot from estimated mixture
if(plots){
print("Check");
tmpL.v = as.vector(rmultinom(1:nL,length(beta1.v),prob=em1.o$eta));
tmpB.v = vector();
for(l in 1:nL){
  tmpB.v = c(tmpB.v,rbeta(tmpL.v[l],em1.o$a[l,1],em1.o$b[l,1]));
}
plot(density(beta1.v),main= paste("Type1fit-", sep=""));
d.o = density(tmpB.v);
points(d.o$x,d.o$y,col="green",type="l")
legend(x=0.5,y=3,legend=c("obs","fit"),fill=c("black","green"),bty="n");
}

### Estimate Modes 
if (  sum(class1.v==1)==1 ){ mod1U= beta1.v[class1.v==1]}
if (  sum(class1.v==3)==1 ){ mod1M= beta1.v[class1.v==3]}
if (  sum(class1.v==1) >1){ 
d1U.o = density(beta1.v[class1.v==1])
mod1U = d1U.o$x[which.max(d1U.o$y)]
}
if (  sum(class1.v==3)>1 ){ 
d1M.o = density(beta1.v[class1.v==3])
mod1M = d1M.o$x[which.max(d1M.o$y)]
}

### BETA 2
for (ii in 1:dim(datM)[[1]] ){
printFlush(paste("ii=",ii))
sampleID=ii
beta2.v = as.numeric(datM[ii,])

d2U.o = density(beta2.v[which(beta2.v<0.4)]);
d2M.o = density(beta2.v[which(beta2.v>0.6)]);
mod2U = d2U.o$x[which.max(d2U.o$y)]
mod2M = d2M.o$x[which.max(d2M.o$y)]

### now deal with type2 fit
th2.v = vector();
th2.v[1] = nth1.v[1] + (mod2U-mod1U);
th2.v[2] = nth1.v[2] + (mod2M-mod1M);

### estimate initial weight matrix 
w0.m = matrix(0,nrow=length(beta2.v),ncol=nL);
w0.m[which(beta2.v <= th2.v[1]),1] = 1;
w0.m[intersect(which(beta2.v > th2.v[1]),which(beta2.v <= th2.v[2])),2] = 1;
w0.m[which(beta2.v > th2.v[2]),3] = 1;

print("Fitting EM beta mixture to input probes");
# I fixed an error in the following line (replaced beta1 by beta2)
set.seed(1)
rand.idx = sample(1:length(beta2.v),min(c(nfit, length(beta2.v)),na.rm=TRUE)   ,replace=FALSE)
em2.o = blc2(Y=matrix(beta2.v[rand.idx],ncol=1),w=w0.m[rand.idx,],maxiter=niter,tol=tol,verbose=TRUE);
print("Done");

### for type II probes assign to state (unmethylated, hemi or full methylation)
subsetclass2.v = apply(em2.o$w,1,which.max);


if (sum(subsetclass2.v==2)>0 ){
subsetth2.v = c(mean(max(beta2.v[rand.idx[subsetclass2.v==1]]),min(beta2.v[rand.idx[subsetclass2.v==2]])),
mean(max(beta2.v[rand.idx[subsetclass2.v==2]]),min(beta2.v[rand.idx[subsetclass2.v==3]])));
}
if (sum(subsetclass2.v==2)==0 ){
subsetth2.v = c(1/2*max(beta2.v[rand.idx[subsetclass2.v==1]])+ 1/2*mean(beta2.v[rand.idx[subsetclass2.v==3]]), 1/3*max(beta2.v[rand.idx[subsetclass2.v==1]])+ 2/3*mean(beta2.v[rand.idx[subsetclass2.v==3]]));
}
class2.v = rep(2,length(beta2.v));
class2.v[which(beta2.v <= subsetth2.v[1])] = 1;
class2.v[which(beta2.v >= subsetth2.v[2])] = 3;

### generate plot
if(plots){
tmpL.v = as.vector(rmultinom(1:nL,length(beta2.v),prob=em2.o$eta));
tmpB.v = vector();
for(lt in 1:nL){
  tmpB.v = c(tmpB.v,rbeta(tmpL.v[lt],em2.o$a[lt,1],em2.o$b[lt,1]));
}
plot(density(beta2.v),  main= paste("Type2fit-",sampleID,sep="")  );
d.o = density(tmpB.v);
points(d.o$x,d.o$y,col="green",type="l")
legend(x=0.5,y=3,legend=c("obs","fit"),fill=c("black","green"),bty="n");
}

classAV1.v = vector();classAV2.v = vector();
for(l in 1:nL){
  classAV1.v[l] =  em1.o$mu[l,1];
  classAV2.v[l] =  em2.o$mu[l,1];
}

### start normalising input probes
print("Start normalising input probes");
nbeta2.v = beta2.v;
### select U probes
lt = 1;
selU.idx = which(class2.v==lt);
selUR.idx = selU.idx[which(beta2.v[selU.idx] > classAV2.v[lt])];
selUL.idx = selU.idx[which(beta2.v[selU.idx] < classAV2.v[lt])];
### find prob according to typeII distribution
p.v = pbeta(beta2.v[selUR.idx],em2.o$a[lt,1],em2.o$b[lt,1],lower.tail=FALSE);
### find corresponding quantile in type I distribution
q.v = qbeta(p.v,em1.o$a[lt,1],em1.o$b[lt,1],lower.tail=FALSE);
nbeta2.v[selUR.idx] = q.v;
p.v = pbeta(beta2.v[selUL.idx],em2.o$a[lt,1],em2.o$b[lt,1],lower.tail=TRUE);
### find corresponding quantile in type I distribution
q.v = qbeta(p.v,em1.o$a[lt,1],em1.o$b[lt,1],lower.tail=TRUE);
nbeta2.v[selUL.idx] = q.v;

### select M probes
lt = 3;
selM.idx = which(class2.v==lt);
selMR.idx = selM.idx[which(beta2.v[selM.idx] > classAV2.v[lt])];
selML.idx = selM.idx[which(beta2.v[selM.idx] < classAV2.v[lt])];
### find prob according to typeII distribution
p.v = pbeta(beta2.v[selMR.idx],em2.o$a[lt,1],em2.o$b[lt,1],lower.tail=FALSE);
### find corresponding quantile in type I distribution
q.v = qbeta(p.v,em1.o$a[lt,1],em1.o$b[lt,1],lower.tail=FALSE);
nbeta2.v[selMR.idx] = q.v;


if(doH){ ### if TRUE also correct type2 hemimethylated probes
### select H probes and include ML probes (left ML tail is not well described by a beta-distribution).
lt = 2;
selH.idx = c(which(class2.v==lt),selML.idx);
minH = min(beta2.v[selH.idx],na.rm=TRUE)
maxH = max(beta2.v[selH.idx],na.rm=TRUE)
deltaH = maxH - minH;
#### need to do some patching
deltaUH = -max(beta2.v[selU.idx],na.rm=TRUE) + min(beta2.v[selH.idx],na.rm=TRUE)
deltaHM = -max(beta2.v[selH.idx],na.rm=TRUE) + min(beta2.v[selMR.idx],na.rm=TRUE)

## new maximum of H probes should be
nmaxH = min(nbeta2.v[selMR.idx],na.rm=TRUE) - deltaHM;
## new minimum of H probes should be
nminH = max(nbeta2.v[selU.idx],na.rm=TRUE) + deltaUH;
ndeltaH = nmaxH - nminH;

### perform conformal transformation (shift+dilation)
## new_beta_H(i) = a + hf*(beta_H(i)-minH);
hf = ndeltaH/deltaH ;
### fix lower point first
nbeta2.v[selH.idx] = nminH + hf*(beta2.v[selH.idx]-minH);

}


### generate final plot to check normalisation
if(plots){
 print("Generating final plot");
 d1.o = density(beta1.v);
 d2.o = density(beta2.v);
 d2n.o = density(nbeta2.v);
 ymax = max(d2.o$y,d1.o$y,d2n.o$y);
plot(density(beta2.v),type="l",ylim=c(0,ymax),xlim=c(0,1), main=paste("CheckBMIQ-",sampleID,sep="") );
 points(d1.o$x,d1.o$y,col="red",type="l");
 points(d2n.o$x,d2n.o$y,col="blue",type="l");
 legend(x=0.5,y=ymax,legend=c("type1","type2","type2-BMIQ"),bty="n",fill=c("red","black","blue"));
}

datM[ii,]= nbeta2.v ;
} # end of for (ii=1 loop
datM
} # end of function BMIQcalibration


BMIQ = function(beta.v,design.v,nL=3,doH=TRUE,nfit=50000,th1.v=c(0.2,0.75),th2.v=NULL,niter=5,tol=0.001,plots=TRUE,sampleID=1,calibrateUnitInterval=TRUE){

if (calibrateUnitInterval) {
rangeBySample=range(beta.v,na.rm=TRUE)
minBySample=rangeBySample[1]
maxBySample=rangeBySample[2]
if ( (minBySample<0 | maxBySample>1) & !is.na(minBySample) & !is.na(maxBySample) ) {
y1=c(0.001,.999) 
x1=c(minBySample,maxBySample)
lm1=lm( y1 ~ x1 )
intercept1=coef(lm1)[[1]]
slope1=coef(lm1)[[2]]
beta.v=intercept1+slope1*beta.v
} # end of if
} # end of if (calibrateUnitInterval


type1.idx = which(design.v==1);
type2.idx = which(design.v==2);

beta1.v = beta.v[type1.idx];
beta2.v = beta.v[type2.idx];


### estimate initial weight matrix from type1 distribution
w0.m = matrix(0,nrow=length(beta1.v),ncol=nL);
w0.m[which(beta1.v <= th1.v[1]),1] = 1;
w0.m[intersect(which(beta1.v > th1.v[1]),which(beta1.v <= th1.v[2])),2] = 1;
w0.m[which(beta1.v > th1.v[2]),3] = 1;

### fit type1
print("Fitting EM beta mixture to goldstandard probes");
set.seed(1)
rand.idx = sample(1:length(beta1.v),min(c(nfit, length(beta1.v))  )   ,replace=FALSE)
em1.o = blc2(Y=matrix(beta1.v[rand.idx],ncol=1),w=w0.m[rand.idx,],maxiter=niter,tol=tol);
subsetclass1.v = apply(em1.o$w,1,which.max);
subsetth1.v = c(mean(max(beta1.v[rand.idx[subsetclass1.v==1]]),min(beta1.v[rand.idx[subsetclass1.v==2]])),mean(max(beta1.v[rand.idx[subsetclass1.v==2]]),min(beta1.v[rand.idx[subsetclass1.v==3]],na.rm=TRUE)));
class1.v = rep(2,length(beta1.v));
class1.v[which(beta1.v < subsetth1.v[1])] = 1;
class1.v[which(beta1.v > subsetth1.v[2])] = 3;
nth1.v = subsetth1.v;
print("Done");

### generate plot from estimated mixture
if(plots){
print("Check");
tmpL.v = as.vector(rmultinom(1:nL,length(beta1.v),prob=em1.o$eta));
tmpB.v = vector();
for(l in 1:nL){
  tmpB.v = c(tmpB.v,rbeta(tmpL.v[l],em1.o$a[l,1],em1.o$b[l,1]));
}

pdf(paste("Type1fit-",sampleID,".pdf",sep=""),width=6,height=4);
plot(density(beta1.v));
d.o = density(tmpB.v);
points(d.o$x,d.o$y,col="green",type="l")
legend(x=0.5,y=3,legend=c("obs","fit"),fill=c("black","green"),bty="n");
dev.off();
}



### Estimate Modes 
if (  sum(class1.v==1)==1 ){ mod1U= beta1.v[class1.v==1]}
if (  sum(class1.v==3)==1 ){ mod1M= beta1.v[class1.v==3]}
if (  sum(class1.v==1) >1){ 
d1U.o = density(beta1.v[class1.v==1])
mod1U = d1U.o$x[which.max(d1U.o$y)]
}
if (  sum(class1.v==3)>1 ){ 
d1M.o = density(beta1.v[class1.v==3])
mod1M = d1M.o$x[which.max(d1M.o$y)]
}


d2U.o = density(beta2.v[which(beta2.v<0.4)]);
d2M.o = density(beta2.v[which(beta2.v>0.6)]);
mod2U = d2U.o$x[which.max(d2U.o$y)]
mod2M = d2M.o$x[which.max(d2M.o$y)]


### now deal with type2 fit
th2.v = vector();
th2.v[1] = nth1.v[1] + (mod2U-mod1U);
th2.v[2] = nth1.v[2] + (mod2M-mod1M);

### estimate initial weight matrix 
w0.m = matrix(0,nrow=length(beta2.v),ncol=nL);
w0.m[which(beta2.v <= th2.v[1]),1] = 1;
w0.m[intersect(which(beta2.v > th2.v[1]),which(beta2.v <= th2.v[2])),2] = 1;
w0.m[which(beta2.v > th2.v[2]),3] = 1;

print("Fitting EM beta mixture to input probes");
set.seed(1)
rand.idx = sample(1:length(beta2.v),min(c(nfit, length(beta2.v)),na.rm=TRUE)   ,replace=FALSE)
em2.o = blc2(Y=matrix(beta2.v[rand.idx],ncol=1),w=w0.m[rand.idx,],maxiter=niter,tol=tol);
print("Done");

### for type II probes assign to state (unmethylated, hemi or full methylation)
subsetclass2.v = apply(em2.o$w,1,which.max);



if (sum(subsetclass2.v==2)>0 ){
subsetth2.v = c(mean(max(beta2.v[rand.idx[subsetclass2.v==1]]),min(beta2.v[rand.idx[subsetclass2.v==2]])),
mean(max(beta2.v[rand.idx[subsetclass2.v==2]]),min(beta2.v[rand.idx[subsetclass2.v==3]])));
}
if (sum(subsetclass2.v==2)==0 ){
subsetth2.v = c(1/2*max(beta2.v[rand.idx[subsetclass2.v==1]])+ 1/2*mean(beta2.v[rand.idx[subsetclass2.v==3]]), 1/3*max(beta2.v[rand.idx[subsetclass2.v==1]])+ 2/3*mean(beta2.v[rand.idx[subsetclass2.v==3]]));
}


class2.v = rep(2,length(beta2.v));
class2.v[which(beta2.v <= subsetth2.v[1])] = 1;
class2.v[which(beta2.v >= subsetth2.v[2])] = 3;


### generate plot
if(plots){
tmpL.v = as.vector(rmultinom(1:nL,length(beta2.v),prob=em2.o$eta));
tmpB.v = vector();
for(lt in 1:nL){
  tmpB.v = c(tmpB.v,rbeta(tmpL.v[lt],em2.o$a[lt,1],em2.o$b[lt,1]));
}
pdf(paste("Type2fit-",sampleID,".pdf",sep=""),width=6,height=4);
plot(density(beta2.v));
d.o = density(tmpB.v);
points(d.o$x,d.o$y,col="green",type="l")
legend(x=0.5,y=3,legend=c("obs","fit"),fill=c("black","green"),bty="n");
dev.off();
}

classAV1.v = vector();classAV2.v = vector();
for(l in 1:nL){
  classAV1.v[l] =  em1.o$mu[l,1];
  classAV2.v[l] =  em2.o$mu[l,1];
}

### start normalising input probes
print("Start normalising input probes");
nbeta2.v = beta2.v;
### select U probes
lt = 1;
selU.idx = which(class2.v==lt);
selUR.idx = selU.idx[which(beta2.v[selU.idx] > classAV2.v[lt])];
selUL.idx = selU.idx[which(beta2.v[selU.idx] < classAV2.v[lt])];
### find prob according to typeII distribution
p.v = pbeta(beta2.v[selUR.idx],em2.o$a[lt,1],em2.o$b[lt,1],lower.tail=FALSE);
### find corresponding quantile in type I distribution
q.v = qbeta(p.v,em1.o$a[lt,1],em1.o$b[lt,1],lower.tail=FALSE);
nbeta2.v[selUR.idx] = q.v;
p.v = pbeta(beta2.v[selUL.idx],em2.o$a[lt,1],em2.o$b[lt,1],lower.tail=TRUE);
### find corresponding quantile in type I distribution
q.v = qbeta(p.v,em1.o$a[lt,1],em1.o$b[lt,1],lower.tail=TRUE);
nbeta2.v[selUL.idx] = q.v;

### select M probes
lt = 3;
selM.idx = which(class2.v==lt);
selMR.idx = selM.idx[which(beta2.v[selM.idx] > classAV2.v[lt])];
selML.idx = selM.idx[which(beta2.v[selM.idx] < classAV2.v[lt])];
### find prob according to typeII distribution
p.v = pbeta(beta2.v[selMR.idx],em2.o$a[lt,1],em2.o$b[lt,1],lower.tail=FALSE);
### find corresponding quantile in type I distribution
q.v = qbeta(p.v,em1.o$a[lt,1],em1.o$b[lt,1],lower.tail=FALSE);
nbeta2.v[selMR.idx] = q.v;


if(doH){ ### if TRUE also correct type2 hemimethylated probes
### select H probes and include ML probes (left ML tail is not well described by a beta-distribution).
lt = 2;
selH.idx = c(which(class2.v==lt),selML.idx);
minH = min(beta2.v[selH.idx],na.rm=TRUE)
maxH = max(beta2.v[selH.idx],na.rm=TRUE)
deltaH = maxH - minH;
#### need to do some patching
deltaUH = -max(beta2.v[selU.idx],na.rm=TRUE) + min(beta2.v[selH.idx],na.rm=TRUE)
deltaHM = -max(beta2.v[selH.idx],na.rm=TRUE) + min(beta2.v[selMR.idx],na.rm=TRUE)

## new maximum of H probes should be
nmaxH = min(nbeta2.v[selMR.idx],na.rm=TRUE) - deltaHM;
## new minimum of H probes should be
nminH = max(nbeta2.v[selU.idx],na.rm=TRUE) + deltaUH;
ndeltaH = nmaxH - nminH;

### perform conformal transformation (shift+dilation)
## new_beta_H(i) = a + hf*(beta_H(i)-minH);
hf = ndeltaH/deltaH ;
### fix lower point first
nbeta2.v[selH.idx] = nminH + hf*(beta2.v[selH.idx]-minH);

}

pnbeta.v = beta.v;
pnbeta.v[type1.idx] = beta1.v;
pnbeta.v[type2.idx] = nbeta2.v;

### generate final plot to check normalisation
if(plots){
 print("Generating final plot");
 d1.o = density(beta1.v);
 d2.o = density(beta2.v);
 d2n.o = density(nbeta2.v);
 ymax = max(d2.o$y,d1.o$y,d2n.o$y);
 pdf(paste("CheckBMIQ-",sampleID,".pdf",sep=""),width=6,height=4)
 plot(density(beta2.v),type="l",ylim=c(0,ymax),xlim=c(0,1));
 points(d1.o$x,d1.o$y,col="red",type="l");
 points(d2n.o$x,d2n.o$y,col="blue",type="l");
 legend(x=0.5,y=ymax,legend=c("type1","type2","type2-BMIQ"),bty="n",fill=c("red","black","blue"));
 dev.off();
}

print(paste("Finished for sample ",sampleID,sep=""));

return(list(nbeta=pnbeta.v,class1=class1.v,class2=class2.v,av1=classAV1.v,av2=classAV2.v,hf=hf,th1=nth1.v,th2=th2.v));

}

CheckBMIQ = function(beta.v,design.v,pnbeta.v){### pnbeta is BMIQ normalised profile

type1.idx = which(design.v==1);
type2.idx = which(design.v==2);

beta1.v = beta.v[type1.idx];
beta2.v = beta.v[type2.idx];
pnbeta2.v = pnbeta.v[type2.idx];
  
} # end of function CheckBMIQ
















CalibrateUnitInterval=function(datM,onlyIfOutside=TRUE){

rangeBySample=data.frame(lapply(data.frame(t(datM)),range,na.rm=TRUE))
minBySample=as.numeric(rangeBySample[1,])
maxBySample=as.numeric(rangeBySample[2,])
if (onlyIfOutside) { indexSamples=which((minBySample<0 | maxBySample>1) & !is.na(minBySample) & !is.na(maxBySample))
   }
if (!onlyIfOutside) { indexSamples=1:length(minBySample)}
if ( length(indexSamples)>=1 ){
for ( i in indexSamples) {
y1=c(0.001,0.999) 
x1=c(minBySample[i],maxBySample[i])
lm1=lm( y1 ~ x1 )
intercept1=coef(lm1)[[1]]
slope1=coef(lm1)[[2]]
datM[i,]=intercept1+slope1*datM[i,]
} # end of for loop
}
datM
} #end of function for calibrating to [0,1]

p.qqplot<-function(pvalues,outf="qq",outfmt="jpg",draw.conf=TRUE,
        conf.col="lightgray",conf.alpha=.95,pch=20,col="black",reducesize=0,...)
{

    conf.alpha=1-conf.alpha
    ## Calculate lambda
    lambda <- qchisq(median(pvalues,na.rm=T), df = 1, lower.tail = F)/qchisq(0.5, 1)
    cat("\nThe genomic inflation factor lambda=",lambda,"\n")

    pvalues=pvalues[!is.na(pvalues)]
    if(!is.numeric(pvalues)){stop("pvalues is not a numeric vector")}
    if(min(pvalues)<=0){stop("Some pvalues is less or equal to 0")}
    if(max(pvalues)>1){stop("Some pvalues is greater than 1")}

    pvalues=sort(pvalues)
    n=length(pvalues)+1
    exp.x <- -log10((1:(n-1))/n)
    pvalues <- -log10(pvalues)

    if(draw.conf){
        conf.points = n-1;
        cpts<-matrix(nrow=conf.points*2, ncol=2)
        for(i in seq(from=1, to=conf.points)) {
           cpts[i,1]<- -log10((i)/n)
           cpts[i,2]<- -log10(qbeta(1-conf.alpha/2, i, n-i))
           cpts[conf.points*2+1-i,1]<- -log10((i)/n)
           cpts[conf.points*2+1-i,2]<- -log10(qbeta(conf.alpha/2, i, n-i))
        }
    }

#    xlab=expression(Expected~~-log[10](italic(P))~~"under NULL hypothesis")
#    ylab=expression(Observed~~-log[10](italic(P)))

     xlab=bquote('Expected  '*'-log'['10']*'(P)'*'  under NULL hypothesis')
     ylab=bquote('Observed  '*'-log'['10']*'(P)')

if(outfmt == "jpg"){
jpeg(paste(outf,".",outfmt,sep=""),height=600,width=600,quality=100)
par(mar=c(5.1, 5.1, 4.1, 2.1))
}else if(outfmt == "eps"){
grDevices::postscript(paste(outf,".",outfmt,sep=""),width=3.5,height=3.5,
    paper = "special",pointsize=5,onefile=FALSE)
par(mar=c(5.1, 5.1, 4.1, 2.1))
while(reducesize>0){
dat=data.frame(ex=exp.x,p=pvalues)
dat=dat[order(dat$ex),]
dat$id=1:nrow(dat)
dat$v1=dat$ex/max(dat$ex)
dat$v2=dat$p/max(dat$p)
dat$v3=c(2,abs(dat$v1[2:nrow(dat)]-dat$v1[1:(nrow(dat)-1)])+
  abs(dat$v2[2:nrow(dat)]-dat$v2[1:(nrow(dat)-1)]))
dat$V3[nrow(dat)]=2
dat=dat[order(dat$v3),]
exid=dat$id[1:round(nrow(dat)/2)]
exid=exid[exid%%2==0]
dat=dat[!(dat$id %in% exid),]
reducesize=reducesize-1
exp.x=dat$ex
pvalues=dat$p
}
}else{stop("Outfmt should be jpg or eps\n")}

    plot(c(0,0),type="n",xlim=c(0,max(exp.x)), ylim=c(0,max(pvalues)), xlab=xlab,ylab=ylab,cex.lab=1.3,...)
    if(draw.conf){polygon(x=cpts[,1],y=cpts[,2], col="lightgray", lty=0)}
    points(exp.x,pvalues,pch=pch,col=col)
    abline(0,1,lty=2)
    dev.off()
}

library(ENmix)
library(minfiData)
library(digest)

digestMatrix <- function(mat) {
    content <- sprintf("%.6f", mat)
    content[content == "-0.000000"] <- "0.000000"
    digest(c(content, rownames(mat), colnames(mat)))
}

#read in raw intensity data
sheet <- read.metharray.sheet(file.path(find.package("minfiData"),
        "extdata"), pattern = "csv$")
rgSet <- read.metharray.exp(targets = sheet,extended = TRUE)
qc<-QCinfo(rgSet)
mdat <- preprocessRaw(rgSet)
mdat.bg=preprocessENmix(rgSet,bgParaEst="oob",nCores=6)
mdat.filter=QCfilter(mdat,qcinfo=qc, samplethre = 0.05, CpGthre = 0.05)
mdat.quantile=normalize.quantile.450k(mdat,method="quantile1")

testDigests <- list(
    qc = list(detP = digestMatrix(qc$detP),
      nbead=digestMatrix(qc$nbead)),
   enmix = list(Meth = digestMatrix(getMeth(mdat.bg)),
      Unmeth = digestMatrix(getUnmeth(mdat.bg))),
    filter = list(Meth = digestMatrix(getMeth(mdat.filter)),
      Unmeth = digestMatrix(getUnmeth(mdat.filter))),
    quantile = list(Meth = digestMatrix(getMeth(mdat.quantile)),
      Unmeth = digestMatrix(getUnmeth(mdat.quantile)))
    )

save(testDigests, file = "../unitTests/testDigests.rda")

gc()
sessionInfo()

rm(list = ls())

test_enmix <- function() {
    stopifnot(require(minfiData))
    stopifnot(require(digest))
    load(file.path(path.package("ENmix"), "unitTests", "testDigests.rda"))

    digestMatrix <- function(mat) {
        content <- sprintf("%.6f", mat)
        content[content == "-0.000000"] <- "0.000000"
        digest(c(content, rownames(mat), colnames(mat)))
    }

    ##read in raw intensity data
    sheet <- read.metharray.sheet(file.path(find.package("minfiData"),
        "extdata"), pattern = "csv$")
##to save time for package build
#    rgSet <- read.metharray.exp(targets = sheet,extended = TRUE)
#    qc<-QCinfo(rgSet)
#    mdat <- preprocessRaw(rgSet)
#    mdat.bg=preprocessENmix(rgSet,bgParaEst="oob",nCores=6)
#    mdat.filter=QCfilter(mdat,qcinfo=qc, samplethre = 0.05, CpGthre = 0.05)
#    mdat.quantile=normalize.quantile.450k(mdat,method="quantile1")

#    checkEquals(testDigests$qc$detP, digestMatrix(qc$detP))
#    checkEquals(testDigests$qc$nbead, digestMatrix(qc$nbead))
#    checkEquals(testDigests$enmix$Meth, digestMatrix(getMeth(mdat.bg)))
#    checkEquals(testDigests$enmix$Unmeth, digestMatrix(getUnmeth(mdat.bg)))
#    checkEquals(testDigests$filter$Meth, digestMatrix(getMeth(mdat.filter)))
#    checkEquals(testDigests$filter$Unmeth, digestMatrix(getUnmeth(mdat.filter)))
#    checkEquals(testDigests$quantile$Meth, digestMatrix(getMeth(mdat.quantile)))
#    checkEquals(testDigests$quantile$Unmeth, digestMatrix(getUnmeth(mdat.quantile)))
}

.guessArrayTypes <- function(nProbes) {
    if (nProbes >= 622000 && nProbes <= 623000) {
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylation450k",
            annotation = "ilmn12.hg19")
    } else if (nProbes >= 1032000 && nProbes <= 1033000) {
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylationEPIC",
            annotation = "ilm10b2.hg19")
    } else if (nProbes >= 1105000 && nProbes <= 1105300) {
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylationEPICv2",
            annotation = "20a1.hg38")
    } else if (nProbes >= 55200 && nProbes <= 55400) {
        arrayAnnotation <- c(
            array = "IlluminaHumanMethylation27k",
            annotation = "ilmn12.hg19")
    } else {
        arrayAnnotation <- c(array = "Unknown", annotation = "Unknown")
    }
    arrayAnnotation
}

readidat <- function(path = NULL,manifestfile=NULL,recursive = TRUE, verbose = FALSE, force=FALSE) {
#    if(!require("illuminaio")){stop("Can not load in illuminaio package")}
    Grn.files <- list.files(path = path, pattern = "_Grn.idat$", recursive = recursive,
        ignore.case = FALSE, full.names = TRUE)
    Red.files <- list.files(path = path,pattern = "_Red.idat$",recursive = recursive,
        ignore.case = FALSE, full.names = TRUE)
    if (length(Grn.files) == 0 || length(Red.files) == 0) {
        stop("[readidat] No files with suffix _Grn.idat or _Red.idat were found")
    }else{message("[readidat] Found ",length(Grn.files)," files with suffix _Grn.idat")
          message("[readidat] Found ",length(Red.files)," files with suffix _Red.idat")}
    commonFiles <- intersect(
        x = sub("_Grn.idat$", "", Grn.files),
        y = sub("_Red.idat$", "", Red.files))
    if (length(commonFiles) == 0) {
        stop("[readidat] No IDAT files with both Red and Green channel were found")
    }
    basenames = commonFiles

    G.files <- paste(basenames, "_Grn.idat", sep = "")
    names(G.files) <- basename(basenames)
    R.files <- paste(basenames, "_Red.idat", sep = "")
    names(R.files) <- basename(basenames)

    G.Quants <- lapply(G.files, function(xx) {
        tr=try(Quants <- readIDAT(xx)[["Quants"]])
	if(class(tr)[1] == "try-error"){stop(paste0("ERROR: file reading error, check or exclude/remove file: ",xx))}
        if (verbose) message("[readidat] Loaded ", basename(xx), ", ",nrow(Quants)," probes")
        Quants[,c("Mean","NBeads")]
    })

    allNProbes <- vapply(G.Quants, nrow, integer(1L))
    arrayTypes <- cbind(do.call(rbind, lapply(allNProbes, .guessArrayTypes)),
                        size = allNProbes)
    sameLength <- (length(unique(allNProbes)) == 1)
    sameArray <- (length(unique(arrayTypes[, "array"])) == 1)
    if(!sameArray){
        print(arrayTypes)
        stop("[readidat] Samples are from different array")
     }else{
      if (!sameLength){
          print(allNProbes)
        if(!force){
          stop("[readidat] Array size are different.  \n Set force=TRUE to force data merging\n")
        }else{message("Arrays with different sizes are forced to merge together\n")}
    }
           arraytype=as.character(arrayTypes[,"array"][1])
           annotation=as.character(arrayTypes[,"annotation"][1])
    }

    commonAddresses <- as.character(
        Reduce("intersect", lapply(G.Quants, rownames)))



    if (verbose) {message("[readidat] Loading array manifestfile...")}
    if(!is.null(manifestfile))
    {
#      manifestfilepath=manifestfile 
#      manifestfile=base::strsplit(manifestfilepath,split="/")[[1]]
#      if(length(manifestfile)>1){
#         manifestfile=manifestfile[length(manifestfile)]
#         if (file.exists(manifestfile))file.remove(manifestfile)
#         system(paste("wget ",manifestfilepath,sep=""))
#      }
      manifest = readmanifest(manifestfile)
    }else{
      manifestpkg=paste(arraytype,"manifest",sep="")
#    manifest = getmanifest (manifestpkg)
      manifest = getmanifest (arraytype,annotation)
    }
    tmp=c(manifest$assay[,"Address"],manifest$ictrl[,"Address"])
    diff=setdiff(tmp,commonAddresses)
    if(length(diff)>8000){stop("IDAT file and manifest file are differed
      by more than 8000 probes")}else{
    commonAddresses=commonAddresses[commonAddresses %in% tmp]}

    if (verbose) {message("[readidat] Creating data matrices at green channel...")}
    GrnMean <- do.call(cbind,
        lapply(G.Quants, function(xx) xx[commonAddresses, "Mean"]))
    if (verbose) {message("[readidat] Creating data matrices for NBeads...")}
    NBeads <- do.call(cbind,
        lapply(G.Quants, function(xx) xx[commonAddresses, "NBeads"]))
    rm(G.Quants)
    
    R.Quants <- lapply(R.files, function(xx) {
        tr=try(Quants <- readIDAT(xx)[["Quants"]])
	if(class(tr)[1] == "try-error"){stop(paste0("ERROR: file reading error, check or exclude/remove file: ",xx))}
        if (verbose) message("[readidat] Loaded ", basename(xx), ", ",nrow(Quants)," probes")
        Quants[,c("Mean")]
    })
    commonAddresses <- as.character(
        Reduce("intersect", lapply(R.Quants, names)))

    tmp=c(manifest$assay[,"Address"],manifest$ictrl[,"Address"])
    diff=setdiff(tmp,commonAddresses)
    if(length(diff)>8000){stop("IDAT file and manifest file are differed
      by more than 8000 probes")}else{
    commonAddresses=commonAddresses[commonAddresses %in% tmp]}

    if(!identical(rownames(GrnMean),commonAddresses)){
      commonAddresses=intersect(rownames(GrnMean),commonAddresses)
      GrnMean=GrnMean[commonAddresses,]
      NBeads=NBeads[commonAddresses,]
    }

    if (verbose) {message("[readidat] Creating data matrices at red channel...")}
    RedMean <- do.call(cbind,
        lapply(R.Quants, function(xx) xx[commonAddresses]))
    message("[readidat] Red Green Channel contains ",nrow(GrnMean)," probes and ",
        ncol(GrnMean)," Samples")

    tmp=manifest$ictrl
    tmp=tmp[tmp$Address %in% commonAddresses,]
    ictrl.anno=tmp
    tmp=manifest$assay[1:nrow(ictrl.anno),]
    tmp$Name=paste("ictrl",1:nrow(ictrl.anno),sep="")
    tmp$Address=ictrl.anno$Address
    tmp$Infinium_Design_Type="ctrl"
    tmp[,-which(names(tmp) %in% c("Name","Address","Infinium_Design_Type"))]=NA

    assay.anno=manifest$assay
    assay.anno=rbind(assay.anno,tmp)
    rownames(assay.anno)=assay.anno$Address
    assay.anno=assay.anno[rownames(GrnMean),]

    rownames(ictrl.anno)=ictrl.anno$Address

#check
    flag=!(assay.anno$Infinium_Design_Type %in% c("IA","IB","snpIA","snpIB") & 
                     !(assay.anno$Color_Channel %in% c("Red","Grn")))
    RedMean=RedMean[flag,]
    GrnMean=GrnMean[flag,]
    NBeads=NBeads[flag,]
    assay.anno=assay.anno[flag,]
    
    rgSet <- rgDataSet(
        Red = RedMean,
        Green = GrnMean,
        NBeads = NBeads,
	rowData=as(assay.anno,"DataFrame"),
	ictrl=as(ictrl.anno,"DataFrame")
    )
    metadata(rgSet)$Array=arraytype
    metadata(rgSet)$annotation=annotation
    rgSet
}

#define a few global variables
if (getRversion() >= "2.15.1") {
  utils::globalVariables(c(
    "DNAmFitnessModels",
    "EpiToc_CpGs",
    "GrimAgeV1_model",
    "GrimAgeV2_model",
    "GrimAge_CpG",
    "MiAgedat",
    "bAge_CpG",
    "cAge_CpG",
    "episcore_model",
    "libraryDataGet",
    "mPOA_Models",
    "methscore_dict",
    "pcc1",
    "pcc2",
    "pcc3",
    "pcc4",
    "refmeth",
    "refmeth2",
    "AdaptAge_model", 
    "CausAge_model", 
    "DamAge_model"
  ))
}

qcfilter <-function(mat,qcscore=NULL,rmoutlier=TRUE,byrow=TRUE,detPthre=0.000001,nbthre=3,
     rmcr=FALSE,rthre=0.05,cthre=0.05,impute=FALSE,imputebyrow=TRUE,fastimpute=FALSE,...)
{
    if(impute){requireNamespace("impute")}
    if(!(is.numeric(mat) & is.matrix(mat))){stop("Input data must be a
     numeric matrix")}

    #If qcsocre is not NULL, low quality data points will be filtered
    if(is.null(qcscore)){}else if((sum(!(rownames(mat) %in%
     rownames(qcscore$detP))) + 
    sum(!(colnames(mat) %in% colnames(qcscore$detP))))>0){
    stop("Wrong qcscore matrix, please check...\n")}else{
    temp <- qcscore$nbead<nbthre | qcscore$detP>detPthre
    temp=temp[rownames(mat),]
    temp=temp[,colnames(mat)]
    mat[temp]=NA
    }
    #remove outliers
    if(rmoutlier){
    if(!byrow){mat=t(mat)}
    q2575 <- apply(mat,1,function(x) quantile(x, probs=c(0.25,0.75),
    na.rm=TRUE))
    qr <- q2575["75%",]-q2575["25%",]
    flag=mat < (q2575["25%",] - 3*qr) | mat > (q2575["75%",] + 3*qr)
    mat[flag] <- NA
    if(!byrow){mat=t(mat)}
    }
    #remove rows and columns that have too many missing values
    if(rmcr){
    if(nrow(mat)>ncol(mat)){
    thre=min(2*cthre,0.3)
    cpercna=apply(is.na(mat),2,sum)/nrow(mat)
    tmat=mat[,cpercna<thre]
    rpercna=apply(is.na(tmat),1,sum)/ncol(tmat)
    rthre=max(rthre,min(3,ncol(tmat))/ncol(tmat))
    tmat=mat[rpercna<rthre,]
    nrout=sum(rpercna>=rthre)
    cpercna=apply(is.na(tmat),2,sum)/nrow(tmat)
    cthre=max(cthre,min(3,nrow(tmat))/nrow(tmat))
    mat=tmat[,cpercna<cthre]
    ncout=sum(cpercna>=cthre)
    cat(nrout," rows with percentage of missing data greater than ",rthre,
    " were excluded\n")
    cat(ncout," columns with percentage of missing data greater than ",cthre,
    " were excluded\n")
    }else{
    thre=min(2*rthre,0.3)
    rpercna=apply(is.na(mat),1,sum)/ncol(mat)
    tmat=mat[rpercna<thre,]
    cpercna=apply(is.na(tmat),2,sum)/nrow(tmat)
    cthre=max(cthre,min(3,nrow(tmat))/nrow(tmat))
    tmat=mat[,cpercna<cthre]
    ncout=sum(cpercna>=cthre)
    rpercna=apply(is.na(tmat),1,sum)/ncol(tmat)
    rthre=max(rthre,min(3,ncol(tmat))/ncol(tmat))
    mat=tmat[rpercna<rthre,]
    nrout=sum(rpercna>=rthre)
    cat(nrout," rows with percentage of missing data greater than ",rthre,
    " were excluded\n")
    cat(ncout," columns with percentage of missing data greater than ",cthre,
    " were excluded\n")
    }
    }

    #impute missing data using knn method
    if(impute)
    {
    if(imputebyrow){mat=t(mat)}

if(fastimpute){
    mm=apply(mat,2,function(x)median(x,na.rm=TRUE))
    idx=which(is.na(mat),arr.ind = TRUE)
    for(i in 1:nrow(idx)){mat[idx[i,1],idx[i,2]]=mm[idx[i,2]]}
}else{
psize=100000
if(ncol(mat)>psize){
    npart=floor(ncol(mat)/psize)
    for(i in 1:npart){
       n1=psize*(i-1)+1
       n2=psize*i
       if(i==npart){n2=ncol(mat)}
       mat1=mat[,n1:n2]
       resu=impute::impute.knn(mat1,...)
       mat[,n1:n2]=resu$data
    }
}else{
      resu=impute::impute.knn(mat,...)
      mat=resu$data
}}

    if(imputebyrow){mat=t(mat)}
    }
    mat
}

bincount <- function(x,breaks)
{
    x <- x[!is.na(x)]
    bc <- table(.bincode(x, breaks, TRUE, TRUE))
    temp=data.frame(id=c(1: (length(breaks)-1)))
    bc=data.frame(id=as.numeric(names(bc)),counts=as.numeric(bc))
    resu=merge(temp,bc,by="id",all.x=TRUE)
    resu$counts[is.na(resu$counts)]=0
    resu$counts[order(resu$id)]
}

multifreqpoly <- function(mat, nbreaks=100, col=1:ncol(mat), xlab="", 
    ylab="Frequency", legend = list(x = "top", fill=col,
    legend = if(is.null(colnames(mat))) paste(1:ncol(mat)) else 
    colnames(mat)),append=FALSE,...)
{
    if(!is.matrix(mat)) stop("Warning: input data is not a numeric matrix\n")
    if(is.null(col)) col="black"
    col=rep(col,ceiling(ncol(mat)/length(col)))
    if(nbreaks > nrow(mat)) nbreaks=min(15,round(nrow(mat)/2))

    breaks <- seq(min(mat,na.rm=TRUE), max(mat,na.rm=TRUE), 
      diff(range(mat,na.rm=TRUE))/nbreaks)
    mids <- 0.5 * (breaks[-1] + breaks[-length(breaks)])
    counts <- sapply(data.frame(mat),bincount,breaks=breaks)
    if(!append){plot(range(mids),c(0,max(counts)),type="n",xlab=xlab,ylab=ylab,...)}
    for(i in 1:ncol(counts)){lines(mids,counts[,i],col=col[i],...)}
    if(is.list(legend)) do.call(graphics::legend, legend)
}

freqpoly <- function(mat, nbreaks=15, col="black", xlab="", ylab="Frequency",
     type="l",append=FALSE,...)
{
    if(!is.numeric(mat)) stop("Warning: input data is not a numeric vector\n")
    if(nbreaks > length(mat)) nbreaks=min(15,round(length(mat)/2))

    breaks <- seq(min(mat,na.rm=TRUE), max(mat,na.rm=TRUE),
      diff(range(mat,na.rm=TRUE))/nbreaks)
    mids <- 0.5 * (breaks[-1] + breaks[-length(breaks)])
    counts <- bincount(mat,breaks=breaks)
    if(!append){plot(range(mids),c(0,max(counts)),type="n",xlab=xlab,
    ylab=ylab,...)}
    lines(mids,counts,col=col,type=type,...)
}

dupicc<-function(dat,dupid,mvalue=FALSE,center=TRUE,nCores=2,qcflag=FALSE,qc=NULL,
       detPthre=0.05,nbthre=3,skipicc=FALSE,corfig=FALSE,model="oneway")
{
if(nCores>detectCores()){nCores=detectCores()}
dupid=data.frame(dupid)

if(mvalue){dat=B2M(dat)}

if(qcflag){
	if(is.null(qc)){stop("Please provide ENmix QCinfo object for quality control")} 
	dat=dat[,!(colnames(dat) %in% qc$badsample)]
	dat=dat[!(rownames(dat) %in% qc$badCpG),]
	qc$detP=qc$detP[rownames(dat),]
	qc$detP=qc$detP[,colnames(dat)]
	qc$nbead=qc$nbead[,colnames(dat)]
	qc$nbead=qc$nbead[rownames(dat),]
	dat[qc$detP>detPthre | qc$nbead<nbthre]<-NA
}      

flag=dupid$id1 %in% colnames(dat) & dupid$id2 %in% colnames(dat) 
dupid=dupid[flag,]

#Shrout & Fleiss (1979), ICC(3,1)
#modified from irr package
#twoway, consistency
icc <-function(meth,model) {
  r0 = 0; conf.level = .95
  meth <- as.matrix(na.omit(meth))
  alpha <- 1-conf.level
  ns <- nrow(meth)
  nr <- ncol(meth)
  SStotal <- var(as.numeric(meth))*(ns*nr-1)
  MSr <- var(apply(meth,1,mean))*nr
  MSw <- sum(apply(meth,1,var)/ns)
  MSc <- var(apply(meth,2,mean))*ns
  MSe <- (SStotal-MSr*(ns-1)-MSc*(nr-1))/((ns-1)*(nr-1))

  if (model=="oneway") {
#Asendorpf & Wallbott, S. 245, ICu; #Bartko (1966)
  coeff  <- (MSr-MSw)/(MSr+(nr-1)*MSw)
  Fvalue <- MSr/MSw*((1-r0)/(1+(nr-1)*r0))
  df1    <- ns-1
  df2    <- ns*(nr-1)
  p.value <- pf(Fvalue, df1, df2, lower.tail=FALSE)
 }else if (model=="twoway") {
  coeff  <- (MSr-MSe)/(MSr+(nr-1)*MSe)
  Fvalue <- MSr/MSe*((1-r0)/(1+(nr-1)*r0))
  df1    <- ns-1
  df2    <- (ns-1)*(nr-1)
  p.value <- pf(Fvalue, df1, df2, lower.tail=FALSE)}
  list(value = coeff,Fvalue = Fvalue,p.value = p.value)
}

icc_all <-function(probe,mat1,mat2)
{
        fit=try(icc(meth=cbind(mat1[probe,],mat2[probe,]),model=model))
        if(class(fit)[1] == "try-error"){return(c(probe,NA,NA,NA))}else{
        icc=fit$value
        p=fit$p.value
        return(c(probe,icc,p))
        }
}

dcor<-function(b1,b2)
{
cc=array()
dist=array()
for(i in 1:ncol(b1))
{
	x1=b1[,i];x2=b2[,i]
	flag=!(is.na(x1) | is.na(x2))
	x1=x1[flag];x2=x2[flag]
        cc[i]=cor(x1,x2,method="pearson",use="pairwise.complete.obs")
        dist[i]=sum(abs(x1-x2))/length(x1)
}
resu=data.frame(id1=colnames(b1),id2=colnames(b2),cor=cc,diff=dist)
resu
}

if(!skipicc){
dupid=dupid[dupid$id1 %in% colnames(dat) & dupid$id2 %in% colnames(dat),]
dat1=dat[,as.vector(dupid$id1)]
dat2=dat[,as.vector(dupid$id2)]
resu1=mclapply(dimnames(dat1)[[1]],icc_all,mat1=dat1,mat2=dat2, mc.cores=nCores)
resu <- do.call(rbind, lapply(resu1, unlist))
resu=as.data.frame(resu)
names(resu)=c("probe","icc","p")
resu$icc=as.numeric(as.vector(resu$icc))
resu$p=as.numeric(as.vector(resu$p))
}

if(center){dat=t(scale(t(dat),center=T,scale=F))}
    dat1=dat[,as.vector(dupid$id1)]
    dat2=dat[,as.vector(dupid$id2)]
    dupcor=dcor(dat1,dat2)
    if(corfig){
    nd1=cor(dat1,method="pearson",use="pairwise.complete.obs");nd1=nd1[lower.tri(nd1)]
    nd2=cor(dat2,method="pearson",use="pairwise.complete.obs");nd2=nd2[lower.tri(nd2)]
    nondupcor=c(nd1,nd2)
      jpeg("dupcorfig.jpg",height=800,width=500,quality=100)
      par(mfrow=c(2,1))
      plot(dupcor$cor,xlab="Duplicate pairs",ylab="Correlation",main="",ylim=c(-1,1))
      plot(nondupcor,xlab="Non-duplicate pairs",ylab="Correlation",main="",ylim=c(-1,1))
      dev.off()
    }

if(!skipicc){list(icc=resu,dupcor=dupcor)}else{dupcor}
}

plotCtrl <- function(rgSet,IDorder=NULL)
{
    if(!is(rgSet,"rgDataSet") & !is(rgSet, "RGChannelSet")){
      stop("object needs to be of class 'rgDataSet' or 'RGChannelSet'")}
    if(!is.null(IDorder))
    {
    if(sum(!(IDorder %in% colnames(rgSet)))>0)
    {
       idmissing=IDorder[!(IDorder %in% colnames(rgSet))]
       cat("The IDs were not found in the input data: ",idmissing,"\n")
       stop("Wrong ids in IDorder, please check")
    }else{
    rgSet=rgSet[,IDorder]
    }
    }

    if(is(rgSet,"rgDataSet")){
        ctrls <- getCGinfo(rgSet,type="ctrl")
    }else if(is(rgSet,"RGChannelSet")){
        ctrls <- getProbeInfo(rgSet,type="Control")}

    ctrls <- ctrls[ctrls$Address %in% rownames(rgSet),]
    ctrl_r <- assays(rgSet)$Red[ctrls$Address,]
    ctrl_g <- assays(rgSet)$Green[ctrls$Address,]
    contlid <- c("STAINING","EXTENSION","HYBRIDIZATION","TARGET REMOVAL",
     "BISULFITE CONVERSION I", "BISULFITE CONVERSION II","SPECIFICITY I",
     "SPECIFICITY II","NON-POLYMORPHIC","NEGATIVE",
     "NORM_A","NORM_C","NORM_G","NORM_T","NORM_ACGT")
    col <- as.vector(ctrls$Color)
    col[col == "-99"] <- NA
    col[col == "Aqua"] <- "aquamarine2"
    col[col == "Crimson"] <- "firebrick2"
    col[col == "Fuchsia"] <- "deeppink1"
    col[col == "Indigo"] <- "darkviolet"
    col[col == "Lime"] <- "yellowgreen"
    col[col == "Olive"] <- "darkolivegreen"
    col[col == "Silver"] <- "azure4"
    col[col == "Teal"] <- "cyan4"
    ctrls$Color <- col

for(ctype in contlid)
{
    fn <- ctype;fn <- gsub(" ","_",fn)
    cat("Plotting ",fn,".jpg","\n")
    jpeg(paste(fn,".jpg",sep=""),width=1100,height=500,quality=100)
    par(mfrow=c(1,2))
    if(ctype == "NORM_ACGT"){
    cc <- ctrls[ctrls$Type %in% c("NORM_A","NORM_C","NORM_G","NORM_T"),]}
    else{cc <- ctrls[ctrls$Type %in% ctype,]}
    red <- ctrl_r[cc$Address,]
    grn <- ctrl_g[cc$Address,]
    ymax <- max(red,grn)*1.01
if(ctype == "NEGATIVE")
{
    par(mar=c(5, 4, 4, 2))
    colnames(red) <- 1:ncol(red)
    colnames(grn) <- 1:ncol(grn)
    boxplot(grn,ylim=c(0,ymax),main=paste(ctype," Green",sep=""),bty="o",
    xlab="Sample",ylab="Intensity",cex.lab=1.2)
    boxplot(red,ylim=c(0,ymax),main=paste(ctype," Red",sep=""),bty="o",
    xlab="Sample",ylab="Intensity",cex.lab=1.2)
}
else if(ctype %in% c("NORM_A","NORM_C","NORM_G","NORM_T"))
{
    par(mar=c(5, 4, 4, 1))
    colnames(red)<-1:ncol(red)
    colnames(grn)<-1:ncol(grn)
    boxplot(grn,ylim=c(0,ymax),col=unique(as.vector(cc$Color)),main=
    paste(ctype," Green",sep=""),bty="o",xlab="Sample",ylab="Intensity",
    cex.lab=1.2)
    boxplot(red,ylim=c(0,ymax),col=unique(as.vector(cc$Color)),main=
    paste(ctype," Red",sep=""),bty="o",xlab="Sample",ylab="Intensity",
    cex.lab=1.2)
}
else if(ctype == "NORM_ACGT")
{
    par(mar=c(5, 4, 4, 8.2))
    label <- c("NORM_A","NORM_C","NORM_G","NORM_T")
    colcode <- c("Red","Green","Blue","Purple")
    idx <- t(replicate(nrow(red),1:ncol(red)));
    loc <- cc$Color;loc[loc == "Red"] <- -0.2
    loc[loc == "Green"] <- -0.1
    loc[loc == "Blue"] <- 0.1
    loc[loc == "Purple"]=0.2;loc=as.numeric(loc)
    loc1 <- matrix(rep(loc,ncol(red)),ncol=ncol(red));idx=idx+loc1
    plot(idx,grn,col=as.vector(cc$Color),ylim=c(0,ymax),main=paste(ctype,
    " Green",sep=""),bty="o",xlab="Sample",ylab="Intensity",cex.lab=1.2)
    legend(x=(ncol(grn)+0.2)*1.05,y=ymax*0.8,xjust = 0,yjust=0.5, bty="o",
    legend=label, col=colcode,xpd=TRUE,pch=15,cex=0.8)
    plot(idx,red,col=as.vector(cc$Color),ylim=c(0,ymax),main=paste(ctype,
    " Red",sep=""),bty="o",xlab="Sample",ylab="Intensity",cex.lab=1.2)
    legend(x=(ncol(red)+0.2)*1.05,y=ymax*0.8,xjust = 0,yjust=0.5, bty="o",
    legend=label, col=colcode,xpd=TRUE,pch=15,cex=0.8)
}
else
{
    par(mar=c(5, 4, 4, 8.2))
    idx <- t(replicate(nrow(red),1:ncol(red)))
    plot(idx,grn,col=as.vector(cc$Color),ylim=c(0,ymax),main=paste(ctype,
    " Green",sep=""),bty="o",xlab="Sample",ylab="Intensity",cex.lab=1.2)
    legend(x=ncol(grn)*1.05,y=ymax*0.8,xjust = 0,yjust=0.5, bty="o",legend=
    as.vector(cc$ExtendedType),col=as.vector(cc$Color),xpd=TRUE,pch=15,cex=0.8)
    plot(idx,red,col=as.vector(cc$Color),ylim=c(0,ymax),main=paste(ctype,
    " Red",sep=""),bty="o",xlab="Sample",ylab="Intensity",cex.lab=1.2)
    legend(x=ncol(red)*1.045,y=ymax*0.8,xjust = 0,yjust=0.5, bty="o",legend=
    as.vector(cc$ExtendedType),col=as.vector(cc$Color),xpd=TRUE,pch=15,cex=0.8)
}
    dev.off()
}
}

methyAge<-function(beta,fastImputation=FALSE,normalize=TRUE,nCores=2)
{
#source("mage_norm_function.R")
#require(sqldf)
#require(impute)
#require(flashClust)
#require(dynamicTreeCut)
#require(RPMM)

if(min(beta,na.rm=TRUE)<0 | max(beta,na.rm=TRUE)>1){
stop("Warning: Methylation beta value input should be within [0,1]")}

horvath=data.frame()
hannum=data.frame()
phenoage=data.frame()

load(system.file("mage_ref.RData",package="ENmix"))
names(refmeth)[which(names(refmeth)=="meth_mean")]="goldstandard"
#load(system.file("mPOA_Models.RData",package="ENmix"))
PACE=DunedinPACE(betas=beta, proportionOfProbesRequired = 0.8)

missing_horvath=as.vector(horvath$cg[-1])[!(as.vector(horvath$cg[-1]) %in% rownames(beta))]
missing_hannum=as.vector(hannum$cg)[!(as.vector(hannum$cg) %in% rownames(beta))]
missing_phenoage=as.vector(phenoage$cg[-1])[!(as.vector(phenoage$cg[-1]) %in% rownames(beta))]
missing=FALSE
if(length(missing_horvath)>0){missing=TRUE
cat(length(missing_horvath), "CpG missed methy age calculation using Horvath method: ",missing_horvath)}
if(length(missing_hannum)>0){missing=TRUE
cat(length(missing_hannum), "CpG missed methy age calculation using Hannum method: ",missing_hannum)}
if(length(missing_phenoage)>0){missing=TRUE
cat(length(missing_phenoage), "CpG missed methy age calculation using phenoage method: ",missing_phenoage)}

tmp=NULL
if(missing){
cat("Warning: Methylation values for missing CpGs will be imputed with reference values, but the results will be less acurate")
clockcg=unique(c(as.vector(horvath$cg)[-1],as.vector(hannum$cg),as.vector(phenoage$cg)[-1]))
missedcg=clockcg[!(clockcg %in%  rownames(beta))]
if(length(missedcg)>0){
  reference=refmeth[as.vector(refmeth$cg) %in% missedcg,]
  tmp=matrix(rep(reference$goldstandard,ncol(beta)),nrow=length(missedcg))
  rownames(tmp)=as.vector(reference$cg)
  colnames(tmp)=colnames(beta)
# beta=rbind(beta,tmp)
tmp=t(tmp)
}
}


refmeth=refmeth[as.vector(refmeth$cg) %in% rownames(beta),]
beta=beta[as.vector(refmeth$cg),]

#STEP 2: Imputing
set.seed(1)
nSamples=ncol(beta)
beta= t(beta)
noMissingPerSample=apply(as.matrix(is.na(beta)),1,sum)
if(!fastImputation & nSamples>1){
if(max(noMissingPerSample,na.rm=TRUE)>0 ){
dimnames1=dimnames(beta)
beta= data.frame(t(impute::impute.knn(t(beta))$data))
dimnames(beta)=dimnames1
}}
if(fastImputation | nSamples==1){
if(max(noMissingPerSample,na.rm=TRUE)>0){
dimnames1=dimnames(beta)
for (i in which(noMissingPerSample>0)){
selectMissing1=is.na(beta[i,])
beta[i,selectMissing1] = as.numeric(refmeth$goldstandard[selectMissing1])
}
dimnames(beta)=dimnames1
}}

#normalization using Horvath modified BMIQ method
if(normalize){
nCores=min(floor(nrow(beta)/2),nCores)
if(nCores>detectCores()){nCores <- detectCores()}
N=ceiling(nrow(beta)/(nCores))
parts=rep(1:nCores,each = N)[1:nrow(beta)]
bbb<-function(s,parts,dat){
  id=which(parts==s)
  dat1=dat[id,]
  BMIQcalibration(datM=dat1,goldstandard.beta= refmeth$goldstandard,plots=FALSE)
}
resu <- mclapply(1:nCores,bbb,parts=parts,dat=beta, mc.cores=nCores)
beta=as.matrix(do.call(rbind, resu))
}
beta=cbind(beta,tmp)

methyAge=data.frame(SampleID=rownames(beta))
datMethClock=beta[,as.character(horvath$cg[-1])]
#Age transformation and probe annotation functions
trafo= function(x,adult.age=20) { x=(x+1)/(1+adult.age); y=ifelse(x<=1, log( x),x-1);y }
anti.trafo= function(x,adult.age=20) { ifelse(x<0, (1+adult.age)*exp(x)-1, (1+adult.age)*x+adult.age) }
mAge=anti.trafo(horvath$coef[1]+as.matrix(datMethClock)%*% as.numeric(horvath$coef[-1]))
methyAge$mAge_Horvath=mAge

datMethClock=beta[,as.character(hannum$cg)]
mAge=as.matrix(datMethClock) %*% as.numeric(hannum$coef)
methyAge$mAge_Hannum=mAge

datMethClock=beta[,as.character(phenoage$cg[-1])]
mAge=phenoage$coef[1]+as.matrix(datMethClock)%*% phenoage$coef[-1]
methyAge$PhenoAge=mAge

methyAge$PACE=PACE[as.vector(methyAge$SampleID)]

return(methyAge)
}

relic <- function(mdat,at_red=NULL,cg_grn=NULL)
{
    if(!is(mdat, "methDataSet") & !is(mdat, "MethylSet"))
    {stop("The input must be an object of methDataSet or MethylSet\n")}
    if(is.null(at_red) | is.null(cg_grn))
    {stop("Internal control intensity matrics at_red and cg_grn must be
     provided\n")}

    # make sure the order is right
    name_grn <- gsub("G","A",gsub("C","T",rownames(cg_grn)))
    common=intersect(name_grn,rownames(at_red))
    if(length(common)<50)
       {stop("Please check Internal control intensity matrics at_red and
     cg_grn.\n Please assign internal control ExtendedType as row name 
     for matrics at_red and cg_grn\n")}
    at_red=at_red[common,];cg_grn=cg_grn[match(common,name_grn),]

    if(is(mdat, "methDataSet")){
      probe_type=rowData(mdat)$Infinium_Design_Type
      col=rowData(mdat)$Color_Channel
      probe_type[probe_type %in% c("I","snpI") & col=="Grn"]="IGrn"
      probe_type[probe_type %in% c("I","snpI") & col=="Red"]="IRed"
      probe_type[probe_type %in% c("snpII")]="II"
    }else if(is(mdat, "MethylSet")){
    probe_type <- getProbeType(mdat, withColor=TRUE)}

    m_grn=assays(mdat)$Meth[probe_type %in% c("IGrn","II"),]
    um_grn=assays(mdat)$Unmeth[probe_type %in% c("IGrn"),]
    for(i in 1:ncol(cg_grn))
    {
    index<-(at_red[,i]>10 & cg_grn[,i]>10)
    X<-cbind(rep(1,sum(index)),log(cg_grn[index,i])); Y<-log(at_red[index,i])
    temp<-solve(t(X)%*%X)%*%t(X)%*%Y;
    m_grn[,i]<-exp(log(m_grn[,i])*temp[2]+temp[1]);
    um_grn[,i]<-exp(log(um_grn[,i])*temp[2]+temp[1]);
    }
    assays(mdat)$Meth[probe_type %in% c("IGrn","II"),] <- m_grn
    assays(mdat)$Unmeth[probe_type %in% c("IGrn"),] <- um_grn
    mdat
}

QCinfo <- function(rgSet,detPthre=0.000001,detPtype="negative",nbthre=3,samplethre=0.05,CpGthre=0.05,
      bisulthre=NULL,outlier=TRUE,distplot=TRUE)
{

    ##number of bead
    if(!is(rgSet, "rgDataSet") & !is(rgSet, "RGChannelSetExtended"))
      stop("[QCinfo] The input should be an object of 'rgDataSet' or 'RGChannelSetExtended'")

    if(is(rgSet, "rgDataSet")){
      cginfo=getCGinfo(rgSet)
      typeI <- cginfo[cginfo$Infinium_Design_Type %in% c("I","snpI"),]
      typeIred=typeI[typeI$Color_Channel=="Red",]
      typeIgrn=typeI[typeI$Color_Channel=="Grn",]
      typeII<-cginfo[cginfo$Infinium_Design_Type %in% c("II","snpII"),]
      locusNames=c(typeIred$Name,typeIgrn$Name,typeII$Name)
      ##detection P value
      detP<-calcdetP(rgSet,detPtype=detPtype)
      ctrls<-metadata(rgSet)$ictrl
    }else if(is(rgSet, "RGChannelSetExtended")){
      typeI <- getProbeInfo(rgSet, type = "I")
      typeII <- getProbeInfo(rgSet, type = "II")
      locusNames <- getManifestInfo(rgSet, "locusNames")
      ##detection P value
      detP<-detectionP(rgSet)
      ctrls<-getProbeInfo(rgSet,type="Control")
    }

    bc_I <- assays(rgSet)$NBeads[typeI$AddressA,]
    flag<-bc_I>assays(rgSet)$NBeads[typeI$AddressB,]
    bc_I[flag] <- assays(rgSet)$NBeads[typeI$AddressB,][flag];
    nbead <- matrix(NA_real_, ncol = ncol(rgSet), nrow = length(locusNames),
    dimnames = list(locusNames, colnames(rgSet)))
    nbead[typeI$Name,]<-bc_I
    nbead[typeII$Name,]<-assays(rgSet)$NBeads[typeII$AddressA,]
    rm(list=c("bc_I","flag"))

    if(!identical(rownames(detP),rownames(nbead))){
        idx=intersect(rownames(detP),rownames(nbead))
        detP=detP[idx,]
        nbead=nbead[idx,]
    }
    if(!identical(colnames(detP),colnames(nbead))){
        idx=intersect(colnames(detP),colnames(nbead))
        detP=detP[,idx]
        nbead=nbead[,idx]
    }

    ## bisulfite conversion internal controls
    ctrls=ctrls[ctrls$Address %in% rownames(rgSet),]
    ctrl_r <- assays(rgSet)$Red[ctrls$Address,]
    ctrl_g <- assays(rgSet)$Green[ctrls$Address,]

cIg=data.frame(ctrl_g[ctrls$Type %in% c("BISULFITE CONVERSION I"),])
cIg[]=lapply(-cIg,rank,ties.method="average")
cIg=sort(apply(cIg,1,sum))
    cc=ctrls[(ctrls$Type %in% c("BISULFITE CONVERSION I")) & (ctrls$Color %in% c("Green","Lime","LimeGreen")),]
    if(all(names(cIg)[1:2] %in% cc$Address)){cAddress=as.vector(cc$Address)}else{cAddress=names(cIg)[1:2]}
    I_green=colMeans(ctrl_g[cAddress,])
cIr=data.frame(ctrl_r[ctrls$Type %in% c("BISULFITE CONVERSION I"),])
cIr[]=lapply(-cIr,rank,ties.method="average")
cIr=sort(apply(cIr,1,sum))
    cc=ctrls[(ctrls$Type %in% c("BISULFITE CONVERSION I")) & (ctrls$Color %in% c("Purple","Red","Tomato")),]
    if(all(names(cIr)[1:2] %in% cc$Address)){cAddress=as.vector(cc$Address)}else{cAddress=names(cIr)[1:2]}
    I_red=colMeans(ctrl_r[cAddress,])
cIIr=data.frame(ctrl_r[ctrls$Type %in% c("BISULFITE CONVERSION II"),])
cIIr[]=lapply(-cIIr,rank,ties.method="average")
cIIr=sort(apply(cIIr,1,sum))
    cc=ctrls[ctrls$Type %in% c("BISULFITE CONVERSION II") & ctrls$Color %in% c("Crimson",
     "DarkMagenta","Red","Orange","Purple","Tomato"),]
    if(all(names(cIIr)[1:2] %in% cc$Address)){cAddress=as.vector(cc$Address)}else{cAddress=names(cIIr)[1:2]}
    II_red=colMeans(ctrl_r[cAddress,])
    bisul=(I_green+I_red+II_red)/3

    #threshold of bisulfite conversion control intensity
    if(is.null(bisulthre)){bisulthre=mean(bisul,na.rm=TRUE)-3*sd(bisul,na.rm=TRUE)}

    ##low quality samples
    qcmat <- nbead<nbthre | detP>detPthre
    badValuePerSample <- apply(qcmat,2,sum)/nrow(qcmat)
    flag <- badValuePerSample > samplethre | bisul < bisulthre
    cat(sum(flag)," samples with percentage of low quality CpG value greater than ",
      samplethre, " or bisulfite intensity less than ", bisulthre, "\n")
    badsample=colnames(qcmat)[flag]

    ##low quality CpGs
    qcmat <- qcmat[,!flag]
    NbadValuePerCpG <- apply(qcmat,1,sum)
    badValuePerCpG <- NbadValuePerCpG/ncol(qcmat)
    flag2 <- badValuePerCpG>CpGthre & NbadValuePerCpG>1
    cat(sum(flag2)," CpGs with percentage of low quality value greater than ",
      CpGthre,"\n")
    badCpG <- rownames(qcmat)[flag2]
    qcmat=qcmat[!flag2,]

    #plotting quality scores
    cat("Ploting qc_sample.jpg ...")
    jpeg(filename="qc_sample.jpg",width=1000,height=1000,quality = 100)
    color=rep("black",length(badValuePerSample))
    color[flag]="red"
    plot(badValuePerSample,bisul,xlab="Percent of low quality data per sample",
       ylab="Average bisulfite conversion intensity",cex=1.5,col=color,
       main=paste(length(badsample)," samples were classified as low quality samples"))
    abline(h=bisulthre,lty=2,col="red")
    abline(v=samplethre,lty=2,col="red")
    dev.off()
    cat("Done\n")

    cat("Ploting qc_CpG.jpg ...")
    jpeg(filename="qc_CpG.jpg",width=1000,height=1000,quality = 100)
    par(mfrow=c(2,1))
    hist(badValuePerCpG,breaks=1000,xlab="Percent of low quality data per CpG",
     main=paste(length(badCpG)," CpGs were classified as low quality CpGs"))
    abline(v=CpGthre,lty=2,col="red")
    hist(badValuePerCpG,breaks=1000,xlim=c(0,0.1),
       xlab="Percent of low quality data per CpG",main="Zoom in view")
    abline(v=CpGthre,lty=2,col="red")
    dev.off()
    cat("Done\n")


    if(outlier | distplot){
    if(is(rgSet, "rgDataSet")){mdat=getmeth(rgSet)
    }else if(is(rgSet, "RGChannelSetExtended")){mdat=preprocessRaw(rgSet)}
    rm(rgSet)}

    beta0=getB(mdat, type="Illumina")
    #Identifying outlier samples
    if(outlier)
    {
    cat("Identifying ourlier samples based on beta or total intensity values...\n")
    mdat=mdat[rownames(qcmat),]
    mdat=mdat[,colnames(qcmat)]
    #outliers based on total intensity values
    mu <- assays(mdat)$Meth+assays(mdat)$Unmeth
    mumean=apply(mu,2,mean,na.rm=TRUE)
    q2575 <- quantile(mumean, probs=c(0.25,0.75), na.rm=TRUE)
    qr <- q2575["75%"]-q2575["25%"]
    low=q2575["25%"] - 3*qr
    flag1=  mumean < low
    cat("After excluding low quality samples and CpGs\n")
    cat(sum(flag1)," samples are outliers based on averaged total intensity value","\n")
rm(mu)
rm(mumean)
rm(mdat)
    #outliers in beta value distribution
#    beta=getB(mdat, type="Illumina")
    beta=beta0[rownames(qcmat),]
    beta=beta[,colnames(qcmat)]

    qq=apply(beta,2,function(x) quantile(x, probs=c(0.25,0.5,0.75), na.rm=TRUE))
    q2575 <- apply(qq,1,function(x) quantile(x, probs=c(0.25,0.75), na.rm=TRUE))
    qr <- q2575["75%",]-q2575["25%",]
    up=q2575["75%",] + 3*qr
    low=q2575["25%",] - 3*qr
    flag=qq > up | qq < low
    flag=apply(flag,2,sum)>0
    cat(sum(flag)," samples are outliers in beta value distribution","\n")
    flag=flag | flag1
    badsample=c(badsample,colnames(beta)[flag])
    outlier_sample=colnames(beta)[flag]
    cat(sum(flag)," outlier samples were added into badsample list\n")
    if(ncol(qcmat)<15){
    cat("WARNING: Sample size may be too small to correctly identify outlier samples!\n")
    cat("RECOMMAND: set outlier=FALSE or double check total intensity and beta value 
     distribution plots to confirm\n")}
    rm(beta)
    }

    if(distplot)
    {
#    if(!outlier){beta=getB(mdat, type="Illumina")}
    cat("Ploting freqpolygon_beta_beforeQC.jpg ...")
    jpeg(filename="freqpolygon_beta_beforeQC.jpg",width=1000,
     height=500,quality = 100)
    color=rep("black",ncol(beta0))
    color[colnames(beta0) %in% badsample]="red"
    multifreqpoly(beta0,cex.lab=1.4,cex.axis=1.5, col=color, legend="",
    cex.main=1.5,main="Beta value distribution",
    xlab="Methylation beta value")
    dev.off()
    cat("Done\n")

    beta0=beta0[!(rownames(beta0) %in% badCpG),]
    beta0=beta0[,!(colnames(beta0) %in% badsample)]
    cat("Ploting freqpolygon_beta_afterQC.jpg ...")
    jpeg(filename="freqpolygon_beta_afterQC.jpg",width=1000,
     height=500,quality = 100)
    multifreqpoly(beta0,cex.lab=1.4,cex.axis=1.5, col="black",legend="",
    cex.main=1.5,main="Beta value distribution",
    xlab="Methylation beta value")
    dev.off()
    cat("Done\n")
    }
#rm(mdat)

    if(outlier)
    {list(detP=detP,nbead=nbead,bisul=bisul,badsample=badsample,badCpG=badCpG,
      outlier_sample=outlier_sample)
    }else{list(detP=detP,nbead=nbead,bisul=bisul,badsample=badsample,badCpG=badCpG)}
}

#create methDataSet from rgDataSet
getmeth <-function(rgSet){
  if(!is(rgSet, "rgDataSet") & !is(rgSet, "RGChannelSet")){
    stop("[getmeth] The input must be a rgDataSet or RGChannelSet")}

if(is(rgSet, "RGChannelSet")){return(preprocessRaw(rgSet))}else{

cginfo=getCGinfo(rgSet)

probeI=cginfo[cginfo$Infinium_Design_Type %in% c("I","snpI"),]
probeIred=probeI[probeI$Color_Channel=="Red",]
probeIgrn=probeI[probeI$Color_Channel=="Grn",]
probeII=cginfo[cginfo$Infinium_Design_Type %in% c("II","snpII"),]

MIRed <- assays(rgSet)$Red[probeIred$AddressB,]
UIRed <- assays(rgSet)$Red[probeIred$AddressA,]
MIGrn <- assays(rgSet)$Green[probeIgrn$AddressB,]
UIGrn <- assays(rgSet)$Green[probeIgrn$AddressA,]

dimnames(MIRed)[[1]]=probeIred$Name
dimnames(UIRed)[[1]]=probeIred$Name
dimnames(MIGrn)[[1]]=probeIgrn$Name
dimnames(UIGrn)[[1]]=probeIgrn$Name

MII <- assays(rgSet)$Green[probeII$AddressA,]
UII <- assays(rgSet)$Red[probeII$AddressA,]

dimnames(MII)[[1]]=probeII$Name
dimnames(UII)[[1]]=probeII$Name

M=rbind(MIRed,MIGrn,MII)
U=rbind(UIRed,UIGrn,UII)

rowData=rbind(probeI,probeII)
dimnames(rowData)[[1]]=rowData$Name
rowData=rowData[dimnames(M)[[1]],]

    out <- methDataSet(
        Meth=M,
        Unmeth = U,
        rowData=as(rowData,"DataFrame")
    )
    out
}}

nm_est <-
function(x,minN=3,modedist=0.2)
{
    x <- x[!is.na(x)]
    if(sum(!is.na(x))>3){
    minN <- max(minN,ceiling(length(x)*0.05))
    y<-density(x)
    sn <- 3;nn <- sn*2+1
    n <- length(y$y)
    v <- matrix(NA,ncol=nn,nrow=n-nn+1)
    for(i in 1:nn){v[,i] <- y$y[i:(n-nn+i)]}
    ix <- sn+which(apply(v<v[,(sn+1)],1,sum)==(nn-1))
#number of samples under each peak must greater than minN
    if(length(ix)>1){
    valley <- array()
    for(i in 1:(length(ix)-1))
    {valley[i] <- y$x[ix[i]:ix[i+1]][which.min(y$y[ix[i]:ix[i+1]])]}
    v1 <- c(min(x),valley,max(x)+0.1)
    nn <- array();
    for(i in 1:(length(v1)-1)){nn[i] <- length(x[x>=v1[i] & x<v1[i+1]])}
    ix=ix[nn>=minN]
    nm <- max(1,length(ix))
    }else{nm=1}
    }else{nm=1}
#nm <- min(5,max(1,length(ix)))
    if(nm>=2){
    flag <- 1;
    xx <- sort(y$x[ix][order(y$y[ix],decreasing=TRUE)[1:nm]])
    }else{flag <- 0}
#distance between peaks must greater than modedist
    while(flag)
    {
    pdist <- xx[2:length(xx)]-xx[1:(length(xx)-1)]

    valley <- array()
    pidx <- which(y$x %in% xx)
    for(i in 1:(length(pidx)-1))
    {valley[i] <- y$x[pidx[i]:pidx[i+1]][which.min(y$y[pidx[i]:pidx[i+1]])]}
    v1 <- c(min(x),valley,max(x)+0.1)
    nn <- array();
    for(i in 1:(length(v1)-1)){nn[i] <- length(x[x>=v1[i] & x<v1[i+1]])}

    if(min(pdist)<modedist)
    {
    id_rm <- which.min(pdist)
    if(nn[id_rm]<nn[id_rm+1]){id_rm <- id_rm+1}
    xx <- xx[-id_rm]
    }else if(min(nn)<minN){
    id_rm <- which.min(nn)
    xx <- xx[-id_rm]
    }else{flag <- 0}
    nm <- length(xx)
    if(nm==1){flag <- 0}
    }
    nm
}

nmode <-function(x,minN=3,modedist=0.2,nCores=1)
{
    if(nCores>detectCores()){
    nCores <- detectCores();
    cat("Only ",detectCores()," Cores avialable, nCores was reset to ",
    detectCores(),"\n")
    }
    cat("Analysis is running, please wait...!","\n")

    CpG <- rownames(x)
    if(!(.Platform$OS.type == "unix")){
    c1 <- makeCluster(nCores)
    registerDoParallel(c1)
    }

    nm=NULL
    N=ceiling(nrow(x)/(nCores*1200))
    parts=rep(1:N,each = ceiling(nrow(x)/N))[1:nrow(x)]
    for(i in 1:N){
    id=which(parts==i)
    x.1=x[id,]

    if(.Platform$OS.type == "unix"){
    nm_est1 <-function(id,beta,minN,modedist)
    {nm_est(beta[id,],minN=minN,modedist=modedist)}
    nm.1=mclapply(1:nrow(x.1),nm_est1,beta=x.1,minN=minN,modedist=modedist,
    mc.cores=nCores)
    nm.1=unlist(nm.1)
    }else{

    nm.1 <- foreach(i = 1:nrow(x.1),.combine=c,.export=c("nm_est")) %dopar% {
     i=i; nm_est(x.1[i,], minN=minN,modedist=modedist)}
    }
    nm=c(nm,nm.1)
    }
    if(!(.Platform$OS.type == "unix")){stopCluster(c1)}
    names(nm) <- CpG
    nm
}

mpreprocess <-function(rgSet,nCores=2,bgParaEst="oob",dyeCorr="RELIC", 
        qc=TRUE,qnorm=TRUE,qmethod="quantile1",
        fqcfilter=FALSE,rmcr=FALSE,impute=FALSE)
{
    if(!is(rgSet, "RGChannelSet") & !is(rgSet, "MethylSet") & !is(rgSet, "rgDataSet")
      & !is(rgSet, "methDataSet")){
    stop("ERROR: rgSet is not an object of class 'rgDataSet','methDataSet', 'RGChannelSet' or 'MethylSet'")}
    if(nCores>detectCores()){
    nCores <- detectCores();
    cat("Only ",nCores," cores are avialable in your computer,", 
       "argument nCores was reset to nCores=",nCores,"\n")
    }
    if(qc){
      if(is(rgSet, "rgDataSet") | is(rgSet, "RGChannelSetExtended")){qc <- QCinfo(rgSet)}else{
          qc=NULL; 
          cat("rgSet is not an object of class 'rgDataSet' or 'RGChannelSetExtended';\n",
              "QC will not be performed")}
        }else{qc=NULL}
    mdat <- preprocessENmix(rgSet, bgParaEst=bgParaEst, QCinfo=qc, 
        dyeCorr=dyeCorr,nCores=nCores)
    if(qnorm){mdat=norm.quantile(mdat,method=qmethod)}
    beta=rcp(mdat,qcscore=qc)
    if(rmcr|impute){fqcfilter=TRUE}
    if(fqcfilter){beta <- qcfilter(beta,qcscore=qc,impute=impute,rmcr=rmcr)}
    beta
}

four.identical<-function(x1,x2,x3,x4){
    return(identical(x1,x2) & identical(x1,x3) & identical(x1,x4))
}

oxBS.MLE<-function(beta.BS,beta.oxBS,N.BS,N.oxBS){
    if (!four.identical(rownames(beta.BS),rownames(beta.oxBS),rownames(N.BS),
    rownames(N.oxBS))){stop("Row names are not consistent!")}
    else if (!four.identical(colnames(beta.BS),colnames(beta.oxBS),
    colnames(N.BS),colnames(N.oxBS))){stop("Column names are not consistent!")
    }else {
    # filter the measurements to do estimation
    index<-!(is.na(beta.BS) | is.na(beta.oxBS) | is.na(N.BS) | is.na(N.oxBS)
       | N.BS==0 | N.oxBS==0)
    weight.BS<-N.BS[index]/(N.BS[index]+N.oxBS[index])
    weight.oxBS<-(1-weight.BS);

    # oxBS-MLE
    beta.5mC<-matrix(NA,nrow(beta.BS),ncol(beta.BS))
    beta.5mC[index]<-ifelse(beta.BS[index]>=beta.oxBS[index],beta.oxBS[index],
    weight.BS*beta.BS[index]+weight.oxBS*beta.oxBS[index])
    beta.5hmC<-matrix(NA,nrow(beta.BS),ncol(beta.BS))
    beta.5hmC[index]<-ifelse(beta.BS[index]>=beta.oxBS[index],beta.BS[index]
    -beta.oxBS[index],0)

    # name the rows and the columns
    rownames(beta.5mC)<-rownames(beta.BS)
    rownames(beta.5hmC)<-rownames(beta.BS)
    colnames(beta.5mC)<-colnames(beta.BS)
    colnames(beta.5hmC)<-colnames(beta.BS);

    # return the estimation
    return(list("5mC"=beta.5mC,"5hmC"=beta.5hmC))
    }
}

#convert Beta value to M value
B2M<-function (x)
{
    x[x == 0] <- min(x[x != 0])
    x[x == 1] <- max(x[x != 1])
    log2(x) - log2(1 - x)
}

#convert M value to Beta value
M2B<-function(x)
{2^x/(1+2^x)}

#extract Beta value
getB <- function(mdat,type="Illumina",offset=100)
{
    if(!is(mdat, "methDataSet") & !is(mdat, "MethylSet"))
        {stop("The input is not an object of methDataSet or MethylSet\n")}
    if(type=="Illumina"){offset=100}
    beta<-assays(mdat)$Meth/(assays(mdat)$Meth+assays(mdat)$Unmeth+offset)
    beta
}

#bed example file simulation
simubed <-function(nprobe=1000)
{
   chrset=c(1:22,"X","Y")
   resu=NULL
   for(i in 1:length(chrset)){
     start=sample(1:(90*nprobe),nprobe)
     end=start+1
     p=runif(nprobe)
     resu=rbind(resu,data.frame(chr=chrset[i],start=start,end=end,p=p))
  }
  resu$probe=paste("cg",1:nrow(resu),sep="")
  return(resu)
}

#remove suffix from CpG names, and combine values for duplicated probes
rm.cgsuffix<-function(datMeth){
    cgid=sapply(strsplit(rownames(datMeth),split="_"),unlist)[1,]
    dupcg=unique(cgid[duplicated(cgid)])
    datMeth2=datMeth[cgid %in% dupcg,]
    cid=sapply(strsplit(rownames(datMeth2),split="_"),unlist)[1,]
    datMeth2=aggregate(datMeth2,by=list(cid),FUN=function(x)mean(x,na.rm=TRUE))
    rownames(datMeth2)=datMeth2[,1];datMeth2=as.matrix(datMeth2[,-1])
    datMeth=datMeth[!(cgid %in% dupcg),]
    rownames(datMeth)=sapply(strsplit(rownames(datMeth),split="_"),unlist)[1,]
    rbind(datMeth,datMeth2)
}

#Modified RCP method to normalize datMeth value to a reference
rcp2 <- function(datMeth,reference,quantile.grid=seq(0.001,0.999,by=0.001)){
    #if datMeth is a vector, convert it to a matrix
    if(is.vector(datMeth)){datMeth=as.matrix(datMeth)}

#    quantile.grid=seq(0.001,0.999,by=0.001)
    datMeth<-B2M(datMeth)

    #common CpG set between reference and datMeth
    cg=intersect(as.vector(reference$cg),rownames(datMeth))
    rownames(reference)=reference$cg
    reference=B2M(reference[cg,]$meth_mean)
    datMeth2=datMeth[cg,]
    if(is.vector(datMeth2)){datMeth2=as.matrix(datMeth2)}

#linear regression
    qtl<-function(x) quantile(x, quantile.grid, na.rm=TRUE)
    M.I=qtl(reference)
    M.II=apply(datMeth2,2,qtl)

#    datMeth.est<-mat.or.vec(2,ncol(datMeth))
    datMeth.est<-matrix(NA,2,ncol(datMeth))

    for (i in 1:ncol(datMeth)){
    index<-(M.II[,i]!=Inf & M.II[,i]!=-Inf & M.I!=Inf & M.I!=-Inf)
    X<-cbind(rep(1,sum(index)),M.II[index,i]); Y<-M.I[index]
    datMeth.est[,i]<-solve(t(X)%*%X)%*%t(X)%*%Y
    }

    datMeth.new<-matrix(NA,nrow(datMeth),ncol(datMeth))
    for (i in 1:ncol(datMeth)){
    datMeth.new[,i]<-datMeth.est[1,i]+datMeth.est[2,i]*datMeth[,i]
    }
    datMeth.new[datMeth==Inf]<-Inf; datMeth.new[datMeth==-Inf]<-(-Inf)
    dimnames(datMeth.new)=dimnames(datMeth)
    M2B(datMeth.new)
}

mhtplot <-function(probe=NULL,chr=NULL, pos=NULL, p=NULL, color="bg",sigthre=NULL,
      sigthre2=NULL,threlty=c(2,1),markprobe=NULL,markcolor="red", outf="mht",
      outfmt="jpg",reducesize=0,...)
{
if(is.null(probe)|is.null(chr)|is.null(pos)|is.null(p)){
    stop("Please provide probe,chr, pos and p\n")}
if(!(length(probe)==length(chr) & length(chr)==length(pos) & length(chr)==length(p))){
     stop("The length of probe, chr, pos and p are different")}
flag=!is.na(probe) & !is.na(chr) & !is.na(pos) & !is.na(p)

dat=data.frame(probe=probe,chr=chr,pos=pos,p=p)
dat=dat[flag,]
dat=dat[!is.na(dat$chr),]

tmp=as.vector(dat$chr); tmp=toupper(tmp)
tmp=sub("CHR","",tmp)

chridx1=mixedsort(names(table(tmp)))
tmp=factor(tmp,levels=chridx1)
tmp=as.numeric(tmp)
dat$chr=tmp

#tmp=toupper(as.vector(dat$chr))
#tmp[tmp %in% "X"]="23"
#tmp[tmp %in% "Y"]="24"
#dat$chr=as.numeric(tmp)

dat$pos=as.numeric(as.vector(dat$pos))
dat=dat[order(dat$chr,dat$pos),]
#dat=dat[!is.na(dat$chr),]

chrlen=aggregate(dat$pos,by=list(dat$chr),FUN=max)
chrlen$cumx=cumsum(chrlen$x)
chrlen$mid=(chrlen$cumx-c(0,chrlen$cumx[1:(nrow(chrlen)-1)]))/2+
            c(0,chrlen$cumx[1:(nrow(chrlen)-1)])
for(i in 2:nrow(chrlen))
{dat$pos[dat$chr %in% chrlen$Group.1[i]]=dat$pos[dat$chr %in% chrlen$Group.1[i]]+
            chrlen$cumx[i-1]}

if(color=="bg"){
dat$col="gray"
dat$col[dat$chr %in% chrlen$Group.1[(1:nrow(chrlen) %% 2)==0]]="black"
}else{dat$col=color}

if(sum(chrlen$Group.1==1:nrow(chrlen))==nrow(chrlen)){chrlen$Group.1=chridx1}
#chrlen$Group.1[chrlen$Group.1 == 23]="X"
#chrlen$Group.1[chrlen$Group.1 == 24]="Y"

dsig=dat[dat$probe %in% markprobe,]

if(outfmt == "jpg"){
jpeg(paste(outf,".",outfmt,sep=""),height=600,width=900,quality=100)
par(mar=c(5.1, 5.1, 4.1, 2.1))
}else if(outfmt == "eps"){
#postscript(paste(outf,".",outfmt,sep=""),height=2,width=3)
grDevices::postscript(paste(outf,".",outfmt,sep=""),width = 4, height = 3.1,
      paper = "special",pointsize=5,onefile=FALSE)
par(mar=c(5.1, 5.1, 4.1, 2.1))
while(reducesize>0){
dat=dat[order(dat$pos),]
dat$id=1:nrow(dat)
dat$vp=-log10(dat$p)
dat$v1=dat$pos/max(dat$pos)
dat$v2=dat$vp/max(dat$vp)
dat$v3=c(2,abs(dat$v1[2:nrow(dat)]-dat$v1[1:(nrow(dat)-1)])+
  abs(dat$v2[2:nrow(dat)]-dat$v2[1:(nrow(dat)-1)]))
dat$V3[nrow(dat)]=2
dat=dat[order(dat$v3),]
exid=dat$id[1:round(nrow(dat)/2)]
exid=exid[exid%%2==0]
dat=dat[!(dat$id %in% exid),]
reducesize=reducesize-1
}
}else{stop("Outfmt should be jpg or eps\n")}

plot(dat$pos,-log10(dat$p),col=dat$col,pch=20,lwd=1,xaxt="n",
    xlab="Chromosome location",
#    ylab=expression(paste(-log[10],"(", italic(P), " value)")),
    ylab=bquote('-log'['10']*'(P) value'),
    cex.lab=1.5,...)

axis(1,at=c(0,chrlen$cumx),las=1,lwd=1,labels=FALSE,cex=0.5,font=2)
if(outfmt=="eps"){
graphics::mtext(chrlen$Group.1,side=1,at=chrlen$mid,las=2,adj=0.5,line=1,cex=0.7)}else{
graphics::mtext(chrlen$Group.1,side=1,at=chrlen$mid,las=2,adj=0.5,line=1)}

if(!is.null(markprobe)){points(dsig$pos,-log10(dsig$p),col=markcolor,pch=20)}

if(!is.null(sigthre)){abline(h=-log10(sigthre),lty=threlty[1])}
if(!is.null(sigthre2)){abline(h=-log10(sigthre2),lty=threlty[2])}
dev.off()
}

repicc<-function(dat,repid,mvalue=FALSE,nCores=2,qcflag=FALSE,qc=NULL,
       detPthre=0.05,nbthre=3)
{
  if(nCores>detectCores()){nCores=detectCores()}
  if(mvalue){dat=B2M(dat)}
  if(qcflag){
    if(is.null(qc)){stop("Please provide ENmix QCinfo object for quality control")} 
    dat=dat[,!(colnames(dat) %in% qc$badsample)]
    dat=dat[!(rownames(dat) %in% qc$badCpG),]
    qc$detP=qc$detP[rownames(dat),]
    qc$detP=qc$detP[,colnames(dat)]
    qc$nbead=qc$nbead[,colnames(dat)]
    qc$nbead=qc$nbead[rownames(dat),]
    dat[qc$detP>detPthre | qc$nbead<nbthre]<-NA
}      

if(!identical(repid$id,colnames(dat))){
  rownames(repid)=repid$id
  id=intersect(repid$id,colnames(dat))
  repid=repid[id,]
  dat=dat[,id]
#remove singletons
x.char=as.character(repid$idx)
flag=x.char %in% names(table(x.char)[table(x.char)<2])
repid=repid[!flag,]
dat=dat[,!flag]
}

icc<-function(probe,meth,x.char){
  y=meth[probe,]
#remove missing data and singletons
  flag=is.na(y);if(sum(flag)>0){y=y[!flag];x.char=x.char[!flag]
flag=x.char %in% names(table(x.char)[table(x.char)<2]);y=y[!flag];x.char=x.char[!flag]
}
  temp<-aggregate(y~x.char,FUN=mean)
  ga<-temp[,2]
  gs<-table(x.char)
  sst<-sum((ga-mean(y))^2*gs[temp$"x.char"])
  sse<-(sum((y-mean(y))^2)-sst)
  n<-length(y); t<-length(ga)
  mse<-sse/(n-t)
  mst<-sst/(t-1)
  c<-(n^2-sum(gs^2))/(n*(t-1))
  sigma.a.sq<-(mst-mse)/c
  c(probe,ifelse(sigma.a.sq<=0,0,sigma.a.sq/(sigma.a.sq+mse)))
}

resu1=mclapply(dimnames(dat)[[1]],icc,meth=dat,x.char=as.character(repid$idx), mc.cores=nCores)
resu <- do.call(rbind, lapply(resu1, unlist))
resu=as.data.frame(resu)
names(resu)=c("probe","icc")
resu$icc=as.numeric(as.vector(resu$icc))
resu
}

# rgSet class

setClass("rgDataSet", contains = "SummarizedExperiment")

# Validity methods

setValidity2("rgDataSet", function(rgSet) {
    if(!all((metadata(rgSet)$ictrl)$Address %in% rownames(rgSet))){
    stop("Some control probes do not have intensity data, please provide correct manifest file")}
    probeIA=rowData(rgSet)$IlmnID[rowData(rgSet)$Infinium_Design_Type %in% c("IA","snpIA")]
    probeIB=rowData(rgSet)$IlmnID[rowData(rgSet)$Infinium_Design_Type %in% c("IB","snpIB")]
    if(!all(probeIA %in% probeIB) | !all(probeIA %in% probeIB)){
    stop("Some type I probes missed reads from channels A or B")}
    NULL
})

# rgSet functions
rgDataSet <- function(Red = new("matrix"), Green = new("matrix"),
    NBeads = new("matrix"),rowData=new("DataFrame"),ictrl= new("DataFrame"),...){
    assays <- SimpleList(Red = Red, Green = Green, NBeads = NBeads)
    metadata <- list(ictrl=ictrl)
    new("rgDataSet",
        SummarizedExperiment(assays = assays,metadata=metadata,rowData=rowData,...)
    )
}

# Exported methods -------------------------------------------------------------

setMethod("show", signature(object = "rgDataSet"), function(object) {
    callNextMethod()
    cat("Array: ", metadata(object)$Array, "\n")
    cat("Annotation: ", metadata(object)$annotation,"\n")
})

predSex <-function(mdat,cutoff=2){
  if(!is(mdat, "rgDataSet") & !is(mdat, "methDataSet")){
    stop("[predSex] The input must be a rgDataSet or methDataSet")}

  if(is(mdat, "rgDataSet")){mdat=getmeth(mdat)}
  xmdat=mdat[rowData(mdat)$chr %in% c("chrX","X")]
  ymdat=mdat[rowData(mdat)$chr %in% c("chrY","Y"),]
  xCN <- log2(assays(xmdat)$Meth + assays(xmdat)$Unmeth)
  yCN <- log2(assays(ymdat)$Meth + assays(ymdat)$Unmeth)
  xMed=apply(xCN,2,median,na.rm=TRUE)    
  yMed=apply(yCN,2,median,na.rm=TRUE)
  diff=xMed -yMed
  sex <- ifelse(diff > cutoff, "F", "M")
  return(data.frame(id=colnames(mdat),sex=sex))

#add in prediction based on beta values
#kmeans clustering, in minfi
}

methscore<-function(datMeth,datPheno=NULL,fastImputation=FALSE,normalize=TRUE,
		    GrimAgeComponent=3,AgeResid=FALSE)
{
#input datMeth value matrix must be in between 0 to 1, percentage of cg methylated
datMeth<-as.matrix(datMeth)
if(min(datMeth,na.rm=TRUE)<0 | max(datMeth,na.rm=TRUE)>1){
    stop("Warning: Methylation datMeth value should be within [0,1]")}

#if GrimAge Components flag
if (!GrimAgeComponent %in% c(1, 2, 3)) {
  stop("Error: GrimAgeComponent must be 1, 2, or 3")
}

#remove CpGs with missing values for all samples
datMeth=datMeth[rowSums(is.na(datMeth)) != ncol(datMeth),]

#check pheno 
if (is.null(datPheno)) {stop("Error: datPheno is missing.")}
datPheno <- as.data.frame(datPheno)
required_cols <- c("Age", "Female", "SampleID")
missing_cols <- setdiff(required_cols, colnames(datPheno))
if (length(missing_cols) > 0) {
  stop(paste0("Error: datPheno must have the following columns: ",
    paste(missing_cols, collapse = ", ")))
}
if (!all(colnames(datMeth) %in% datPheno$SampleID)) {
  stop("Error: some samples in datMeth do not have phenotype in datPheno")
}
rownames(datPheno) <- datPheno$SampleID
datPheno <- datPheno[colnames(datMeth), , drop = FALSE]
if (any(is.na(datPheno))) {
  stop("Error: missing value in phenotype data is not allowed")
}

#load in all model and reference data
load(system.file("mage_ref.RData",package="ENmix"))
load(system.file("mPOA_Models.RData",package="ENmix"))
#load(system.file("pcclock_model.RData",package="ENmix"))
#pcc1=pcc2=pcc3=pcc4=NULL
#load in PC related clocks
load(system.file("pcclock_model1.RData",package="ENmix"))
load(system.file("pcclock_model2.RData",package="ENmix"))
load(system.file("pcclock_model3.RData",package="ENmix"))
load(system.file("pcclock_model4.RData",package="ENmix"))
pcc=c(pcc1,pcc2,pcc3,pcc4)

#check whether datMeth probe names include suffix, such as in EPICv2 array
if(sum(as.character(refmeth$cg) %in% rownames(datMeth))<50){datMeth=rm.cgsuffix(datMeth)}

#combine reference CpG
#normalize to Horvath refmeth before combine
refcg <- unique(c(
  as.character(refmeth$cg),
  as.character(mPOA_Models$gold_standard_probes),
  as.character(episcore_model$CpG_Site),
  as.character(frailty_model$cg[-1]),
  names(pcc$pcCpGs),
  as.character(refmeth2$cg),
  as.character(cAge_CpG$means$cpg),
  as.character(bAge_CpG$cpgs$CpG_Site),
  as.character(DNAmFitnessModels$AllCpGs)
))

#remove unnecessary probes for estimation
datMeth=datMeth[rownames(datMeth) %in% refcg,,drop=FALSE]

#imputation to fill missing values
noMissingPerCpG <- rowSums(is.na(datMeth))
if(max(noMissingPerCpG)>0 & ncol(as.matrix(datMeth))>1){
message("Imputation to fill missing values...")
if(!fastImputation){
    set.seed(1)
    datMeth <- t(impute::impute.knn(t(datMeth))$data)
}else{
    for (i in which(noMissingPerCpG>0)){
        idx <- is.na(datMeth[i,])
        datMeth[i,idx] <- mean(datMeth[i,],na.rm=TRUE)
}}}

#Count of model required CpGs that are missing in user provided data
## Count of model-required CpGs that are missing in user-provided data
cgcheck <- data.frame(
  predictor = character(),
  nCpG_required = integer(),
  nCpG_present = integer(),
  nCpG_missing = integer(),
  stringsAsFactors = FALSE
)
add_check <- function(name, required, present) {
  cgcheck <<- rbind(
    cgcheck,
    data.frame(
      predictor = name,
      nCpG_required = required,
      nCpG_present = present,
      stringsAsFactors = FALSE
    )
  )
}
rownames_meth <- rownames(datMeth)
add_check("HorvathAge", nrow(horvath) - 1,
          sum(as.character(horvath$cg) %in% rownames_meth))
add_check("PhenoAge", nrow(phenoage) - 1,
          sum(as.character(phenoage$cg) %in% rownames_meth))
add_check("HannumAge",
          nrow(hannum),
          sum(as.character(hannum$cg) %in% rownames_meth))
add_check("PACE",length(mPOA_Models$model_probes),
          sum(as.character(mPOA_Models$model_probes) %in% rownames_meth))
add_check("Frailty",nrow(frailty_model) - 1,
          sum(as.character(frailty_model$cg[-1]) %in% rownames_meth))
add_check("DNAmTL",nrow(DNAmTL_CpGs),
          sum(as.character(DNAmTL_CpGs$ID) %in% rownames_meth))
add_check("pcgtAge",length(EpiToc_CpGs),
          sum(EpiToc_CpGs %in% rownames_meth))
add_check("TNSC",nrow(EpiToc2_CpGs),
          sum(rownames(EpiToc2_CpGs) %in% rownames_meth))
add_check("TNSC2",nrow(EpiToc2_CpGs),
          sum(rownames(EpiToc2_CpGs) %in% rownames_meth))
add_check("Zhang10CpG",nrow(Zhang_10_CpG),
          sum(as.character(Zhang_10_CpG$Marker) %in% rownames_meth))
add_check("Horvath2",nrow(Horvath2_CpGs),
          sum(as.character(Horvath2_CpGs$ID) %in% rownames_meth))
add_check("MiAge", nrow(MiAgedat$MiAge_CpGs),
          sum(as.character(MiAgedat$MiAge_CpGs$CpGs) %in% rownames_meth))
add_check("PedBE",nrow(PEDBE_CpGs), 
	  sum(as.character(PEDBE_CpGs$ID) %in% rownames_meth))
add_check("GACPC",nrow(GACPC_CpGs),
          sum(as.character(GACPC_CpGs$CpG) %in% rownames_meth))
add_check("GARPC",nrow(GARPC_CpGs),
          sum(as.character(GARPC_CpGs$CpG) %in% rownames_meth))
add_check("GARRPC",nrow(GARRPC_CpGs),
          sum(as.character(GARRPC_CpGs$CpG) %in% rownames_meth))
add_check("BohlinGAge", nrow(Bohlin_CpGs),
          sum(as.character(Bohlin_CpGs$CpG) %in% rownames_meth))
add_check("KnightGAge", nrow(Knight_CpGs),
          sum(as.character(Knight_CpGs$CpG) %in% rownames_meth))
#pcclock
pcvar <- c(
  "PCHorvath1", "PCHorvath2", "PCHannum", "PCPhenoAge",
  "PCDNAmTL", "PCPACKYRS", "PCADM", "PCB2M", "PCCystatinC",
  "PCGDF15", "PCLeptin", "PCPAI1", "PCTIMP1", "PCGrimAge"
)
pc_cpgs <- names(pcc$pcCpGs)
for (vv in pcvar) {
  add_check(vv,length(pc_cpgs), sum(pc_cpgs %in% rownames_meth))
}

## cAge
add_check("cAge",nrow(cAge_CpG$means),
  sum(as.character(cAge_CpG$means$cpg) %in% rownames_meth))
## bAge
add_check("bAge",nrow(bAge_CpG$cpgs),
  sum(as.character(bAge_CpG$cpgs$CpG_Site) %in% rownames_meth))
add_check("bAge_Years", nrow(bAge_CpG$cpgs),
  sum(as.character(bAge_CpG$cpgs$CpG_Site) %in% rownames_meth))
## episcore
for (ms in unique(episcore_model$Predictor)) {
  tmp <- episcore_model[episcore_model$Predictor == ms, , drop = FALSE]
  add_check(ms, nrow(tmp), sum(as.character(tmp$CpG_Site) %in% rownames_meth))
}
## DNAm Fit Age
fa <- c(
  "DNAmGait_noAge", "DNAmGrip_noAge", "DNAmVO2max",
  "DNAmGait_wAge", "DNAmGrip_wAge", "DNAmFEV1_wAge",
  "DNAmFitAge"
)
for (i in seq_along(fa)) {
  add_check(fa[i], length(DNAmFitnessModels$AllCpGs),
    sum(as.character(DNAmFitnessModels$AllCpGs) %in% rownames_meth)
  )
}

#GrimAge
add_check("DNAmPACKYRS", length(GrimAgeV1_model$PACKYRS.model[-1]),
          sum(names(GrimAgeV1_model$PACKYRS.model) %in% rownames_meth))
add_check("DNAmADM", length(GrimAgeV1_model$ADM.model[-1]),
          sum(names(GrimAgeV1_model$ADM.model) %in% rownames_meth))
add_check("DNAmB2M", length(GrimAgeV1_model$B2M.model[-1]),
          sum(names(GrimAgeV1_model$B2M.model) %in% rownames_meth))
add_check("DNAmCystatinC", length(GrimAgeV1_model$CystatinC.model[-1]),
          sum(names(GrimAgeV1_model$CystatinC.model) %in% rownames_meth))
add_check("DNAmGDF15", length(GrimAgeV1_model$GDF15.model[-1]),
          sum(names(GrimAgeV1_model$GDF15.model) %in% rownames_meth))
add_check("DNAmLeptin", length(GrimAgeV1_model$Leptin.model),
          sum(names(GrimAgeV1_model$Leptin.model) %in% rownames_meth))
add_check("DNAmPAI1", length(GrimAgeV1_model$PAI1.model),
          sum(names(GrimAgeV1_model$PAI1.model) %in% rownames_meth))
add_check("DNAmTIMP1", length(GrimAgeV1_model$TIMP1.model[-1]),
          sum(names(GrimAgeV1_model$TIMP1.model) %in% rownames_meth))
add_check("DNAmlogCRP", length(GrimAgeV2_model$logCRP.model),
          sum(names(GrimAgeV2_model$logCRP.model) %in% rownames_meth))
add_check("DNAmlogA1C", length(GrimAgeV2_model$logA1C.model[-1]),
          sum(names(GrimAgeV2_model$logA1C.model) %in% rownames_meth))
add_check("GrimAgeV1", length(GrimAge_CpG),
	  sum(GrimAge_CpG %in% rownames_meth))
add_check("GrimAgeV2", length(GrimAge_CpG),
          sum(GrimAge_CpG %in% rownames_meth))

#AdaptAge_CpGs
add_check("AdaptAge", length(AdaptAge_model$CpG),
          sum(AdaptAge_model$CpG %in% rownames_meth))

#AdaptAge_CpGs
add_check("CausAge", length(CausAge_model$CpG),
          sum(CausAge_model$CpG %in% rownames_meth))

#DamAge_CpGs
add_check("DamAge", length(DamAge_model$CpG),
          sum(DamAge_model$CpG %in% rownames_meth))


message("Number of CpGs missing for estimates\n")
cgcheck$nCpG_missing=cgcheck$nCpG_required-cgcheck$nCpG_present
print(cgcheck)
if(max(cgcheck$nCpG_missing)>0){
	message("Missing probes will be imputed with reference values\n")}
message("See file summary_methscore_CpG.csv for the CpG count summary 
	and citation for each predictor")
cgcheck=merge(methscore_dict,cgcheck,by="predictor")
rownames(cgcheck)=cgcheck$predictor
cgcheck=cgcheck[as.character(methscore_dict$predictor),]
write.csv(cgcheck,file="summary_methscore_CpG.csv",row.names=FALSE)

#function for normalization using a modified RCP method, and imputing missing CpGs
norm_impute <-function(dat,refdat,normalize = TRUE, missedcg = character(0)){
   if(normalize){dat=rcp2(dat,refdat)}
    #replace missing probes with reference values
    missedcg <- missedcg[!(missedcg %in% rownames(dat))]
    if(length(missedcg)>0){
        refsub=refdat[as.character(refdat$cg) %in% missedcg, , drop = FALSE]
        missedcg.dat=matrix(rep(refsub$meth_mean,ncol(dat)),
			    nrow=length(missedcg),ncol = ncol(dat))
        rownames(missedcg.dat)=as.character(refsub$cg)
        colnames(missedcg.dat)=colnames(dat)
        dat=rbind(dat,missedcg.dat)
     }
     return(dat)    
}

#Age transformation and probe annotation functions for Hovath age
trafo= function(x,adult.age=20) { x=(x+1)/(1+adult.age); 
    y=ifelse(x<=1, log( x),x-1);y }
anti.trafo= function(x,adult.age=20) {
       	ifelse(x<0, (1+adult.age)*exp(x)-1, (1+adult.age)*x+adult.age) }


###Methylation predictors
#methylation ages
message("Calculating methylation scores ...")
mScore=data.frame(SampleID=colnames(datMeth))

modelcg=unique(c(as.character(horvath$cg[-1]),as.character(phenoage$cg[-1]),
		 as.character(hannum$cg)))
missedcg=modelcg[!(modelcg %in% rownames(datMeth))]
refdat=refmeth
datMeth2=norm_impute(datMeth,refdat=refdat,normalize,missedcg)

#HorvathAge
intercept=horvath$coef[horvath$cg=="(Intercept)"]
horvath=horvath[as.character(horvath$cg) %in% rownames(datMeth2), , drop = FALSE]
mAge=anti.trafo(
    colSums(as.numeric(horvath$coef) * datMeth2[as.character(horvath$cg), , drop = FALSE],
		na.rm=T)+intercept)
mScore$HorvathAge=mAge[as.character(mScore$SampleID)]
#Hannum Age
hannum=hannum[hannum$cg %in% rownames(datMeth2), , drop = FALSE]
mAge=colSums(hannum$coef * datMeth2[as.character(hannum$cg), , drop = FALSE],na.rm=TRUE)
mScore$HannumAge=mAge[as.character(mScore$SampleID)]
#PhenoAge
intercept=phenoage$coef[phenoage$cg=="intercept"]
phenoage=phenoage[phenoage$cg %in% rownames(datMeth2), , drop = FALSE]
mAge=colSums(phenoage$coef * datMeth2[as.character(phenoage$cg), , drop = FALSE],na.rm=T)+intercept
mScore$PhenoAge=mAge[as.character(mScore$SampleID)]

#PACE estimation
modelcg=unique(as.character(mPOA_Models$model_probes))
missedcg=modelcg[!(modelcg %in% rownames(datMeth))]
refdat=data.frame(cg=mPOA_Models$gold_standard_probes,meth_mean=mPOA_Models$gold_standard_means)
datMeth2=norm_impute(datMeth,refdat=refdat,normalize,missedcg)

PACE=DunedinPACE(betas=datMeth2, proportionOfProbesRequired = 0.8)
mScore$PACE=PACE[as.character(mScore$SampleID)]

#Frailty
modelcg=unique(as.character(frailty_model$cg[-1]))
missedcg=modelcg[!(modelcg %in% rownames(datMeth))]
refdat=frailty_model[-1,c("cg","meth_mean")]
datMeth2=norm_impute(datMeth,refdat=refdat,normalize,missedcg)

intercept=frailty_model$coef[frailty_model$cg=="intercept"]
frailty_model=frailty_model[frailty_model$cg %in% rownames(datMeth2), , drop = FALSE]
eFRS=colSums(frailty_model$coef * datMeth2[as.character(frailty_model$cg), , drop = FALSE],na.rm=T)+intercept
mScore$Frailty=eFRS[as.character(mScore$SampleID)]


#PEDBE,EpiToc,EpiToc2,Zhang10CpG,Horvath2,MiAge,DNAmTL,PEDBE,GACPC,GARPC,GARRPC,Bohlin,Knight
modelcg <- unique(c(
  as.character(PEDBE_CpGs$ID),
  as.character(EpiToc_CpGs),
  rownames(EpiToc2_CpGs),
  as.character(Zhang_10_CpG$Marker),
  as.character(Horvath2_CpGs$ID),
  as.character(MiAgedat$MiAge_CpGs$CpGs),
  as.character(DNAmTL_CpGs$ID),
  as.character(GACPC_CpGs$CpG),
  as.character(GARPC_CpGs$CpG),
  as.character(GARRPC_CpGs$CpG),
  as.character(Bohlin_CpGs$CpG),
  as.character(Knight_CpGs$CpG)
))
missedcg <- setdiff(modelcg, rownames(datMeth))
refdat=refmeth2; names(refdat)=c("cg","meth_mean")
datMeth2=norm_impute(datMeth,refdat=refdat,normalize,missedcg)

#DNAmTL
DNAmTL_CpGs=DNAmTL_CpGs[DNAmTL_CpGs$ID %in% rownames(datMeth2), , drop = FALSE]
DNAmTL=colSums(DNAmTL_CpGs$Coef * datMeth2[as.character(DNAmTL_CpGs$ID), , drop = FALSE],na.rm=TRUE)+7.924780053
mScore$DNAmTL=DNAmTL[as.character(mScore$SampleID)]

#EpiTOC
pcgtAge <- colMeans(datMeth2[rownames(datMeth2) %in% EpiToc_CpGs, , drop = FALSE],na.rm=TRUE)
mScore$pcgtAge=pcgtAge[as.character(mScore$SampleID)]

#EpiTOC2
EpiToc2_CpGs=EpiToc2_CpGs[rownames(EpiToc2_CpGs) %in% rownames(datMeth2), , drop = FALSE]
tmp.m=datMeth2[rownames(EpiToc2_CpGs), , drop = FALSE]
TNSC <- 2*colMeans((1/(EpiToc2_CpGs[,1]*(1-EpiToc2_CpGs[,2]))) * (tmp.m - EpiToc2_CpGs[,2]),na.rm=TRUE)
#approximated = T
TNSC2 <- 2*colMeans((1/EpiToc2_CpGs[,1]) * tmp.m,na.rm=TRUE)
mScore$TNSC=TNSC[as.character(mScore$SampleID)]
mScore$TNSC2=TNSC2[as.character(mScore$SampleID)]

#Zhang 10 CpG clock
Zhang_10_CpG=Zhang_10_CpG[as.character(Zhang_10_CpG$Marker) %in% rownames(datMeth2), , drop = FALSE]
Zhang10CpG=colSums(Zhang_10_CpG$coef * datMeth2[as.character(Zhang_10_CpG$Marker), , drop = FALSE],na.rm=T)
mScore$Zhang10CpG=Zhang10CpG[as.character(mScore$SampleID)]

#Horvath2
Horvath2_CpGs=Horvath2_CpGs[as.character(Horvath2_CpGs$ID) %in% rownames(datMeth2), , drop = FALSE]
Horvath2=anti.trafo(colSums(Horvath2_CpGs$Coef * datMeth2[as.character(Horvath2_CpGs$ID), , drop = FALSE],na.rm=T)-0.447119319)
mScore$Horvath2=Horvath2[as.character(mScore$SampleID)]

#MiAge mitotic age
#function used by the original MiAge code http://www.columbia.edu/~sw2206/softwares.htm
MiAge_fr2 <- function (x, b, c, d, betaj)
{
    nj = x
    return(sum((c + b^(nj - 1) * d - betaj)^2, na.rm = T))
}
MiAge_grr2 <-function (x, b, c, d, betaj)
{
    nj = x
    return(2 * sum((c + b^(nj - 1) * d - betaj) * b^(nj - 1) *
        log(b) * d, na.rm = T))
}
mitotic.age <- function (beta, b, c, d)
{
#    library(methods)
    upperage = 10000
    lowerage = 10
    n = rep(500, ncol(beta));names(n)=colnames(beta)
    no.initial.n = 5
    for (j in 1:ncol(beta)) {
        current.value = MiAge_fr2(n[j], b, c, d, beta[, j])
        columnoptim = vector("list", no.initial.n)
        val = rep(NA, no.initial.n)
        for (jj in 1:(no.initial.n - 1)) {
            temp = try(optim(par = lowerage + jj * (upperage -
                lowerage)/no.initial.n, fn = MiAge_fr2, gr = MiAge_grr2,
                b = b, c = c, d = d, betaj = beta[, j], method = "L-BFGS-B",
                lower = lowerage, upper = upperage, control = list(factr = 1)),
                silent = T)
            if (!is(temp, "try-error")) {
                columnoptim[[jj]] = temp
                val[jj] = temp$value
            }
        }
        temp = try(optim(par = n[j], fn = MiAge_fr2, gr = MiAge_grr2,
            b = b, c = c, d = d, betaj = beta[, j], method = "L-BFGS-B",
            lower = lowerage, upper = upperage, control = list(factr = 1)),
            silent = T)
        if (!is(temp, "try-error")) {
            columnoptim[[no.initial.n]] = temp
            val[no.initial.n] = temp$value
        }
        temp = columnoptim[[which(val == min(val, na.rm = T))[1]]]
        if (!is(temp, "try-error")) {
            n[j] = temp$par
        }
        else {
            print(2)
            print(temp)
        }
    }
    return(n)
}

MiAge=mitotic.age(
		  datMeth2[na.omit(match(MiAgedat$MiAge_CpGs$CpGs,
		  rownames(datMeth2))),,drop=FALSE],MiAgedat$MiAge_parameters[[1]],
		  MiAgedat$MiAge_parameters[[2]],
		  MiAgedat$MiAge_parameters[[3]])
mScore$MiAge=MiAge[as.character(mScore$SampleID)]

#PedBE The Pediatric-Buccal-Epigenetic (PedBE) clock
PEDBE_CpGs=PEDBE_CpGs[PEDBE_CpGs$ID %in% rownames(datMeth2),]
PedBE=anti.trafo(colSums(PEDBE_CpGs$Coef * datMeth2[as.character(PEDBE_CpGs$ID),,drop=FALSE],na.rm=T)-2.10)
mScore$PedBE=PedBE[as.character(mScore$SampleID)]

#Placental epigenetic clocks
#Control placental clock (CPC)
GACPC_CpGs=GACPC_CpGs[GACPC_CpGs$CpG %in% rownames(datMeth2),]
GACPC=colSums(GACPC_CpGs$coef * datMeth2[as.character(GACPC_CpGs$CpG),,drop=FALSE],na.rm = T)+13.06182
mScore$GACPC=GACPC[as.character(mScore$SampleID)]

#Robust placental clock (RPC)
GARPC_CpGs=GARPC_CpGs[GARPC_CpGs$CpG %in% rownames(datMeth2),]
GARPC=colSums(GARPC_CpGs$coef * datMeth2[as.character(GARPC_CpGs$CpG),,drop=FALSE], na.rm = T)+24.99772
mScore$GARPC=GARPC[as.character(mScore$SampleID)]

#Refined robust placental clock for uncomplicated term pregnancies
GARRPC_CpGs=GARRPC_CpGs[GARRPC_CpGs$CpG %in% rownames(datMeth2),]
GARRPC=colSums(GARRPC_CpGs$coef * datMeth2[as.character(GARRPC_CpGs$CpG),,drop=FALSE], na.rm = T)+ 30.74966
mScore$GARRPC=GARRPC[as.character(mScore$SampleID)]

#Bohlin Gestational age
Bohlin_CpGs=Bohlin_CpGs[Bohlin_CpGs$CpG %in% rownames(datMeth2),]
BohlinGAge=colSums(Bohlin_CpGs$coef * datMeth2[as.character(Bohlin_CpGs$CpG),,drop=FALSE],na.rm=T)+ 277.2421
mScore$BohlinGAge=BohlinGAge[as.character(mScore$SampleID)]

#Knight Gestational age
Knight_CpGs=Knight_CpGs[Knight_CpGs$CpG %in% rownames(datMeth2),]
KnightGAge=colSums(Knight_CpGs$coef * datMeth2[as.character(Knight_CpGs$CpG),,drop=FALSE],na.rm=T) + 41.7
mScore$KnightGAge=KnightGAge[as.character(mScore$SampleID)]

#PC clocks
modelcg=names(pcc$pcCpGs)
missedcg=modelcg[!(modelcg %in% rownames(datMeth))]
refdat=data.frame(cg=names(pcc$pcCpGs),meth_mean=pcc$pcCpGs)
datMeth2=norm_impute(datMeth,refdat=refdat,normalize,missedcg)

if(!is.null(datPheno)){
    pcclock=calcPCClocks(methdat=datMeth2,phenodat=datPheno,pcc)
    pcclock=pcclock[,!(names(pcclock) %in% c("Age","Female"))]
    mScore <- cbind(mScore, pcclock[match(mScore$SampleID, pcclock$SampleID), setdiff(colnames(pcclock), "SampleID"),drop=FALSE])
}

# Calculate GrimAge Clocks
# adopted code from methylCIPHER
modelcg <- GrimAge_CpG
missedcg <- setdiff(modelcg, rownames(datMeth))
datMeth2=norm_impute(datMeth,refdat=refmeth2,normalize,missedcg)
if(!is.null(datPheno)){
DNAm <- datMeth2[GrimAge_CpG, , drop = FALSE]
pp <- datPheno
rownames(pp)=pp$SampleID
DNAm <- DNAm [,rownames(pp),drop=FALSE]
DNAm <- rbind(DNAm, Age= pp[colnames(DNAm), "Age"])
pp$DNAmPACKYRS <- drop(GrimAgeV1_model$PACKYRS.model %*% DNAm[names(GrimAgeV1_model$PACKYRS.model),,drop = FALSE]) + GrimAgeV1_model$PACKYRS.intercept
pp$DNAmADM <- drop(GrimAgeV1_model$ADM.model %*% DNAm[names(GrimAgeV1_model$ADM.model), , drop = FALSE]) +  GrimAgeV1_model$ADM.intercept
pp$DNAmB2M <- drop(GrimAgeV1_model$B2M.model %*% DNAm[names(GrimAgeV1_model$B2M.model), , drop = FALSE]) +  GrimAgeV1_model$B2M.intercept
pp$DNAmCystatinC <- drop(GrimAgeV1_model$CystatinC.model %*% DNAm[names(GrimAgeV1_model$CystatinC.model), , drop = FALSE]) +  GrimAgeV1_model$CystatinC.intercept
pp$DNAmGDF15 <- drop(GrimAgeV1_model$GDF15.model %*% DNAm[names(GrimAgeV1_model$GDF15.model), , drop = FALSE]) +  GrimAgeV1_model$GDF15.intercept
pp$DNAmLeptin <- drop(GrimAgeV1_model$Leptin.model %*% DNAm[names(GrimAgeV1_model$Leptin.model), , drop = FALSE]) +  GrimAgeV1_model$Leptin.intercept
pp$DNAmPAI1 <- drop(GrimAgeV1_model$PAI1.model %*% DNAm[names(GrimAgeV1_model$PAI1.model), , drop = FALSE]) +  GrimAgeV1_model$PAI1.intercept
pp$DNAmTIMP1 <- drop(GrimAgeV1_model$TIMP1.model %*% DNAm[names(GrimAgeV1_model$TIMP1.model), , drop = FALSE]) +  GrimAgeV1_model$TIMP1.intercept
pp$GrimAgeV1 <- drop(as.matrix(pp[, GrimAgeV1_model$components, drop = FALSE]) %*%    GrimAgeV1_model$GrimAge.model)
y <- pp$GrimAgeV1
pp$GrimAgeV1 <- (((y - GrimAgeV1_model$GrimAge.transform[3]) /  GrimAgeV1_model$GrimAge.transform[4]) *  GrimAgeV1_model$GrimAge.transform[2]) +  GrimAgeV1_model$GrimAge.transform[1]
pp$DNAmlogA1C <- drop(GrimAgeV2_model$logA1C.model %*% DNAm[names(GrimAgeV2_model$logA1C.model), , drop = FALSE]) +  GrimAgeV2_model$logA1C.intercept
pp$DNAmlogCRP <- drop(GrimAgeV2_model$logCRP.model %*% DNAm[names(GrimAgeV2_model$logCRP.model), , drop = FALSE]) +  GrimAgeV2_model$logCRP.intercept
pp$GrimAgeV2 <- drop(as.matrix(pp[, GrimAgeV2_model$components, drop = FALSE]) %*%    GrimAgeV2_model$GrimAge.model)
y <- pp$GrimAgeV2
pp$GrimAgeV2 <- (((y - GrimAgeV2_model$GrimAge.transform[3]) /  GrimAgeV2_model$GrimAge.transform[4]) *  GrimAgeV2_model$GrimAge.transform[2]) +  GrimAgeV2_model$GrimAge.transform[1]
pp <- pp[match(mScore$SampleID, rownames(pp)), , drop = FALSE]
mScore <- cbind( mScore, pp[, !(colnames(pp) %in% c("Age", "Female", "SampleID")), drop = FALSE])
}


#reference value for AdaptAge CausAge DamAge are in refmeth2
modelcg <- c(AdaptAge_model$CpG,CausAge_model$CpG,DamAge_model$CpG)
missedcg <- setdiff(modelcg, rownames(datMeth))
datMeth2=norm_impute(datMeth,refdat=refmeth2,normalize,missedcg)
#AdaptAge
AdaptAge <- drop(AdaptAge_model$Beta %*% datMeth2[AdaptAge_model$CpG,,drop=FALSE]) - 511.9742762
mScore$AdaptAge <- AdaptAge[match(mScore$SampleID,names(AdaptAge))]

#CausAge
CausAge <- drop(CausAge_model$Beta %*% datMeth2[CausAge_model$CpG,,drop=FALSE]) + 86.80816381
mScore$CausAge <- CausAge[match(mScore$SampleID,names(CausAge))]

#DamAge
DamAge <- drop(DamAge_model$Beta %*% datMeth2[DamAge_model$CpG,,drop=FALSE]) + 543.4315887
mScore$DamAge <- DamAge[match(mScore$SampleID,names(DamAge))]


#cAge and bAge
modelcg=unique(as.character(cAge_CpG$means$cpg))
missedcg=modelcg[!(modelcg %in% rownames(datMeth))]
refdat=data.frame(cg=as.character(cAge_CpG$means$cpg),meth_mean=cAge_CpG$means$mean)
datMeth2=norm_impute(datMeth,refdat=refdat,normalize,missedcg)
mScore$cAge=calc_cAge(datMeth2,cAge_CpG)[as.character(mScore$SampleID)]

#use episcore reference data for bAge
modelcg=unique(as.character(bAge_CpG$cpgs$CpG_Site))
missedcg=modelcg[!(modelcg %in% rownames(datMeth))]
tmp0=episcore_model[!duplicated(episcore_model$CpG_Site),]
refdat=data.frame(cg=tmp0$CpG_Site,meth_mean=tmp0$Mean_Beta_Value)
datMeth2=norm_impute(datMeth,refdat=refdat,normalize,missedcg)
tmp=calc_bAge(datMeth2,datPheno,mScore,bAge_CpG,GrimAgeComponent)[as.character(mScore$SampleID),]
mScore$bAge=tmp$bAge
mScore$bAge_Years=tmp$bAge_Years

#episcore
modelcg=unique(as.character(episcore_model$CpG_Site))
missedcg=modelcg[!(modelcg %in% rownames(datMeth))]
tmp0=episcore_model[!duplicated(episcore_model$CpG_Site),]
refdat=data.frame(cg=tmp0$CpG_Site,meth_mean=tmp0$Mean_Beta_Value)
datMeth2=norm_impute(datMeth,refdat=refdat,normalize,missedcg)

out <- data.frame(SampleID=colnames(datMeth2))
for(i in unique(episcore_model$Predictor)){
  tmp_coef = episcore_model[episcore_model$Predictor %in% i, ]
  tmp_coef=tmp_coef[tmp_coef$CpG_Site %in% rownames(datMeth2),]
  if(nrow(tmp_coef) > 1) {
    out[,i]=colSums(tmp_coef$Coefficient * datMeth2[as.character(tmp_coef$CpG_Site),],na.rm=T)
  } else {
    out[,i] = tmp_coef$Coefficient * datMeth2[as.character(tmp_coef$CpG_Site),]
  }
}
out$'Epigenetic Age (Zhang)' <- out$'Epigenetic Age (Zhang)' + 65.79295
mScore <- cbind(mScore, out[match(mScore$SampleID, out$SampleID), setdiff(colnames(out), "SampleID"), drop = FALSE])

#DNAmFitAge
FitAge=calc_DNAmFitAge(datMeth,datPheno,mScore,GrimAgeComponent,DNAmFitnessModels)
mScore <- cbind(mScore, FitAge[match(mScore$SampleID, FitAge$SampleID), setdiff(colnames(FitAge), "SampleID"), drop = FALSE])



#column order
mScore=mScore[,c("SampleID",as.character(methscore_dict$predictor))]
###################Age residuals
if(AgeResid){
if(!is.null(datPheno)){
    rownames(datPheno)=datPheno$SampleID
    datPheno=datPheno[as.character(mScore$SampleID),]
    clockColumns=colnames(mScore)[!(colnames(mScore)=="SampleID")]
    mScore[,"Age"] = datPheno$Age
    for (i in clockColumns){
        mScore[,paste0(i,"Resid")] = resid(lm(mScore[,i] ~ datPheno$Age))
    }
}}
return(mScore)
}

#calculation of PC Clocks
#Higgins-Chen et al, PMID 36277076,https://www.ncbi.nlm.nih.gov/pmc/articles/PMC9586209/
calcPCClocks <- function(methdat, phenodat,pcc=NULL){

  if(!("Age" %in% colnames(phenodat))){
    stop("Error: phenodat must have a column named Age")
  }
  if(!("Female" %in% colnames(phenodat))){
    stop("Error: phenodat must have a column named Female")
  }
  if(!("SampleID" %in% colnames(phenodat))){
    stop("Error: phenodat must have a column named SampleID")
  }else if(sum(colnames(methdat) %in% phenodat$SampleID)<ncol(methdat)){
     stop("Error: some samples in methdat do not have phenotype in phenodat")
  }else{
     rownames(phenodat)=phenodat$SampleID
     phenodat=phenodat[colnames(methdat),]
  }

anti.trafo <-function(x,adult.age=20) { ifelse(x<0, (1+adult.age)*exp(x)-1, (1+adult.age)*x+adult.age) }

  #check number of missing CpGs
  flag=apply(is.na(methdat),1,sum)!=ncol(methdat)
  methdat=methdat[flag,]
#  load(file = paste(path_to_PCClocks_directory,"pcclock_model.RData", sep = ""))
#  if(is.null(pcc)){load(system.file("pcclock_model.RData",package="ENmix"))}
  CpGs=names(pcc$pcCpGs)
  methdat=methdat[rownames(methdat) %in% CpGs,]
  missingCpGs=CpGs[!(CpGs %in% rownames(methdat))]
#  message(paste0("PC clocks: Missing ",length(missingCpGs)," out of ",length(CpGs)," CpGs required for PC clocks calculation"))
  #impute missing probes and values
  tmp=matrix(rep(pcc$pcCpGs[missingCpGs],ncol(methdat)),ncol=ncol(methdat))
  rownames(tmp)=missingCpGs
  methdat=rbind(methdat,tmp)
  meanimpute <- function(x) ifelse(is.na(x),mean(x,na.rm=T),x)
  methdat <- apply(methdat,1,meanimpute)
  methdat=methdat[,CpGs]
  phenodat=phenodat[rownames(methdat),]
  #Calculate PC Clocks
  #Initialize a data frame for PC clocks
  DNAmAge <- data.frame(phenodat)
#  message("Calculating PC Clocks now")
  DNAmAge$PCHorvath1 <- as.numeric(anti.trafo(sweep(as.matrix(methdat),2,pcc$PCHorvath1$center) %*% pcc$PCHorvath1$model + pcc$PCHorvath1$intercept))
  DNAmAge$PCHorvath2 <- as.numeric(anti.trafo(sweep(as.matrix(methdat),2,pcc$PCHorvath2$center) %*% pcc$PCHorvath2$model + pcc$PCHorvath2$intercept))
  DNAmAge$PCHannum <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCHannum$center) %*% pcc$PCHannum$model + pcc$PCHannum$intercept)
  DNAmAge$PCPhenoAge <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCPhenoAge$center) %*% pcc$PCPhenoAge$model + pcc$PCPhenoAge$intercept)
  DNAmAge$PCDNAmTL <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCDNAmTL$center) %*% pcc$PCDNAmTL$model + pcc$PCDNAmTL$intercept)
  pp=cbind(Female = DNAmAge$Female,Age = DNAmAge$Age)
  DNAmAge$PCPACKYRS <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCGrimAge$center) %*% pcc$PCPACKYRS$model + pp %*% pcc$PCPACKYRS$pfa + pcc$PCPACKYRS$intercept)
  DNAmAge$PCADM <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCGrimAge$center) %*% pcc$PCADM$model + pp %*% pcc$PCADM$pfa + pcc$PCADM$intercept)
  DNAmAge$PCB2M <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCGrimAge$center) %*% pcc$PCB2M$model + as.matrix(pp[,2]) %*% pcc$PCB2M$pfa + pcc$PCB2M$intercept)
  DNAmAge$PCCystatinC <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCGrimAge$center) %*% pcc$PCCystatinC$model + pp %*% pcc$PCCystatinC$pfa + pcc$PCCystatinC$intercept)
  DNAmAge$PCGDF15 <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCGrimAge$center) %*% pcc$PCGDF15$model + as.matrix(pp[,2]) %*% pcc$PCGDF15$pfa + pcc$PCGDF15$intercept)
  DNAmAge$PCLeptin <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCGrimAge$center) %*% pcc$PCLeptin$model + pp %*% pcc$PCLeptin$pfa + pcc$PCLeptin$intercept)
  DNAmAge$PCPAI1 <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCGrimAge$center) %*% pcc$PCPAI1$model + as.matrix(pp[,1]) %*% pcc$PCPAI1$pfa + pcc$PCPAI1$intercept)
  DNAmAge$PCTIMP1 <- as.numeric(sweep(as.matrix(methdat),2,pcc$PCGrimAge$center) %*% pcc$PCTIMP1$model + pp %*% pcc$PCTIMP1$pfa + pcc$PCTIMP1$intercept)
  DNAmAge$PCGrimAge <- as.numeric(as.matrix(DNAmAge[,pcc$PCGrimAge$components]) %*% pcc$PCGrimAge$model + pcc$PCGrimAge$intercept)

 return(DNAmAge)
}

#cAge
calc_cAge <- function(data,cAge_CpG){
#load(system.file("cAge.RData",package="ENmix"))
coef_linear=cAge_CpG$coef_linear
coef_log=cAge_CpG$coef_log
intercept=cAge_CpG$intercept
intercept_log=cAge_CpG$intercept_log
means=cAge_CpG$means
rownames(coef_linear)=coef_linear$CpG_Site
rownames(coef_log)=coef_log$CpG_Site

data <- data[rownames(data) %in% rownames(means),]
## impute missing CpGs
if (nrow(data)!=nrow(means)) {
    missing_cpgs <- means[!rownames(means) %in% rownames(data),]
    mat=matrix(rep(missing_cpgs$mean,ncol(data)),ncol=ncol(data))
    rownames(mat)=rownames(missing_cpgs)
    colnames(mat)=colnames(data)
    data <- rbind(data,mat)
}
## impute missing values
na_to_mean <-function(x) {x[is.na(x)] <- mean(x, na.rm=T);x}
data <- t(apply(data,1,na_to_mean))

## Prep for linear predictor
cg1=as.character(coef_linear$CpG_Site[-grep('_2', coef_linear$CpG_Site)])
cg2=as.character(coef_linear$CpG_Site[grep('_2', coef_linear$CpG_Site)])
cg2=gsub("_2","",cg2)
tmp <- data[cg2,]^2;rownames(tmp)=paste0(rownames(tmp),"_2")
scores_linear <- rbind(data[cg1,],tmp)

## Prep for log predictor
cg1=as.character(coef_log$CpG_Site[-grep('_2', coef_log$CpG_Site)])
cg2=as.character(coef_log$CpG_Site[grep('_2', coef_log$CpG_Site)])
cg2=gsub("_2","",cg2)
tmp <- data[cg2,]^2;rownames(tmp)=paste0(rownames(tmp),"_2")
scores_log <- rbind(data[cg1,],tmp)

## Calculate cAge with linear model
coef_linear <- coef_linear[rownames(scores_linear),]
pred_linear_pp <- colSums(scores_linear * coef_linear$Coefficient) + intercept

## Identify any individuals predicted as under 20s, and re-run with model trained on log(age)
over20s <- names(pred_linear_pp[pred_linear_pp > 20])
pred_linear_pp1 <- pred_linear_pp[over20s]
under20s <- names(pred_linear_pp[pred_linear_pp < 20])

## Now re-run model for those
coef_log <- coef_log[rownames(scores_log),]
pred_log_pp <- colSums(scores_log * coef_log$Coefficient) + intercept_log
pred_log_pp <- exp(pred_log_pp[under20s])

c(pred_log_pp, pred_linear_pp1)
}

#bAge
calc_bAge<-function(data,phenodat,grim,bAge_CpG,GrimAgeComponent){
grim$DNAmGrimAge=grim$PCGrimAge
rownames(grim)=grim$SampleID
rownames(phenodat)=phenodat$SampleID
cpgs=bAge_CpG$cpgs
coefficients=bAge_CpG$coefficients
##standardize CpG by CpG
coef <- data[rownames(data) %in% as.character(cpgs$CpG_Site),]
ids <- colnames(coef)
coef <- t(apply(coef, 1, scale))
colnames(coef) <- ids

##impute missing CpGs
if (nrow(coef) != length(unique(cpgs$CpG_Site))) {
    missing_cpgs = cpgs[-which(cpgs$CpG_Site %in% rownames(coef)), c("CpG_Site", "Mean_Beta_Value")]
    missing_cpgs=missing_cpgs[!duplicated(as.character(missing_cpgs$CpG_Site)),]
    mat=matrix(rep(missing_cpgs$Mean_Beta_Value,ncol(coef)),ncol=ncol(coef))
    rownames(mat)=as.character(missing_cpgs$CpG_Site)
    colnames(mat)=colnames(coef)
    coef = rbind(coef,mat)
}
## impute NA with CpG means
na_to_mean <-function(methyl) {
  methyl[is.na(methyl)] <- mean(methyl, na.rm=T)
  return(methyl)
}
coef <- t(apply(coef,1,function(x) na_to_mean(x)))

#Calculate Episcores
loop <- unique(cpgs$Predictor)
out <- data.frame()
for(i in loop){
  tmp_coef = cpgs[cpgs$Predictor %in% i, ]
  tmp=coef[as.character(tmp_coef$CpG_Site),]
  if(nrow(tmp_coef) > 1) {
    out[colnames(coef),i]=colSums(tmp_coef$Coefficient*tmp,na.rm=T)
  } else {
    out[colnames(coef),i] = tmp*tmp_coef$Coefficient
  }
}

###### Standadize GrimAge components
samples <- rownames(out)
grim_pred <- grim[samples, c("DNAmGrimAge"), drop = FALSE]
if(GrimAgeComponent %in% c(1,2)){
  grim <- grim[samples, c("DNAmADM", "DNAmB2M", "DNAmCystatinC", "DNAmGDF15", "DNAmLeptin", "DNAmPACKYRS", "DNAmPAI1", "DNAmTIMP1")]
}else{
    grim <- grim[samples, c("PCADM", "PCB2M", "PCCystatinC", "PCGDF15", "PCLeptin", "PCPACKYRS", "PCPAI1", "PCTIMP1")]
}
colnames(grim)=c("DNAmADM", "DNAmB2M", "DNAmCystatinC", "DNAmGDF15", "DNAmLeptin", "DNAmPACKYRS", "DNAmPAI1", "DNAmTIMP1")
grim <- scale(grim)
## Calculate bAge
scores <- cbind(Age=phenodat[samples, c("Age")], grim, out)
scores <- scores[, coefficients$Variable]
pred_pp <- colSums(coefficients[,"Coefficient"] * t(scores))

## Scale to same scale as age
scale_pred <- function(x, mean_pred, sd_pred, mean_test, sd_test) {
  scaled <- mean_test + (x - mean_pred)*(sd_test/sd_pred)
  return(scaled)
}
# Scale to same Z scale
scale_Z <- function(x, mean_pred, sd_pred) {
  scaled <- (x - mean_pred)/sd_pred
  return(scaled)
}

mean_pred <- mean(pred_pp)
mean_test <- mean(phenodat$Age) # Mean age in testing data
sd_pred <- sd(pred_pp)
sd_test <- sd(phenodat$Age) # SD age in testing data

bAge <- scale_Z(pred_pp, mean_pred, sd_pred)
bAge_Years <- scale_pred(pred_pp, mean_pred, sd_pred, mean_test, sd_test)

pred=data.frame(SampleID=names(bAge),bAge=bAge,bAge_Years=bAge_Years)
rownames(pred)=pred$SampleID
return(pred)
}


#######DNAmFitAge begin
calc_DNAmFitAge<-function(datMeth,datPheno,mScore,GrimAgeComponent,DNAmFitnessModels){

	if (GrimAgeComponent == 1) {
           grim1 <- mScore[, c("SampleID", "GrimAgeV1")]
        } else if (GrimAgeComponent == 2) {
           grim1 <- mScore[, c("SampleID", "GrimAgeV2")]
        } else {
           grim1 <- mScore[, c("SampleID", "PCGrimAge")]
	}
        names(grim1)[2] <- "DNAmGrimAge"

	pheno=merge(datPheno,grim1,by="SampleID")
	pheno=pheno[,c("SampleID","Age","Female","DNAmGrimAge")]
	rownames(pheno)=pheno$SampleID
	pheno=pheno[colnames(datMeth),]

data_prep <- function(dataset,pheno){
  dataset <- dataset[rownames(dataset) %in% DNAmFitnessModels$AllCpGs,]
  if(nrow(dataset) != length(DNAmFitnessModels$AllCpGs)){
      cpgs_toadd <- DNAmFitnessModels$AllCpGs[!DNAmFitnessModels$AllCpGs %in% rownames(dataset)]

      #separate by sex to impute missing CpGs
      output=NULL
      rownames(pheno)=pheno$SampleID
      cid=intersect(colnames(dataset), rownames(pheno))
      dataset=dataset[,cid];pheno=pheno[cid,]
      if(sum(pheno$Female==1)>0){
         dat <- dataset[,pheno$Female == 1]
         dat=rbind(as.matrix(dat),matrix(rep(DNAmFitnessModels$Female_Medians_All[cpgs_toadd],ncol(dat)),ncol=ncol(dat),dimnames=list(cpgs_toadd,colnames(dat))))
         output=cbind(output,dat)
      }
      if(sum(pheno$Female==0)>0){
         dat <- dataset[,pheno$Female == 0]
         dat=rbind(as.matrix(dat),matrix(rep(DNAmFitnessModels$Male_Medians_All[cpgs_toadd],ncol(dat)),ncol=ncol(dat),dimnames=list(cpgs_toadd,colnames(dat))))
         output=cbind(output,dat)
      }

#      print(paste0("Total ", length(cpgs_toadd)," Missing CpGs that are assigned median values from training data"))
      dataset=output
  }
  return(dataset)
}
# Function to provide estimates for any 1 DNAm fitness models
# TidyModel is a specific model in DNAmFitnessModels list
DNAmEstimatorAnyModel <- function(dataset, TidyModel){
  intercept=matrix(rep(1.0, ncol(dataset)),ncol=ncol(dataset),dimnames=list("(Intercept)",colnames(dataset)))
  dataset=rbind(intercept,dataset)
  dataset <- dataset[as.character(TidyModel$term),]
  dm=dimnames(dataset)
  dataset=matrix(as.numeric(dataset),nrow=nrow(dataset))
  dimnames(dataset)=dm
  estimate <- colSums(TidyModel$estimate  * dataset)
  return(estimate)
}

# Function to calculate all DNAm fitness estimates #
DNAmFitnessEstimators <- function(data, pheno){
  rownames(pheno)=pheno$SampleID
  pheno=pheno[colnames(data),]
  data=rbind(t(pheno),data)

  if(sum(pheno$Female ==1)>0){
  data_fem <- data[,pheno$Female ==1]
  fem_est1 <- DNAmEstimatorAnyModel(dataset = data_fem, TidyModel = DNAmFitnessModels$Gait_noAge_Females) # gait without age
  fem_est2 <- DNAmEstimatorAnyModel(dataset = data_fem, TidyModel = DNAmFitnessModels$Grip_noAge_Females) # grip
  fem_est3 <- DNAmEstimatorAnyModel(dataset = data_fem, TidyModel = DNAmFitnessModels$VO2maxModel) # vo2max
  fem_est4 <- DNAmEstimatorAnyModel(dataset = data_fem, TidyModel = DNAmFitnessModels$Gait_wAge_Females) # gait w age
  fem_est5 <- DNAmEstimatorAnyModel(dataset = data_fem, TidyModel = DNAmFitnessModels$Grip_wAge_Females) # grip w age
  fem_est6 <- DNAmEstimatorAnyModel(dataset = data_fem, TidyModel = DNAmFitnessModels$FEV1_wAge_Females) # fev1 w age
  }
  if(sum(pheno$Female ==0)>0){
  data_male <- data[,pheno$Female ==0]
  male_est1 <- DNAmEstimatorAnyModel(dataset = data_male, TidyModel = DNAmFitnessModels$Gait_noAge_Males) # gait
  male_est2 <- DNAmEstimatorAnyModel(dataset = data_male, TidyModel = DNAmFitnessModels$Grip_noAge_Males) # grip
  male_est3 <- DNAmEstimatorAnyModel(dataset = data_male, TidyModel = DNAmFitnessModels$VO2maxModel) # vo2max
  male_est4 <- DNAmEstimatorAnyModel(dataset = data_male, TidyModel = DNAmFitnessModels$Gait_wAge_Males) # gait
  male_est5 <- DNAmEstimatorAnyModel(dataset = data_male, TidyModel = DNAmFitnessModels$Grip_wAge_Males) # grip
  male_est6 <- DNAmEstimatorAnyModel(dataset = data_male, TidyModel = DNAmFitnessModels$FEV1_wAge_Males) # fev1
  }
  if(sum(pheno$Female ==1)==0){est1=male_est1;est2=male_est2;est3=male_est3;est4=male_est4;est5=male_est5;est6=male_est6
    }else if(sum(pheno$Female ==0)==0){est1=fem_est1;est2=fem_est2;est3=fem_est3;est4=fem_est4;est5=fem_est5;est6=fem_est6
    }else{
          est1 <- c(fem_est1, male_est1)
          est2 <- c(fem_est2, male_est2)
          est3 <- c(fem_est3, male_est3)
          est4 <- c(fem_est4, male_est4)
          est5 <- c(fem_est5, male_est5)
          est6 <- c(fem_est6, male_est6)
    }
  data_and_est=data.frame(DNAmGait_noAge=est1,DNAmGrip_noAge=est2,DNAmVO2max=est3,DNAmGait_wAge=est4,DNAmGrip_wAge=est5,DNAmFEV1_wAge=est6)
  data_and_est$SampleID=rownames(data_and_est)
  return(data_and_est)
}


# Function to calculate DNAmFitAge
FitAgeEstimator <- function(data){
  # prep dataset for fitage- dataset by sex
  fem <- data[data$Female == 1,]
  male <- data[data$Female == 0,]

  # can only estimate FitAge if all variables are present, remove those without them #
  fem_comcase <- fem[complete.cases(fem), ]
  male_comcase <- male[complete.cases(male), ]

  # Female FitAge
  female_fitest <- 0.1044232 * ((fem_comcase$DNAmVO2max - 46.825091) / (-0.13620215)) +
                0.1742083 * ((fem_comcase$DNAmGrip_noAge - 39.857718) / (-0.22074456)) +
                0.2278776 * ((fem_comcase$DNAmGait_noAge - 2.508547) / (-0.01245682))  +
                0.4934908 * ((fem_comcase$DNAmGrimAge - 7.978487) / (0.80928530))

  # Male FitAge
  male_fitest <- 0.1390346 * ((male_comcase$DNAmVO2max - 49.836389) / (-0.141862925)) +
                0.1787371 * ((male_comcase$DNAmGrip_noAge - 57.514016) / (-0.253179827)) +
                0.1593873 * ((male_comcase$DNAmGait_noAge - 2.349080) / (-0.009380061))  +
                0.5228411 * ((male_comcase$DNAmGrimAge - 9.549733) / (0.835120557))

  fem <- data.frame(fem_comcase, DNAmFitAge = female_fitest)
  male <- data.frame(male_comcase, DNAmFitAge = male_fitest)

  resu <- rbind(fem, male)
  resu=resu[,!names(resu) %in% c("Age","Female","DNAmGrimAge")]
  return(resu)
}

sample_data_prep <- data_prep(dataset = datMeth,pheno=pheno)
sample_data_FitnessEst <- DNAmFitnessEstimators(sample_data_prep, pheno)
sample_data_FitAge_prep <- merge(pheno, sample_data_FitnessEst, by = "SampleID")
FitAge <- FitAgeEstimator(sample_data_FitAge_prep)
return(FitAge)
}
##### end DNAmFitAge

ctrlsva<-function(rgSet,percvar=0.95,npc=1,flag=1)
{
    if(!is(rgSet, "rgDataSet") & !is(rgSet, "RGChannelSet")){
       stop("Input needs to an rgDataSet or an RGChannelSet\n")}
    if(percvar<0 | percvar>1){stop("Percentage of variation threshold needs 
    to be between 0 and 1\n")}

    if(is(rgSet, "rgDataSet")){ctrls<-metadata(rgSet)$ictrl
    }else if(is(rgSet, "RGChannelSet")){
      ctrls<-getProbeInfo(rgSet,type="Control")
      ctrls <- ctrls[ctrls$Address %in% featureNames(rgSet),]
    }
    ctrl_r <- assays(rgSet)$Red[ctrls$Address[!(ctrls$Type %in% "NEGATIVE")],]
    ctrl_g <- assays(rgSet)$Green[ctrls$Address[!(ctrls$Type %in% "NEGATIVE")],]
    ctrl_nneg=rbind(ctrl_r,ctrl_g)
#PCA of control-probe intensities
    pca <- prcomp(t(ctrl_nneg))
    eigenvalue=pca$sdev^2
    perc=eigenvalue/sum(eigenvalue)
    if(flag==1){
    npc=1
    while(sum(perc[1:npc])<percvar){npc <- npc+1}
    npc
    ctrlsva=pca$x[,1:npc]
    }else{ctrlsva=pca$x[,1:npc]}
    cat(npc," surrogate variables explain ",sum(perc[1:npc])*100, "% of 
    data variation\n")
    ctrlsva
}

lmmatrix <- function(y,cov)
{
    p <- matrix(NA,ncol=ncol(cov),nrow=ncol(y))
    for(j in 1:ncol(cov))
    {
    x <- cov[,j]
    for(i in 1:ncol(y))
    {
    fit <- summary(lm(y[,i]~x,na.action=na.omit))
    f <- fit$fstatistic
    p[i,j] <- pf(f["value"],f["numdf"],f["dendf"],lower.tail=FALSE)
    }
    }
    colnames(p) <- names(cov)
    p
}

plotp<-function(jpgfile,p,yaxis,xmax,title)
{
    jpeg(filename=jpgfile,width=1000,height=700,quality = 100)
    par(mar=c(5, 4.2, 4, 2) + 0.1)
    plot(1,xlim=c(0,xmax),ylim=c(0,length(yaxis)+1),type="n",bty="n",
    axes=FALSE,xlab="Principal Component",ylab="",main=title)
    axis(1,at=c(1:xmax),pos=0.5,las=1,lwd=3)
    for(i in 1:length(yaxis)){text(0.3,i, yaxis[i],xpd=TRUE,adj=1)}
    for(i in 1:ncol(p)){
    for(j in 1:nrow(p)){
    pp <- p[j,i]
    colcode <- "white"
    if(pp <= 10e-10){colcode="darkred"}
    else if (pp <= 10e-5){colcode="red"}
    else if (pp <= 0.01){colcode="orange"}
    else if(pp <= 0.05){colcode="pink"}
    polygon(c(j-0.5,j-0.5,j+0.5,j+0.5), c(i-0.5,i+0.5,i+0.5,i-0.5),col=colcode,
    border=NA)
    }}
    legend("topright",c("<0.05","<0.01","<10E-5","<10E-10"),
    col=c("pink","orange","red","darkred"),
    pch=15,pt.cex=2,bty="o",horiz=TRUE,xpd=TRUE)
    dev.off()
}

pcrplot<-function(beta,cov,npc=50)
{
    if(!is.matrix(beta)){stop("beta is not a data matirx")}
    if(!is.data.frame(cov)){stop("cov is not a data frame")}
    if(ncol(beta)!=nrow(cov))
    {stop("number of columns in beta is not equal to number of rows in cov")}
    cat("Analysis is running, please wait...!","\n")
    npc <- min(ncol(beta),npc)

#    library("irlba")
    svd<- prcomp_irlba(t(beta), n=npc, center=TRUE,scale.=TRUE,retx=TRUE)
    jpeg(filename="svdscreeplot.jpg",width=1000,height=500,quality = 100)
    screeplot(svd,npc,type="barplot")
    dev.off()
    cat("svdscreeplot.jpg was plotted","\n")
    prop=summary(svd)$importance["Cumulative Proportion", npc]*100
    cat("Top ",npc," principal components explain ", prop, "% of data
    variation","\n")
    p <- lmmatrix(svd$x[,1:npc],cov)
    yaxis <- colnames(p)
    plotp("pcr_diag.jpg",p,yaxis,npc,
    title="Principal Component Regression Analysis")
    cat("pcr_diag.jpg was plotted","\n")
}

estimateCellProp <-function(userdata,refdata="FlowSorted.Blood.450k",cellTypes=NULL,nonnegative = TRUE,nProbes=50,normalize=TRUE,refplot=FALSE){
probeSelect="both"
  if(!is(userdata, "rgDataSet") & !is(userdata, "methDataSet") & !is.matrix(userdata) &
     !is(userdata, "MethylSet") & !is(userdata, "RGChannelSet")){
    stop("[predSex] The input must be a rgDataSet, a methDataSet or a matrix")}
    if(is(userdata, "rgDataSet")){
      userdata=getmeth(userdata)}else if(is(userdata, "RGChannelSet")){
      userdata=preprocessRaw(userdata)}else if(is.matrix(userdata)){
      if(min(userdata,na.rm=TRUE)<0 | max(userdata,na.rm=TRUE)>1){stop("[estimateCellProp] Methylation data should within [0,1]")}
      if(normalize){normalize=FALSE; message("rgDataSet or methDataSet required for normalization, reset normalize=FALSE")}
    }

#edit
if(refdata=="FlowSorted.Blood.450k"){
  library(refdata, character.only = TRUE)
  refdata=get(refdata)
  refdata=preprocessRaw(refdata)
#CellType: Bcell  CD4T  CD8T   Eos  Gran  Mono   Neu    NK  PBMC   WBC, 6 each
# flag=refdata$CellType %in% c("PBMC","WBC","Eos","Neu")
  flag=refdata$CellType %in% c("PBMC","WBC","Gran")
  refdata=refdata[,!flag]
}else if(refdata=="FlowSorted.DLPFC.450k"){
  library(refdata, character.only = TRUE)
  refdata=get(refdata)
  refdata=preprocessRaw(refdata)
#CellType: NeuN_neg, NeuN_pos, 29 each
}else if(refdata=="FlowSorted.CordBlood.450k"){
  library(refdata, character.only = TRUE)
  refdata=get(refdata)
  refdata=preprocessRaw(refdata)
#CellType: Bcell CD4T  CD8T Gran Mono NK nRBC WholeBlood
#          15    15    14   12   15  14   4     15
 flag=refdata$CellType %in% c("WholeBlood")
 refdata=refdata[,!flag]
}else if(refdata=="FlowSorted.CordBloodNorway.450k"){
  library(refdata, character.only = TRUE)
  refdata=get(refdata)
  refdata=preprocessRaw(refdata)
#Bcell  CD4T  CD8T  Gran  Mono    NK   WBC, 11 each
 flag=refdata$CellType %in% c("WBC")
 refdata=refdata[,!flag]
}else if(refdata=="FlowSorted.Blood.EPIC"){
  newdir <- tools::R_user_dir("ExperimentHub", which="cache")
  if (!dir.exists(newdir)) {
      dir.create(newdir, recursive = TRUE)
  }
  library(refdata, character.only = TRUE)
  FlowSorted.Blood.EPIC<-libraryDataGet(refdata)
#  hub <- ExperimentHub()
#  query(hub, refdata)
#  FlowSorted.Blood.EPIC <- hub[["EH1136"]]
  refdata=get(refdata)
  refdata=preprocessRaw(refdata)
  flag=refdata$CellType %in% c("MIX")
  refdata=refdata[,!flag]
#need to check
}else if(refdata=="FlowSorted.CordBloodCombined.450k"){
  newdir <- tools::R_user_dir("ExperimentHub", which="cache")
  if (!dir.exists(newdir)) {
      dir.create(newdir, recursive = TRUE)
  }
  library(refdata, character.only = TRUE)
  FlowSorted.CordBloodCombined.450k <-libraryDataGet(refdata)
#  hub <- ExperimentHub()
#  query(hub, refdata)
#  FlowSorted.CordBloodCombined.450k <- hub[["EH2256"]]
  refdata=get(refdata)
  refdata=preprocessRaw(refdata)
#table(refdata$CellType)
#Bcell  CD4T  CD8T  Gran  Mono    NK  nRBC   WBC
#   42    41    33    43    48    45    11    26
 flag=refdata$CellType %in% c("WBC")
 refdata=refdata[,!flag]
}

    if(!is.null(cellTypes)) {
        if(!all(cellTypes %in% pd$CellType))
            stop("elements of argument 'cellTypes' is not part of 'mSet$CellType'")
        keep <- which(colData(refdata)$CellType %in% cellTypes)
        refdata <- refdata[,keep]
    }
    pd <- as.data.frame(colData(refdata))
    pd$CellType <- factor(pd$CellType)

if(sum(colnames(userdata) %in% colnames(refdata)) > 0){
   tmp=colnames(userdata)[colnames(userdata) %in% colnames(refdata)]
   stop("The following column names in user data are the same with refdata: ",tmp,"\nPlease change to different names")}

if(refplot){
#distribution of intensity data
    jpeg(filename="refdata_distribution.jpg",width=700, height=1000,quality = 100)
    par(mfrow=c(2,1))
    color=as.numeric(pd$CellType)
    tmp=rbind(assays(refdata)$Meth,assays(refdata)$Unmeth);rownames(tmp)=1:nrow(tmp)
    multifreqpoly(tmp,col=color,legend="",
    cex.main=1.5,main="Reference data intensity value distribution",xlab="Intensity value")
    ntype=length(levels(pd$CellType))
    legend("top",legend=levels(pd$CellType),lty=1,lwd=3,col=1:ntype,text.col=1:ntype,bty="n")
    rm(tmp)
#distribution of methylation data
    color=as.numeric(pd$CellType)
    multifreqpoly(assays(refdata)$Meth/(assays(refdata)$Meth+assays(refdata)$Unmeth+100),col=color,legend="",
    cex.main=1.5,main="Reference data Methylation value distribution",xlab="Methylation value")
    ntype=length(levels(pd$CellType))
    legend("top",legend=levels(pd$CellType),lty=1,lwd=3,col=1:ntype,text.col=1:ntype,bty="n")
    dev.off()
}

#check whether probe names in userdata are standard CG name
#if not, remove suffix and combined replicated CGs
if(sum(rownames(refdata) %in% rownames(userdata))<50){
if(is.matrix(userdata)){userdata=rm.cgsuffix(userdata)}else{
    Meth=rm.cgsuffix(assays(userdata)$Meth)
    Unmeth=rm.cgsuffix(assays(userdata)$Unmeth)
    rowData=data.frame(Name=dimnames(Meth)[[1]])
    rownames(rowData)=rowData[,1]
    userdata <-methDataSet(Meth=Meth,Unmeth=Unmeth,rowData=as(rowData,"DataFrame"))
}}


commonprobe=intersect(rownames(userdata),rownames(refdata))
userdata=userdata[commonprobe,]
refdata=refdata[commonprobe,]

#quantile normalization
if(normalize){
Meth=cbind(assays(refdata)$Meth,assays(userdata)$Meth)
Unmeth=cbind(assays(refdata)$Unmeth,assays(userdata)$Unmeth)
rname <- rownames(Meth)
cname <- colnames(Meth)
Meth  <-  normalize.quantiles2(Meth)
Unmeth  <-  normalize.quantiles2(Unmeth)
methy=Meth/(Meth+Unmeth+100)
rownames(methy)=rname
colnames(methy)=cname
ref=methy[,colnames(refdata)]
userdata=methy[,colnames(userdata)]
}else{
ref=getB(refdata)
if(is(userdata,"methDataSet")){userdata=getB(userdata)
}else if(is(userdata,"MethylSet")){userdata=getB(userdata)}
}

if(refplot){
#distribution of methylation data
    jpeg(filename="refdata_distribution_after_qc.jpg",width=700, height=500,quality = 100)
    color=as.numeric(pd$CellType)
    multifreqpoly(ref,col=color,legend="",
    cex.main=1.5,main="Reference data Methylation value distribution",xlab="Methylation value")
    ntype=length(levels(pd$CellType))
    legend("top",legend=levels(pd$CellType),lty=1,lwd=3,col=1:ntype,text.col=1:ntype,bty="n")
    dev.off()
}

#select a set of CpG probes best differentiating celltypes
#require(genefilter)
Ftest=genefilter::rowFtests(ref, pd$CellType)
ref=ref[!is.na(Ftest$p.value) & Ftest$p.value<1e-8,]

    tIndexes <- split(x=seq(length(pd$CellType)),f=factor(pd$CellType))
    tstatList <- lapply(tIndexes, function(i) {
        x <- rep(0,ncol(ref))
        x[i] <- 1
        genefilter::rowttests(ref, factor(x))
    })
    if (probeSelect == "any"){
        probeList <- lapply(tstatList, function(x) {
            y <- x[x[,"p.value"] < 1e-8,]
            yAny <- y[order(abs(y[,"dm"]), decreasing=TRUE),]
            c(rownames(yAny)[1:(nProbes*2)])
        })
    } else {
        probeList <- lapply(tstatList, function(x) {
            y <- x[x[,"p.value"] < 1e-8,]
            yUp <- y[order(y[,"dm"], decreasing=TRUE),]
            yDown <- y[order(y[,"dm"], decreasing=FALSE),]
            c(rownames(yUp)[1:nProbes], rownames(yDown)[1:nProbes])
        })
    }
    trainingProbes <- unique(unlist(probeList))
    trainingProbes=trainingProbes[!is.na(trainingProbes)]
    ref <- ref[trainingProbes,]
    userdata = userdata[trainingProbes,]
    
    refmean <- sapply(split(x=seq(length(pd$CellType)),f=factor(pd$CellType)), 
               function(i) rowMeans(ref[,i]))
    refmedian <- sapply(split(x=seq(length(pd$CellType)),f=factor(pd$CellType)), 
               function(i) Biobase::rowMedians(ref[,i]))
    rownames(refmedian)=rownames(ref)

if(refplot){
#if (!require("gplots")){stop("Please install gplots for methylation heatmap")}
jpeg("refdata_heatmap.jpg",width=700,height=700,quality=100)
gplots::heatmap.2(ref,labCol=pd$CellType,labRow="",col=colorRampPalette(c("blue", "white", "red"))(256),key=TRUE,trace="none")
dev.off()
}

#Estimate mean methylation level for each celltypes
if(FALSE)
{
    modelFix <- as.formula(sprintf("y ~ %s - 1", paste(levels(pd$CellType), collapse="+")))
    phenoDF <- as.data.frame(model.matrix(~pd$CellType-1))
    colnames(phenoDF) <- sub("^pd\\$CellType", "", colnames(phenoDF))
    if(ncol(phenoDF) == 2) { # two group solution
        X <- as.matrix(phenoDF)
        coefEsts <- t(solve(t(X) %*% X) %*% t(X) %*% t(ref))
    } else { # > 2 group solution

    modelBatch=NULL

    N <- dim(phenoDF)[1]
    phenoDF$y <- rep(0, N)
    xTest <- model.matrix(modelFix, phenoDF)
    sizeModel <- dim(xTest)[2]
    M <- dim(ref)[1]

    coefEsts <- matrix(NA, M, sizeModel)

    for(j in 1:M) {
        ii <- !is.na(ref[j,])
        phenoDF$y <- ref[j,]
        try({ # Try to fit a mixed model to adjust for plate
            if(!is.null(modelBatch)) {
                fit <- try(lme(modelFix, random=modelBatch, data=phenoDF[ii,]))
                OLS <- inherits(fit,"try-error") # If LME can't be fit, just use OLS
            } else{OLS <- TRUE}

            if(OLS) {
                fit <- lm(modelFix, data=phenoDF[ii,])
                fitCoef <- fit$coef
            } else {
                fitCoef <- fit$coef$fixed
            }
            coefEsts[j,] <- fitCoef
        })
    }
    rownames(coefEsts) <- rownames(ref)
    colnames(coefEsts) <- names(fitCoef)
    }
}
coefEsts=refmean
#Estimate cell type proportions
Xmat = coefEsts
nCol = dim(Xmat)[2]
nSubj = dim(userdata)[2]

  mixCoef = matrix(0, nSubj, nCol)
  rownames(mixCoef) = colnames(userdata)
  colnames(mixCoef) = colnames(Xmat)

  if(nonnegative){
#   if(!require(quadprog))stop("Can not load package quadprog")

    Amat = diag(nCol)
    b0vec = rep(0,nCol)

    for(i in 1:nSubj){
      obs = which(!is.na(userdata[,i]))
      Dmat = t(Xmat[obs,])%*%Xmat[obs,]
      mixCoef[i,] = quadprog::solve.QP(Dmat, t(Xmat[obs,])%*%userdata[obs,i], Amat, b0vec)$sol
    }
  }else{
    for(i in 1:nSubj){
      obs = which(!is.na(userdata[,i]))
      Dmat = t(Xmat[obs,])%*%Xmat[obs,]
      mixCoef[i,] = solve(Dmat, t(Xmat[obs,]) %*% userdata[obs,i])
    }
  }

  data.frame(Sample_Name=rownames(mixCoef),mixCoef)
}

libraryDataGet <- function(title) {
    assign(title,ExperimentHub()[[query(ExperimentHub(),
                                        title)$ah_id]])
}

log.erfc  <- function(x){
    log(2)+pnorm(x*sqrt(2),lower.tail=FALSE,log.p=TRUE)
}

loglik.x  <- function(x,background,lambda,mu2,sigma2,p1,p2,mu,sigma){
    if (background){
    temp.1 <- lambda*(2*mu+lambda*sigma^2-2*x)/2+log.erfc((mu+lambda*
    sigma^2-x)/(sqrt(2)*sigma))
    temp.2 <- dnorm(x,mu+mu2,sqrt(sigma^2+sigma2^2),log=TRUE)
    return(sum(ifelse(temp.1>temp.2,temp.1+log(p1*lambda/2+p2*exp(temp.2-
    temp.1)),temp.2+log(p1*lambda/2*exp(temp.1-temp.2)+p2))))
    }
    else{
    temp.1 <- dexp(x,lambda,log=TRUE)
    temp.2 <- dnorm(x,mu2,sigma2,log=TRUE)
    return(sum(ifelse(temp.1>temp.2,temp.1+log(p1+p2*exp(temp.2-temp.1)),
    temp.2+log(p1*exp(temp.1-temp.2)+p2))))
    }
}

EM_estimate  <- 
function(x,start=c(max(density(x)$y),mean(range(x)),diff(range(x))/6,0.5),
    epsilon=c(0.0001,0.001,0.001,0.001)){
    lambda <- start[1]
    mu <- start[2]
    sigma.sq <- start[3]^2
    p1 <- start[4]
    n <- length(x)

    diff <- TRUE
    while (diff){
    z <- p1/(p1+(1-p1)*exp(dnorm(x,mu,sqrt(sigma.sq),log=TRUE)-dexp(x,
    lambda,log=TRUE)))

    lambda.new <- sum(z)/sum(x*z)
    mu.new <- sum((1-z)*x)/sum(1-z)
    sigma.sq.new <- sum(((x-mu.new)^2)*(1-z))/sum(1-z)
    p1.new <- sum(z)/n

    diff <- !(all(c(abs(lambda.new-lambda),abs(mu.new-mu),
    abs(sigma.sq.new-sigma.sq),abs(p1.new-p1))<epsilon))

    lambda <- lambda.new; mu <- mu.new; sigma.sq <- sigma.sq.new; p1 <- p1.new
    }
    list(lambda,mu,sqrt(sigma.sq),p1)
}

new_cm  <- 
function(lambda,mu2,sigma2,p1,p2,mu,sigma,s){
    a <- mu+lambda*sigma^2
    C <- (sigma2^2*(s-mu)+sigma^2*mu2)/(sigma2^2+sigma^2)
    D <- sigma2*sigma/sqrt(sigma2^2+sigma^2)

    temp_1 <- p1*lambda*exp(lambda^2*sigma^2/2-lambda*(s-mu))*(pnorm(s,a,sigma)-
     pnorm(0,a,sigma))
    temp_2 <- p2*dnorm(s,mu2+mu,sqrt(sigma2^2+sigma^2))/(1-pnorm(0,mu2,sigma2))

    num_1 <- temp_1
    num_2 <- temp_2*(pnorm(s,C,D)-pnorm(0,C,D))

    denom_1 <- temp_1*(s-a+sigma*(dnorm((s-a)/sigma)-dnorm(a/sigma))/(
     pnorm(s,a,sigma)-pnorm(0,a,sigma)))
    denom_2 <- temp_2*(C*(pnorm(s,C,D)-pnorm(0,C,D))+D*(exp(-C^2/(2*D^2))
     -exp(-(s-C)^2/(2*D^2)))/sqrt(2*pi))

    (denom_1+denom_2)/(num_1+num_2)
}

firstpeak <- function(x,y,sn,dat)
{
    n <- length(y)
##sn to avoid using small peak position
    nn <- sn*2+1
    v <- matrix(NA,ncol=nn,nrow=n-nn+1)
    for(i in 1:nn){v[,i]=y[i:(n-nn+i)]}
    ix <- sn+which(apply(v<v[,(sn+1)],1,sum) == (nn-1))
    if(length(ix)>0){mu=x[ix[1]]}
    if(length(ix) == 0 | sum(dat<mu)/length(dat)>=0.15)
    {
    dat <- dat[dat<quantile(dat,p=0.10)]
    temp <- density(dat)
    flag <- temp$x>=min(dat) & temp$x<=max(dat)
    temp$x <- temp$x[flag];temp$y=temp$y[flag]
    mu <- temp$x[which.max(temp$y)]
    }
    mu
}


enmix_adj <- function(meth_i=NULL,bg_i=NULL,bgParaEst)
{
    if(sum(is.na(meth_i))>0)
    {stop("ENmix background correction does not allow missing value")}
    meth_i[meth_i <= 0]=1
    mu <- bg_i$mu[1]
    sigma <- bg_i$sigma[1]
if(bgParaEst == "est" | bgParaEst == "neg" | bgParaEst == "oob")
{
    x <- (meth_i[meth_i>=mu]-mu)
    temp <- EM_estimate(x)
    lambda <- temp[[1]]
    mu2 <- temp[[2]]
    sigma2<-ifelse(temp[[3]] <= sigma, 0.1, sqrt(temp[[3]]^2-sigma^2))

    p1 <- (sum(meth_i<mu)+temp[[4]]*length(x))/length(meth_i)
    p2 <- 1-p1
    meth_adj <- new_cm(lambda,mu2,sigma2,p1,p2,mu,sigma,meth_i)
    meth_adj[meth_adj <= 0]=0.01 #restrict to positive values, only a few
}else if (bgParaEst == "subtract_neg" | bgParaEst == "subtract_estBG"
    | bgParaEst == "subtract_q5neg" | bgParaEst == "subtract_oob"){
    meth_adj=meth_i-mu
    meth_adj[meth_adj <= 0]=0.01 #restrict to positive values
}
    meth_adj
}

enmix <- function(meth,bg,bgParaEst,nCores)
{
    colnm <- colnames(meth)
    meth.o <- foreach(i=1:ncol(meth),.combine=cbind,.export=c("EM_estimate",
     "new_cm","enmix_adj")) %dopar% {
      i=i;enmix_adj(meth[,i],bg[i,],bgParaEst)}
    if(is.matrix(meth.o)){if(sum(is.na(meth.o))>0){stop("Computation ran out
    of memory, try to set nCores with a smaller value")}}else{
     stop("Computation ran out of memory, try to set nCores with a smaller
      value")}
    colnames(meth.o)=colnm
    gc(); meth.o
}

#copy from MASS package
huber<-function (y, k = 1.5, tol = 1e-06)
{
    y <- y[!is.na(y)]
    n <- length(y)
    mu <- median(y)
    s <- mad(y)
    if (s == 0)
        stop("cannot estimate scale: MAD is zero for this sample")
    repeat {
        yy <- pmin(pmax(mu - k * s, y), mu + k * s)
        mu1 <- sum(yy)/n
        if (abs(mu - mu1) < tol * s)
            break
        mu <- mu1
    }
    list(mu = mu, s = s)
}

huber_mus <- function(x){ests <- try(huber(x)); if(class(ests)[1]=="try-error"){
    cat("Warning:Check negtive control data, or do quality control before ENmix\n");
    c(mu=median(x,na.rm=TRUE),s=sd(x,na.rm=TRUE))
    }else{c(mu=ests$mu,s=ests$s)}}
huber_mu <- function(x){ests <- try(huber(x));if(class(ests)[1]=="try-error"){
    cat("Warning: Check NORM control data, or do quality control before ENmix\n");
    median(x,na.rm=TRUE)
    }else{ests$mu}}

estBG  <- function(meth_i)
{
    meth_i[meth_i<=0]=1e-06
    temp <- density(meth_i)
    temp <- density(meth_i[meth_i<temp$x[which.max(temp$y)]])
    flag <- temp$x>=min(meth_i) & temp$x<=max(meth_i)
    temp$x <- temp$x[flag];temp$y=temp$y[flag]
    mu <- temp$x[which.max(temp$y)]
    ##first mode
    if((sum(meth_i<mu)/length(meth_i))>=0.15){mu=firstpeak(temp$x,temp$y,
      sn=5,meth_i)}
    perc <- sum(meth_i<mu)/length(meth_i)
    sigma <- sqrt(sum((meth_i[meth_i<mu]-mu)^2)/sum(meth_i<mu))
    c(mu,sigma,perc)
}

##background correction
preprocessENmix  <- function(rgSet, bgParaEst="oob", dyeCorr="RELIC",
    QCinfo=NULL, exQCsample=TRUE,
    exQCcpg=TRUE, exSample=NULL, exCpG=NULL, nCores=2)
{
    if(is(rgSet, "rgDataSet")){
    if(!is.null(QCinfo) & exQCsample){exSample=unique(c(QCinfo$badsample, exSample))}
    exSample=exSample[exSample %in% colnames(rgSet)]
    if(length(exSample)>0){
    rgSet=rgSet[,!(colnames(rgSet) %in% exSample)]
    cat(length(exSample), " samples were excluded before ENmix correction\n")
    }
    mdat <- getmeth(rgSet)
    }else if(is(rgSet, "RGChannelSet")){
    if(!is.null(QCinfo) & exQCsample){exSample=unique(c(QCinfo$badsample, exSample))}
    exSample=exSample[exSample %in% colnames(rgSet)]
    if(length(exSample)>0){
    rgSet=rgSet[,!(colnames(rgSet) %in% exSample)]
    cat(length(exSample), " samples were excluded before ENmix correction\n")
    }
    mdat <- preprocessRaw(rgSet)
    }else if(is(rgSet, "methDataSet") | is(rgSet, "MethylSet")){
    if(!is.null(QCinfo) & exQCsample){exSample=unique(c(QCinfo$badsample,
     exSample))}
    exSample=exSample[exSample %in% colnames(rgSet)]
    if(length(exSample)>0){
    rgSet=rgSet[,!(colnames(rgSet) %in% exSample)]
    cat(length(exSample), " samples were excluded before ENmix correction\n")
    }
    mdat=rgSet; bgParaEst="est"; 
    if(!(dyeCorr=="none")){cat("Warning: Input data need to be a rgDataSet or 
        RGChannelSet to perform dye-bias correction\n");
       cat("Warning: dye-bias correction will not be performed\n")}
    dyeCorr="none"
    }else{stop("Error: object needs to be of class 'RGChannelSet' or 
      'MethylSet'")}
    if(nCores>detectCores()){
    nCores=detectCores(); 
    cat("Only ",detectCores(), " cores avialable, nCores was reset to ",
     detectCores(),"\n")
    }
    if(!is.null(QCinfo) & exQCcpg) {exCpG=unique(c(exCpG, QCinfo$badCpG))}
    exCpG=exCpG[exCpG %in% rownames(mdat)]
    if(length(exCpG)>0){
    mdat=mdat[!(rownames(mdat) %in% exCpG),]
    cat(length(exCpG), " CpGs were excluded before ENmix correction\n")
    }
    rm(QCinfo)
    
    if(is(mdat, "methDataSet")){
      probe_type=rowData(mdat)$Infinium_Design_Type
      col=rowData(mdat)$Color_Channel
      probe_type[probe_type %in% c("I","snpI") & col=="Grn"]="IGrn"
      probe_type[probe_type %in% c("I","snpI") & col=="Red"]="IRed"
      probe_type[probe_type %in% c("snpII")]="II"
    }else if(is(mdat, "MethylSet")){
    probe_type <- getProbeType(mdat, withColor=TRUE)}
    cat("Analysis is running, please wait...!\n")
##estimate background parameters
    if(bgParaEst == "neg" | bgParaEst == "subtract_neg")
    {
    if(is(rgSet,"rgDataSet")){
        ctrls <- getCGinfo(rgSet,type="ctrl")
    }else if(is(rgSet,"RGChannelSet")){
        ctrls <- getProbeInfo(rgSet,type="Control")}

    ctrls <- ctrls[ctrls$Address %in% rownames(rgSet),]
    ctrl_address <- as.vector(ctrls$Address[ctrls$Type %in% "NEGATIVE"])

    ctrl_r <- assays(rgSet)$Red[ctrl_address,]
    ctrl_g <- assays(rgSet)$Green[ctrl_address,]
    ctrl_r[ctrl_r<=0]=1e-06;ctrl_g[ctrl_g<=0]=1e-06
    temp <- apply(ctrl_r,2,huber_mus)
    mu <- temp["mu",];sigma <- temp["s",]
    bgRI <- as.data.frame(cbind(mu,sigma))
    temp <- apply(ctrl_g,2,huber_mus);
    mu <- temp["mu",];sigma <- temp["s",]
    bgGI <- as.data.frame(cbind(mu,sigma))
    bgRII <- bgRI;bgGII <- bgGI
    }else if (bgParaEst == "oob" | bgParaEst == "subtract_oob")
    {
    if(is(rgSet,"rgDataSet")){
        I_probe <- getCGinfo(rgSet,type="I")
        I_probe=I_probe[I_probe$Color_Channel=="Red",]
    }else if(is(rgSet,"RGChannelSet")){
        I_probe <- getProbeInfo(rgSet, type="I-Red")}
    I_green_bg_M <-  assays(rgSet)$Green[I_probe$AddressB,]
    I_green_bg_U <-  assays(rgSet)$Green[I_probe$AddressA,]
    ctrl_g <- rbind(I_green_bg_M,I_green_bg_U)
    if(is(rgSet,"rgDataSet")){
        I_probe <- getCGinfo(rgSet,type="I")
        I_probe=I_probe[I_probe$Color_Channel=="Grn",]
    }else if(is(rgSet,"RGChannelSet")){
        I_probe <- getProbeInfo(rgSet, type="I-Green")}
    I_red_bg_M <-  assays(rgSet)$Red[I_probe$AddressB,]
    I_red_bg_U <-  assays(rgSet)$Red[I_probe$AddressA,]
    ctrl_r <- rbind(I_red_bg_M,I_red_bg_U)
    ctrl_r[ctrl_r<=0]=1e-06;ctrl_g[ctrl_g<=0]=1e-06
    temp <- apply(ctrl_r,2,huber_mus)
    mu <- temp["mu",];sigma=temp["s",]
    bgRI <- as.data.frame(cbind(mu,sigma))
    temp <- apply(ctrl_g,2,huber_mus);
    mu <- temp["mu",];sigma=temp["s",]
    bgGI <- as.data.frame(cbind(mu,sigma))
    bgRII <- bgRI;bgGII=bgGI
    rm(list=c("I_green_bg_M","I_green_bg_U","ctrl_g","I_red_bg_M","I_red_bg_U",
     "ctrl_r"))

    }else if (bgParaEst == "subtract_q5neg"){
    if(is(rgSet,"rgDataSet")){
        ctrls <- getCGinfo(rgSet,type="ctrl")
    }else if(is(rgSet,"RGChannelSet")){
        ctrls <- getProbeInfo(rgSet,type="Control")}

    ctrls <- ctrls[ctrls$Address %in% rownames(rgSet),]
    ctrl_address <- as.vector(ctrls$Address[ctrls$Type %in% "NEGATIVE"])
    ctrl_r <- assays(rgSet)$Red[ctrl_address,]
    ctrl_g <- assays(rgSet)$Green[ctrl_address,]
    ctrl_r[ctrl_r<=0]=1e-06;ctrl_g[ctrl_g<=0]=1e-06  ## may not need this
    mu <- apply(ctrl_r,2,function(x) quantile(x,probs=0.05,na.rm=TRUE));
    sigma <- apply(ctrl_r,2,function(x)sd(x,na.rm=TRUE));
    bgRI <- as.data.frame(cbind(mu,sigma))
    mu <- apply(ctrl_g,2,function(x) quantile(x,probs=0.05,na.rm=TRUE));
    sigma <- apply(ctrl_g,2,function(x)sd(x,na.rm=TRUE));
    bgGI <- as.data.frame(cbind(mu,sigma))
    bgRII <- bgRI;bgGII=bgGI
    }else if(bgParaEst == "est" | bgParaEst == "subtract_estBG"){
    
    mdat_subset <- mdat[probe_type == "IRed",]
    m_I_red <- rbind(assays(mdat_subset)$Meth,assays(mdat_subset)$Unmeth)
    mdat_subset <- mdat[probe_type == "IGrn",]
    m_I_grn <- rbind(assays(mdat_subset)$Meth,assays(mdat_subset)$Unmeth)
    mdat_subset <- mdat[probe_type == "II",]
    mII <- rbind(assays(mdat_subset)$Meth,assays(mdat_subset)$Unmeth)
    rm(mdat_subset)
    bgRI <- as.data.frame(t(apply(m_I_red,2,estBG)));names(bgRI) <- c("mu",
     "sigma","perc")
    bgGI <- as.data.frame(t(apply(m_I_grn,2,estBG)));names(bgGI) <- c("mu",
     "sigma","perc")
    bgII <- as.data.frame(t(apply(mII,2,estBG)));names(bgII) <- c("mu",
     "sigma","perc")
    ##empirically adjusting the estimates
    pp <- apply(cbind(bgRI$perc,bgGI$perc,bgII$perc),1,max)-
     apply(cbind(bgRI$perc,bgGI$perc,bgII$perc),1,min)
    avgp <- apply(cbind(bgRI$perc,bgGI$perc,bgII$perc),1,mean)
    for(i in 1:nrow(bgGI)){if(pp[i]>=0.04){
    bgRI$mu[i] <- quantile(m_I_red[,i],probs=avgp[i])
    bgGI$mu[i] <- quantile(m_I_grn[,i],probs=avgp[i])
    bgII$mu[i] <- quantile(mII[,i],probs=avgp[i])
    bgRI$perc[i] <- avgp[i];bgGI$perc[i]=avgp[i];bgII$perc[i]=avgp[i]
    }}
    bgRI <- bgRI[,c("mu","sigma")]
    bgGI <- bgGI[,c("mu","sigma")]
    bgII <- bgII[,c("mu","sigma")]
    A1=sum(bgII$mu)*2/(sum(bgRI$mu)+sum(bgGI$mu))
    A2=sum(bgII$sigma)*2/(sum(bgRI$sigma)+sum(bgGI$sigma))
    bgGII=bgGI;bgGII$mu=bgGII$mu*A1;bgGII$sigma=bgGII$sigma*A2
    bgRII=bgRI;bgRII$mu=bgRII$mu*A1;bgRII$sigma=bgRII$sigma*A2
    rm(list=c("m_I_red","m_I_grn","mII"))
    }
    c1 <- makeCluster(nCores)
    registerDoParallel(c1)

    if (dyeCorr == "mean"){
    if(is(rgSet,"rgDataSet")){
        ctrls <- getCGinfo(rgSet,type="ctrl")
    }else if(is(rgSet,"RGChannelSet")){
        ctrls <- getProbeInfo(rgSet,type="Control")}
      ctrls <- ctrls[ctrls$Address %in% rownames(rgSet),]
      ctrl_r <- assays(rgSet)$Red[ctrls$Address,]
      ctrl_g <- assays(rgSet)$Green[ctrls$Address,]
      CG.controls <- ctrls$Type %in% c("NORM_C", "NORM_G")
      AT.controls <- ctrls$Type %in% c("NORM_A", "NORM_T")

      cg_grn=ctrl_g[CG.controls,]
      at_red=ctrl_r[AT.controls,]
      cg_grn <- enmix(cg_grn,bgGI,bgParaEst,nCores)
      at_red <- enmix(at_red,bgRI,bgParaEst,nCores)

      Green.avg <- apply(cg_grn,2,huber_mu)
      Red.avg <- apply(at_red,2,huber_mu)
      ref <- mean(c(Red.avg,Green.avg))
      Grn.factor <- ref/Green.avg
      Red.factor <- ref/Red.avg
      }else if(dyeCorr =="RELIC"){
    if(is(rgSet,"rgDataSet")){
        ctrls <- getCGinfo(rgSet,type="ctrl")
    }else if(is(rgSet,"RGChannelSet")){
        ctrls <- getProbeInfo(rgSet,type="Control")}

    ctrls<-ctrls[ctrls$Address %in% rownames(rgSet),]
    ctrl_r<-assays(rgSet)$Red[ctrls$Address,]
    ctrl_g<-assays(rgSet)$Green[ctrls$Address,]
    CG.controls<-ctrls$Type %in% c("NORM_C","NORM_G")
    AT.controls<-ctrls$Type %in% c("NORM_A","NORM_T")
    cg_grn<-ctrl_g[CG.controls,];rownames(cg_grn)=
     ctrls$ExtendedType[CG.controls]
    at_red<-ctrl_r[AT.controls,];rownames(at_red)=
     ctrls$ExtendedType[AT.controls]
      cg_grn <- enmix(cg_grn,bgGI,bgParaEst,nCores)
      at_red <- enmix(at_red,bgRI,bgParaEst,nCores)
      }
    rm(rgSet)

    methData <- assays(mdat)$Meth
    N=ceiling(ncol(methData)/(nCores*10))
    parts=rep(1:N,each = ceiling(ncol(methData)/N))[1:ncol(methData)]
    for(i in 1:N){
    id=which(parts==i)
    methD=methData[,id]
    methD[probe_type == "IGrn",] <- enmix(methD[probe_type == "IGrn",],
     bgGI[id,], bgParaEst, nCores)
    methD[probe_type == "IRed",] <- enmix(methD[probe_type == "IRed",],
     bgRI[id,], bgParaEst, nCores)
    methD[probe_type == "II",] <- enmix(methD[probe_type == "II",],
     bgGII[id,], bgParaEst, nCores)
    methData[,id]=methD;
    }
    if (dyeCorr == "mean"){
      methData[probe_type == "IGrn",] <- sweep(methData[probe_type == "IGrn",],
       2, FUN="*", Grn.factor)
      methData[probe_type == "II",] <- sweep(methData[probe_type == "II",], 2,
       FUN="*", Grn.factor)
      methData[probe_type == "IRed",] <- sweep(methData[probe_type == "IRed",],
       2, FUN="*", Red.factor)
    }
    assays(mdat)$Meth <- methData
    rm(methData)

    unmethData <- assays(mdat)$Unmeth
    for(i in 1:N){
    id=which(parts==i)
    unmethD=unmethData[,id]
    unmethD[probe_type == "IGrn",] <- enmix(unmethD[probe_type == "IGrn",],
     bgGI[id,], bgParaEst, nCores)
    unmethD[probe_type == "IRed",] <- enmix(unmethD[probe_type == "IRed",],
     bgRI[id,], bgParaEst, nCores)
    unmethD[probe_type == "II",] <- enmix(unmethD[probe_type == "II",],
     bgRII[id,], bgParaEst, nCores)
    unmethData[,id]=unmethD;
    }
    if (dyeCorr == "mean"){
      unmethData[probe_type == "IGrn",] <- sweep(unmethData[probe_type ==
       "IGrn",], 2, FUN="*", Grn.factor)
      unmethData[probe_type == "IRed",] <- sweep(unmethData[probe_type ==
       "IRed",], 2, FUN="*", Red.factor)
      unmethData[probe_type == "II",] <- sweep(unmethData[probe_type ==
       "II",], 2, FUN="*", Red.factor)
    }
    assays(mdat)$Unmeth <- unmethData
    rm(unmethData)
    stopCluster(c1)
    if(dyeCorr =="RELIC"){mdat=relic(mdat,at_red,cg_grn)}

    annotation=c(paste("Backgroud_corr: ENmix,",bgParaEst,sep=""),
         paste("dyeBiasCorrection: ",dyeCorr,sep=""))
    if(is(mdat, "methDataSet")){
      metadata(mdat)$preprocessMethod=annotation
    }else if(is(mdat, "MethylSet")){
    mdat@preprocessMethod <- annotation
    }
    mdat
}

#detection P values
calcdetP <- function(rgSet, detPtype = "negative") {
    if(!is(rgSet, "rgDataSet") & !is(rgSet, "RGChannelSetExtended"))
      stop("[QCinfo] The input should be an object of 'rgDataSet' or 'RGChannelSetExtended'")

    if(is(rgSet, "RGChannelSetExtended")){
    locusNames <- getManifestInfo(rgSet, "locusNames")
    detP <- matrix(NA_real_, ncol = ncol(rgSet), nrow = length(locusNames),
                   dimnames = list(locusNames, colnames(rgSet)))
    if(detPtype == "negative"){
    controlIdx <- getControlAddress(rgSet, controlType = "NEGATIVE")
    rBg <- assays(rgSet)$Red[controlIdx,,drop=FALSE]
    gBg <- assays(rgSet)$Green[controlIdx,,drop=FALSE]
    }else if(detPtype == "oob"){
       Igrn=getProbeInfo(rgSet,type="I-Green")
       Igrn=c(Igrn$AddressA,Igrn$AddressB)
       Ired=getProbeInfo(rgSet,type="I-Red")
       Ired=c(Ired$AddressA,Ired$AddressB)
    rBg <- assays(rgSet)$Red[Igrn,,drop=FALSE]
    gBg <- assays(rgSet)$Green[Ired,,drop=FALSE]
    }
    rMu <- matrixStats::colMedians(rBg)
    rSd <- matrixStats::colMads(rBg)
    gMu <- matrixStats::colMedians(gBg)
    gSd <- matrixStats::colMads(gBg)

    TypeII <- getProbeInfo(rgSet, type = "II")
    TypeI.Red <- getProbeInfo(rgSet, type = "I-Red")
    TypeI.Green <- getProbeInfo(rgSet, type = "I-Green")

    Ired_int=assays(rgSet)$Red[TypeI.Red$AddressA,] + assays(rgSet)$Red[TypeI.Red$AddressB,]
    Igrn_int=assays(rgSet)$Green[TypeI.Green$AddressA,] + assays(rgSet)$Green[TypeI.Green$AddressB,]
    II_int=assays(rgSet)$Red[TypeII$AddressA, ] + assays(rgSet)$Green[TypeII$AddressA, ]
    for (i in 1:ncol(rgSet)) {
        detP[TypeI.Red$Name, i] <- 1-pnorm(Ired_int[,i], mean=rMu[i]*2, sd=rSd[i]*sqrt(2))
        detP[TypeI.Green$Name, i] <- 1-pnorm(Igrn_int[,i], mean=gMu[i]*2, sd=gSd[i]*sqrt(2))
        detP[TypeII$Name, i] <- 1-pnorm(II_int[,i], mean=rMu[i]+gMu[i], sd=sqrt(rSd[i]^2+gSd[i]^2))
    }
}else if(is(rgSet, "rgDataSet")){
    if(detPtype == "negative"){
        ctrl=metadata(rgSet)$ictrl
        controlIdx <- ctrl$Address[ctrl$Type %in% "NEGATIVE"]
        rBg <- assays(rgSet)$Red[controlIdx,,drop=FALSE]
        gBg <- assays(rgSet)$Green[controlIdx,,drop=FALSE]
    }else if(detPtype=="oob"){
        anno=rowData(rgSet)
        rBg <- assays(rgSet)$Red[anno$Address[anno$Color_Channel %in% "Grn"],]
        gBg <- assays(rgSet)$Green[anno$Address[anno$Color_Channel %in% "Red"],]
    }

    rMu <- matrixStats::colMedians(rBg)
    rSd <- matrixStats::colMads(rBg)
    gMu <- matrixStats::colMedians(gBg)
    gSd <- matrixStats::colMads(gBg)

    cginfo=getCGinfo(rgSet)
    probeI=cginfo[cginfo$Infinium_Design_Type %in% c("I","snpI"),]
    probeIred=probeI[probeI$Color_Channel=="Red",]
    probeIgrn=probeI[probeI$Color_Channel=="Grn",]
    probeII=cginfo[cginfo$Infinium_Design_Type %in% c("II","snpII"),]

    Ired_int=assays(rgSet)$Red[probeIred$AddressA,]+assays(rgSet)$Red[probeIred$AddressB,]
    Igrn_int=assays(rgSet)$Green[probeIgrn$AddressA,]+assays(rgSet)$Green[probeIgrn$AddressB,]
    II_int=assays(rgSet)$Red[probeII$AddressA,]+assays(rgSet)$Green[probeII$AddressA,]
detP_Ired=matrix(NA_real_, ncol = ncol(Ired_int), nrow = nrow(Ired_int),dimnames =list(probeIred$Name,colnames(Ired_int)))
detP_Igrn=matrix(NA_real_, ncol = ncol(Igrn_int), nrow = nrow(Igrn_int),dimnames =list(probeIgrn$Name,colnames(Igrn_int)))
detP_II=matrix(NA_real_, ncol = ncol(II_int), nrow = nrow(II_int),dimnames =list(probeII$Name,colnames(II_int)))
for (i in 1:ncol(rgSet)) {
detP_Ired[,i]=1-pnorm(Ired_int[,i], mean=rMu[i]*2, sd=rSd[i]*sqrt(2))
detP_Igrn[,i]=1-pnorm(Igrn_int[,i], mean=gMu[i]*2, sd=gSd[i]*sqrt(2))
detP_II[,i]=1-pnorm(II_int[,i], mean=rMu[i]+gMu[i], sd=sqrt(rSd[i]^2+gSd[i]^2))
}
detP=rbind(detP_Ired,detP_Igrn,detP_II)
    }
    detP
}

# Exported classes -------------------------------------------------------------

setClass(
    "methDataSet",
    contains = "SummarizedExperiment"
)

# Exported functions -----------------------------------------------------------

methDataSet <- function(Meth = new("matrix"), Unmeth = new("matrix"),
    rowData=new("DataFrame"),...) {
    assays <- SimpleList(Meth = Meth, Unmeth = Unmeth)
    metadata <- list(preprocessMethod="")
    new("methDataSet",
        SummarizedExperiment(assays = assays, metadata=metadata,
        rowData=rowData,...)
    )
}

# Exported methods -------------------------------------------------------------

setMethod("show", signature(object = "methDataSet"), function(object) {
    callNextMethod()
    cat("PreprocessMethod: ", metadata(object)$preprocessMethod,"\n")
})

## ----pipeline, eval=FALSE-----------------------------------------------------
# rgSet <- readidat(datapath)
# beta=mpreprocess(rgSet)

## ----example1, eval=FALSE-----------------------------------------------------
# library(ENmix)
# #read in data
# require(minfiData)
# #read in IDAT files
# path <- file.path(find.package("minfiData"),"extdata")
# rgSet <- readidat(path = path,recursive = TRUE)
# #data pre-processing
# beta=mpreprocess(rgSet,nCores=6)
# #quality control, data pre-processing and imputation
# beta=mpreprocess(rgSet,nCores=6,qc=TRUE,fqcfilter=TRUE,
#      rmcr=TRUE,impute=TRUE)
# 
# #CpG information (from Illumina manifest file) is self-contained in rgDataSet or methDataSet
# cginfo=getCGinfo(rgSet)
# #It has the same infomation if extracted from methDataSet
# meth=getmeth(rgSet)
# cginfo1=rowData(meth)
# #Probe information for internal controls
# ictrl=getCGinfo(rgSet,type="ctrl")

## ----example2, eval=FALSE-----------------------------------------------------
# library(ENmix)
# #read in data
# path <- file.path(find.package("minfiData"),"extdata")
# rgSet <- readidat(path = path,recursive = TRUE)
# #QC info
# qc<-QCinfo(rgSet)
# #background correction and dye bias correction
# #if qc info object is provided, the low quality or outlier samples
# # and probes will be excluded before background correction
# mdat<-preprocessENmix(rgSet, bgParaEst="oob", dyeCorr="RELIC",
#                      QCinfo=qc, nCores=6)
# #between-array normalization
# mdat<-norm.quantile(mdat, method="quantile1")
# #probe-type bias adjustment
# beta<-rcp(mdat,qcscore=qc)
# #set low quality and outlier data points as NA (missing value)
# #remove samples and probes with too many (eg. >5%) missing data
# #perform imputation to replace missing values
# beta <- qcfilter(beta,qcscore=qc,rmcr=TRUE,impute=TRUE)
# #remove suffix and combine duplicated CpGs with rm.cgsuffix(), for EPIC v2
# #beta2=rm.cgsuffix(beta)

## ----example3, eval=FALSE-----------------------------------------------------
# library(ENmix)
# #read in data
# path <- file.path(find.package("minfiData"),"extdata")
# rgSet <- readidat(path = path,recursive = TRUE)
# 
# #attach some phenotype info for later use
# sheet <- read.metharray.sheet(file.path(find.package("minfiData"),
#          "extdata"),pattern = "csv$")
# rownames(sheet)=paste(sheet$Slide,sheet$Array,sep="_")
# colData(rgSet)=as(sheet[colnames(rgSet),],"DataFrame")
# 
# #generate internal control plots to inspect quality of experiments
# plotCtrl(rgSet)
# 
# #generate quality control (QC) information and plots,
# #identify outlier samples in data distribution
# #if set detPtype="negative", recommend to set depPthre to <= 10E-6
# #if set detPtype="oob", depPthre of 0.01 or 0.05 may be sufficient
# #see Heiss et al. PMID: 30678737 for details
# qc<-QCinfo(rgSet,detPtype="negative",detPthre=0.000001)
# 
# ###data preprocessing
# #background correction and dye bias correction
# mdat<-preprocessENmix(rgSet, bgParaEst="oob", dyeCorr="RELIC",
#                       QCinfo=qc, nCores=6)
# #between-array normalization
# mdat<-norm.quantile(mdat, method="quantile1")
# #attach phenotype data for later use
# colData(mdat)=as(sheet[colnames(mdat),],"DataFrame")
# #probe-type bias adjustment
# beta<-rcp(mdat,qcscore=qc)
# 
# #Search for multimodal CpGs, so called gap probes
# #beta matrix before qcfilter() step should be used for this purpose
# nmode<-nmode(beta, minN = 3, modedist=0.2, nCores = 6)
# 
# #filter low quality and outlier data points for each probe
# #rows and columns with too many missing values will be removed if specify
# #Imputation will be performed to fill missing data if specify.
# beta <- qcfilter(beta,qcscore=qc,rmcr=TRUE, rthre=0.05,
#                    cthre=0.05,impute=TRUE)
# 
# #If for epigenetic mutation analysis, outliers should be kept
# beta <- qcfilter(beta,qcscore=qc,rmoutlier=FALSE,rmcr=TRUE, rthre=0.05,
#                    cthre=0.05,impute=FALSE)
# 
# #remove suffix and combine duplicated CpGs with rm.cgsuffix(), for EPIC v2
# #beta2=rm.cgsuffix(beta)
# 
# ##Principal component regression analysis plot to check data variance structure
# #phenotypes to be studied in the plot
# cov<-data.frame(group=colData(mdat)$Sample_Group,
#     slide=factor(colData(mdat)$Slide))
# #missing data is not allowed in the analysis!
# pcrplot(beta, cov, npc=6)
# 
# #Non-negative control surrogate variables, which can be used in
# # downstream association analysis to control for batch effects.
# csva<-ctrlsva(rgSet)
# 
# #estimate cell type proportions
# #rgDataset or methDataSet can also be used for the estimation
# celltype=estimateCellProp(userdata=beta,
#          refdata="FlowSorted.Blood.450k",
#          nonnegative = TRUE,normalize=TRUE)
# 
# #predic sex based on rgDataSet or methDataset
# sex=predSex(rgSet)
# sex=predSex(mdat)
# 
# #Methylation predictors (>150), including methylation age calculation
# pheno=data.frame(SampleID=colnames(beta),Age=c(23,45,52,36,58,43),Female=c(1,0,1,1,0,1))
# mscore=methscore(datMeth=beta,datPheno=pheno)
# 
# #ICC analysis can be performed to study measurement reliability if
# # have duplicates, see manual
# #dupicc()
# 
# #After association analysis, the P values can be used for DMR analysis
# #simulate a small example file in BED format
# dat=simubed()
# #using ipdmr method, low seed value (0.01 or 0.05) should be used in real study.
# ipdmr(data=dat,seed=0.1)
# #using comb-p method
# combp(data=dat,seed=0.1)

## ----readdata, eval=FALSE-----------------------------------------------------
# library(ENmix)
# rgSet <- readidat(path = "directory path for idat files",
#                  recursive = TRUE)
# 
# #Download manifestfile for HumanMethylation450 beadchip
# system("wget https://webdata.illumina.com/downloads/productfiles/humanmethylation450/humanmethylation450_15017482_v1-2.csv")
# mf="HumanMethylation450_15017482_v1-2.csv"
# rgSet <- readidat(path = "path to idat files",manifestfile=mf,recursive = TRUE)

## ----readdata2, eval=FALSE----------------------------------------------------
# M<-matrix_for_methylated_intensity
# U<-matrix_for_unmethylated_intensity
# pheno<-as.data.frame(cbind(colnames(M), colnames(M)))
# names(pheno)<-c("Basename","filenames")
# rownames(pheno)<-pheno$Basename
# pheno<-AnnotatedDataFrame(data=pheno)
# anno<-c("IlluminaHumanMethylation450k", "ilmn12.hg19")
# names(anno)<-c("array", "annotation")
# mdat<-MethylSet(Meth = M, Unmeth = U, annotation=anno,
# phenoData=pheno)

## ----ctrlplot, eval=FALSE-----------------------------------------------------
# plotCtrl(rgSet)

## ----ctrlplot2, eval=FALSE----------------------------------------------------
# 
# path <- file.path(find.package("minfiData"),"extdata")
# rgSet <- readidat(path = path,recursive = TRUE)
# sheet <- read.metharray.sheet(file.path(find.package("minfiData"),
#          "extdata"), pattern = "csv$")
# rownames(sheet)=paste(sheet$Slide,sheet$Array,sep="_")
# colData(rgSet)=as(sheet[colnames(rgSet),],"DataFrame")
# #control plots
# IDorder=rownames(colData(rgSet))[order(colData(rgSet)$Slide,
#         colData(rgSet)$Array)]
# plotCtrl(rgSet,IDorder)

## ----distplot, eval=FALSE-----------------------------------------------------
# 
# mraw <- getmeth(rgSet)
# #total intensity plot is userful for data quality inspection
# #and identification of outlier samples
# multifreqpoly(assays(mraw)$Meth+assays(mraw)$Unmeth,
#              xlab="Total intensity")
# #Compare frequency polygon plot and density plot
# beta<-getB(mraw)
# anno=rowData(mraw)
# beta1=beta[anno$Infinium_Design_Type=="I",]
# beta2=beta[anno$Infinium_Design_Type=="II",]
# library(geneplotter)
# jpeg("dist.jpg",height=900,width=600)
# par(mfrow=c(3,2))
# multidensity(beta,main="Multidensity")
# multifreqpoly(beta,main="Multifreqpoly",xlab="Beta value")
# multidensity(beta1,main="Multidensity: Infinium I")
# multifreqpoly(beta1,main="Multifreqpoly: Infinium I",
# xlab="Beta value")
# multidensity(beta2,main="Multidensity: Infinium II")
# multifreqpoly(beta2,main="Multifreqpoly: Infinium II",
# xlab="Beta value")
# dev.off()

## ----filter, eval=FALSE-------------------------------------------------------
# qc<-QCinfo(rgSet)
# #QCinfo returns "detP","nbead","bisul","badsample","badCpG","outlier_sample"
# #Samples with low quality data
# qc$badsample
# #CpGs with low quality data
# qc$badCpG
# #outlier samples
# qc$outlier_sample
# 
# #Low quality samples and probes should be excluded before data preprocesssing
# #by specifying `QCinfo` in `preprocessENmix()`
# mdat<-preprocessENmix(rgSet, QCinfo=qc, nCores=6)

## ----rmoutlier, eval=FALSE----------------------------------------------------
# #filter outlier values only
# b1=qcfilter(beta)
# #filter low quality and outlier values
# b2=qcfilter(beta,qcscore=qc)
# #filter low quality and outlier values, remove rows
# #and columns with too many missing values
# b3=qcfilter(beta,qcscore=qc,rmcr=TRUE)
# #filter low quality and outlier values, remove rows and
# #columns with too many (rthre=0.05,cthre=0.05, 5% in default) missing values,
# # and then do imputation
# b3=qcfilter(beta,qcscore=qc,rmcr=TRUE,rthre=0.05,
#               cthre=0.05,impute=TRUE)

## ----preprocessENmix, eval=FALSE----------------------------------------------
# qc=QCinfo(rgSet)
# mdat<-preprocessENmix(rgSet, bgParaEst="oob", dyeCorr="RELIC",
# QCinfo=qc, exCpG=NULL, nCores=6)

## ----normalize.quantile, eval=FALSE-------------------------------------------
# mdat<-norm.quantile(mdat, method="quantile1")

## ----rcp, eval=FALSE----------------------------------------------------------
# beta<-rcp(mdat)

## ----ctrlsva, eval=FALSE------------------------------------------------------
# sva<-ctrlsva(rgSet)

## ----pcrplot, eval=FALSE------------------------------------------------------
# require(minfiData)
# mdat <- preprocessRaw(RGsetEx)
# beta=getBeta(mdat, "Illumina")
# group=pData(mdat)$Sample_Group
# slide=factor(pData(mdat)$Slide)
# cov=data.frame(group,slide)
# pcrplot(beta,cov,npc=6)

## ----nmode, eval=FALSE--------------------------------------------------------
# nmode<- nmode(beta, minN = 3, modedist=0.2, nCores = 5)

## ----celltype, eval=FALSE-----------------------------------------------------
# require(minfidata)
# path <- file.path(find.package("minfiData"),"extdata")
# #based on rgDataset
# rgSet <- readidat(path = path,recursive = TRUE)
# celltype=estimateCellProp(userdata=rgSet,
#          refdata="FlowSorted.Blood.450k",
#          nonnegative = TRUE,normalize=TRUE)

## ----mage, eval=FALSE---------------------------------------------------------
# require(minfidata)
# path <- file.path(find.package("minfiData"),"extdata")
# #based on rgDataset
# rgSet <- readidat(path = path,recursive = TRUE)
# meth=getmeth(rgSet)
# beta=getB(meth)
# mage=methyAge(beta)
# 
# pheno=data.frame(SampleID=colnames(beta),Age=c(23,45,52,36,58,43),Female=c(1,0,1,1,0,1))
# mscore=methscore(datMeth=beta,datPheno=pheno)

## ----dmr, eval=FALSE----------------------------------------------------------
# dat=simubed()
# names(dat)
# ipdmr(data=dat,seed=0.1)
# combp(data=dat,seed=0.1)

## ----icc, eval=FALSE----------------------------------------------------------
# require(minfiData)
# path <- file.path(find.package("minfiData"),"extdata")
# rgSet <- readidat(path = path,recursive = TRUE)
# mdat=getmeth(rgSet)
# beta=getB(mdat,"Illumina")
# #assuming list in id1 are corresponding duplicates of id2
# dupidx=data.frame(id1=c("5723646052_R02C02","5723646052_R04C01",
#                        "5723646052_R05C02"),
#                   id2=c("5723646053_R04C02","5723646053_R05C02",
#                        "5723646053_R06C02"))
# iccresu<-dupicc(dat=beta,dupid=dupidx)

## ----oxbs, eval=FALSE---------------------------------------------------------
# load(system.file("oxBS.MLE.RData",package="ENmix"))
# resu<-oxBS.MLE(beta.BS,beta.oxBS,N.BS,N.oxBS)
# dim(resu[["5mC"]])
# dim(resu[["5hmC"]])

## ----ENmixAndminfi, eval=FALSE------------------------------------------------
# library(ENmix)
# #minfi functions to read in data
# sheet <- read.metharray.sheet(file.path(find.package("minfiData"),
#  "extdata"), pattern = "csv$")
# rgSet <- read.metharray.exp(targets = sheet, extended = TRUE)
# #ENmix function for control plot
# plotCtrl(rgSet)
# #minfi functions to extract methylation and annotation data
# mraw <- preprocessRaw(rgSet)
# beta<-getB(mraw, "Illumina")
# anno=getAnnotation(rgSet)
# #ENmix function for fast and accurate distribution plot
# multifreqpoly(beta,main="Data distribution")
# multifreqpoly(beta[anno$Type=="I",],main="Data distribution, type I")
# multifreqpoly(beta[anno$Type=="II",],main="Data distribution, type II")
# #ENmix background correction
# mset<-preprocessENmix(rgSet, bgParaEst="oob", dyeCorr="RELIC", nCores=6)
# #minfi functions for further preprocessing and analysis
# gmSet <- preprocessQuantile(mset)
# bumps <- bumphunter(gmSet, design = model.matrix(~ gmSet$status), B = 0,
# type = "Beta", cutoff = 0.25)

## ----sessionInfo, echo=FALSE--------------------------------------------------
sessionInfo()

rcp <- function(mdat,dist=25,quantile.grid=seq(0.001,0.999,by=0.001),
    qcscore=NULL,nbthre=3, detPthre=0.000001)
{
    if(!is(mdat, "methDataSet") & !is(mdat, "MethylSet")){
       stop("the input needs to be of class 'methDataSet' or 'MethylSet'")}

    beta<-getB(mdat,type="Illumina")
    raw.M<-B2M(beta)

# find therby pairs of type I probes and type II probes
    if(is(mdat, "methDataSet")){
      annotation=rowData(mdat)
      names(annotation)[which(names(annotation)=="Infinium_Design_Type")]="Type"
      annotation=annotation[annotation$Type %in% c("I","II"),]
      rownames(annotation)=annotation$Name
    }else if(is(mdat, "MethylSet")){
      annotation<-getAnnotation(mdat)
    }
 
    annotation=annotation[as.vector(annotation$Name) %in% rownames(beta),]

    probe.II.Name=annotation$Name[annotation$Type=="II"]
    annotation=annotation[order(annotation$chr,annotation$pos),]
    anno1=annotation[1:(nrow(annotation)-1),]
    anno2=annotation[2:nrow(annotation),]

    if("anno1$Relation_to_Island" %in% names(anno1)){
    flag=(abs(anno1$pos-anno2$pos)<dist & anno1$chr==anno2$chr & 
     anno1$Relation_to_Island==anno2$Relation_to_Island & anno1$Type !=
     anno2$Type)
    }else{
    flag=(abs(anno1$pos-anno2$pos)<dist & anno1$chr==anno2$chr &
     anno1$Type !=  anno2$Type)
    }
    anno1=anno1[flag,]
    anno2=anno2[flag,]
    probe.I=anno1$Name
    probe.II=anno2$Name
    probe.I[anno2$Type=="I"]=anno2$Name[anno2$Type=="I"]
    probe.II[anno1$Type=="II"]=anno1$Name[anno1$Type=="II"]

    raw.M.t=raw.M[c(probe.I,probe.II),]

#remove low quality data
    if(is.null(qcscore)){}else if((sum(!(rownames(raw.M.t) %in% 
    rownames(qcscore$detP))) +
    sum(!(colnames(raw.M.t) %in% colnames(qcscore$detP))))>0){
    stop("Wrong qcscore matrix, please check...\n")}else{
    temp <- qcscore$nbead<nbthre | qcscore$detP>detPthre
    temp=temp[rownames(raw.M.t),]
    temp=temp[,colnames(raw.M.t)]
    raw.M.t[temp]=NA
    }

#linear regression
    M.II<-raw.M.t[probe.II,]
    M.I<-raw.M.t[probe.I,]
    
    qtl<-function(x) quantile(x, quantile.grid, na.rm=TRUE)
    M.I=apply(M.I,2,qtl)
    M.II=apply(M.II,2,qtl)

    beta.est<-mat.or.vec(2,ncol(beta))

    for (i in 1:ncol(beta)){
    index<-(M.II[,i]!=Inf & M.II[,i]!=-Inf & M.I[,i]!=Inf & M.I[,i]!=-Inf)
    X<-cbind(rep(1,sum(index)),M.II[index,i]); Y<-M.I[index,i]
    beta.est[,i]<-solve(t(X)%*%X)%*%t(X)%*%Y
    }

    M.II.all<-raw.M[probe.II.Name,]
    M.II.new<-mat.or.vec(nrow(M.II.all),ncol(M.II.all))
    for (i in 1:ncol(M.II.all)){
    M.II.new[,i]<-beta.est[1,i]+beta.est[2,i]*M.II.all[,i]
    }
    M.II.new[M.II.all==Inf]<-Inf; M.II.new[M.II.all==-Inf]<-(-Inf)

    beta[probe.II.Name,]<-M2B(M.II.new)
    beta
}

#get probe info from rgDataSet
getCGinfo <-function(rgSet,type="IandII"){
    if(!is(rgSet, "rgDataSet")){stop("[getCGinfo] The input object must be a rgDataSet")}
    if(!(type %in% c("I","II","IandII","ctrl"))){stop("[getRawMeth] Type must be I,II,IandII,or ctrl")}
    probeinfo=rowData(rgSet)
probeIA=probeinfo[probeinfo$Infinium_Design_Type %in% c("IA"),]
probeIB=probeinfo[probeinfo$Infinium_Design_Type %in% c("IB"),]
names(probeIA)[which(names(probeIA)=="Address")]="AddressA"
probeIA$Infinium_Design_Type="I"
probeIB=probeIB[,c("Name","Address")]
names(probeIB)=c("Name","AddressB")
probeI=merge(probeIA,probeIB,by="Name")
var1=c("Name","AddressA","AddressB","Infinium_Design_Type","Color_Channel","chr","pos","Relation_to_Island")
var2=names(probeinfo)[!(names(probeinfo) %in% c(var1,"Address"))]
var=c(var1,var2)
var=var[var %in% names(probeI)]
probeI=probeI[,var]

probesnpIA=probeinfo[probeinfo$Infinium_Design_Type %in% c("snpIA"),]
probesnpIB=probeinfo[probeinfo$Infinium_Design_Type %in% c("snpIB"),]
names(probesnpIA)[which(names(probesnpIA)=="Address")]="AddressA"
probesnpIB=probesnpIB[,c("Name","Address")]
names(probesnpIB)=c("Name","AddressB")
probesnpI=merge(probesnpIA,probesnpIB,by="Name")
probesnpI=probesnpI[,var]
probesnpI$Infinium_Design_Type="snpI"

probeI=rbind(probeI,probesnpI)

probeII=probeinfo[probeinfo$Infinium_Design_Type %in% c("II","snpII"),]
names(probeII)[which(names(probeII)=="Address")]="AddressA"
probeII$AddressB=probeII$AddressA
probeII=probeII[,var]

probectrl=probeinfo[probeinfo$Infinium_Design_Type %in% c("ctrl"),]
probectrl=merge(probectrl,metadata(rgSet)$ictrl,by="Address")

if(type=="I"){
return(probeI)
}else if(type=="II"){
return(probeII)
}else if (type=="ctrl"){
return(probectrl)
}else if (type=="IandII"){
return(rbind(probeI,probeII))
}
}

readmanifest <- function(file) {
    control.line <- try(system(sprintf("grep -n \\\\[Controls\\\\] %s", file), intern = TRUE),silent=TRUE)
    if(class(control.line)[1]=="try-error"){control.line <- system(sprintf("findstr -n \\[Controls\\] %s", file), intern = TRUE)}

    control.line <- as.integer(sub(":.*", "", control.line))
    stopifnot(length(control.line) == 1 && is.integer(control.line) && !is.na(control.line))

    assay.line <- try(system(sprintf("grep -n \\\\[Assay\\\\] %s", file), intern = TRUE),silent=TRUE)
    if(class(assay.line)[1]=="try-error"){assay.line <- system(sprintf("findstr -n \\[Assay\\] %s", file), intern = TRUE)}

    assay.line <- as.integer(sub(":.*", "", assay.line))
    stopifnot(length(assay.line) == 1 && is.integer(assay.line) && !is.na(assay.line))
    colNames <- readLines(file, n = assay.line + 1L)[assay.line + 1L]
    colNames <- strsplit(colNames, ",")[[1]]
    colClasses <- rep("character", length(colNames))
    names(colClasses) <- colNames
    names(colClasses) <- make.names(names(colClasses))
    colClasses[c("MAPINFO")] <- "integer"

    manifest <- read.table(file, header = FALSE, col.names = names(colClasses),
                           sep = ",", comment.char = "", quote = "",
                           skip = assay.line + 1L, colClasses = colClasses,
                           nrows = control.line - assay.line - 2L)


    if("IlmnID" %in% names(manifest)){
        if("Name"  %in% names(manifest)){names(manifest)[which(names(manifest)=="Name")]="Name1"}
        names(manifest)[which(names(manifest)=="IlmnID")]="Name"
    }
    if(!("Name"  %in% names(manifest))){stop("No IlmnID or Name in manifest file")}

    manifest$AddressA_ID <- gsub("^0*", "", manifest$AddressA_ID)
    manifest$AddressB_ID <- gsub("^0*", "", manifest$AddressB_ID)

if(sum(table(manifest$Name)>1)>0){manifest$Name=paste0(manifest$Name,"_",manifest$AddressA_ID)}

#fix Illumina coding in mouse array manifest file
manifest$Infinium_Design_Type=as.character(as.vector(manifest$Infinium_Design_Type))
manifest$Infinium_Design_Type[manifest$Infinium_Design_Type=="1"] ="I"
manifest$Infinium_Design_Type[manifest$Infinium_Design_Type=="2"] ="II"

typeI=manifest[manifest$Infinium_Design_Type=="I",]
typeI$Infinium_Design_Type[grep("^rs",typeI$Name)]="snpI"
typeII=manifest[manifest$Infinium_Design_Type=="II",]
typeII$Infinium_Design_Type[grep("^rs",typeII$Name)]="snpII"
typeIA=typeI
typeIB=typeI
typeIA$Address=typeIA$AddressA_ID
typeIB$Address=typeIB$AddressB_ID
typeII$Address=typeII$AddressA_ID
typeIA$Infinium_Design_Type=paste(typeIA$Infinium_Design_Type,"A",sep="")
typeIB$Infinium_Design_Type=paste(typeIB$Infinium_Design_Type,"B",sep="")

var1=c("Name","Address","Infinium_Design_Type","Color_Channel","CHR","MAPINFO","Relation_to_UCSC_CpG_Island")
var2=names(manifest)[!(names(manifest) %in% c(var1,"AddressB_ID","AddressA_ID","IlmnID"))]
var12=c(var1,var2)
var12=var12[var12 %in% names(typeIA)]
typeIA=typeIA[,var12]
typeIB=typeIB[,var12]
typeII=typeII[,var12]

assay.anno=rbind(typeIA,typeIB,typeII)
names(assay.anno)[which(names(assay.anno)=="CHR")]="chr"
names(assay.anno)[which(names(assay.anno)=="MAPINFO")]="pos"
names(assay.anno)[which(names(assay.anno)=="Relation_to_UCSC_CpG_Island")]="Relation_to_Island"

    controls <- read.table(
        file = file,
        skip = control.line,
        sep = ",",
        comment.char = "",
        quote = "",
        colClasses = c(rep("character", 5)))[, 1:5]
    ictrl.anno <- controls[, 1:4]
    names(ictrl.anno) <- c("Address", "Type", "Color", "ExtendedType")

flag=(assay.anno$Address=="") | is.na(assay.anno$Address)
if(sum(flag)>0){assay.anno=assay.anno[!flag,]}
flag=(ictrl.anno$Address=="") | is.na(ictrl.anno$Address)
if(sum(flag)>0){ictrl.anno=ictrl.anno[!flag,]}

#check duplicate entries
assay.anno=assay.anno[!duplicated(assay.anno$Address),]
ictrl.anno=ictrl.anno[!duplicated(ictrl.anno$Address),]
assay.anno=assay.anno[!(assay.anno$Address %in% ictrl.anno$Address),]

return(list(assay=assay.anno,ictrl=ictrl.anno))
}

getmanifest<- function (arraytype,annotation)
{
#arraytype="IlluminaHumanMethylation450k"
#annotation="ilmn12.hg19"
manifestpkg=paste(arraytype,"manifest",sep="")
annopak=paste(arraytype,"anno.",annotation,sep="")


if(!require(manifestpkg, character.only = TRUE)){
    stop(sprintf("cannot load manifest package %s", manifestpkg))}
manifest=get(manifestpkg)

if(!require(annopak,character.only = TRUE)){
    stop(sprintf("cannot load annotation package %s", annopak))}
anno=get(annopak)
anno <- do.call(cbind, lapply(anno@defaults, get))

idname="Name"
    if("IlmnID" %in% names(anno)){
        idname="IlmnID"
        if("Name"  %in% names(anno)){names(anno)[which(names(anno)=="Name")]="Name1"}
        names(anno)[which(names(anno)=="IlmnID")]="Name"
    }
    if(!("Name"  %in% names(anno))){stop("No IlmnID or Name in manifest file")}

typeI=anno[anno$Type=="I",]
typeII=anno[anno$Type=="II",]
typeIA=typeI
typeIB=typeI
typeIA$Address=typeIA$AddressA
typeIB$Address=typeIB$AddressB
typeII$Address=typeII$AddressA

var=names(typeII)[!(names(typeII) %in% c("AddressB","AddressA","Color","Type","Name"))]
var=var[var %in% names(typeIA)]

typeIA=typeIA[,var]
typeIB=typeIB[,var]
typeII=typeII[,var]
anno=rbind(typeIA,typeIB,typeII)

#manifest
typeI=manifest@data$TypeI
typeIsnp=manifest@data$TypeSnpI
typeII=manifest@data$TypeII
typeIIsnp=manifest@data$TypeSnpII
ictrl.anno=manifest@data$TypeControl

typeI$Infinium_Design_Type="I"
typeIsnp$Infinium_Design_Type="snpI"
typeII$Infinium_Design_Type="II"
typeIIsnp$Infinium_Design_Type="snpII"
typeI=rbind(typeI,typeIsnp)
typeII=rbind(typeII,typeIIsnp)

typeIA=typeI[,c(idname,"AddressA","Infinium_Design_Type","Color")]
typeIA$Infinium_Design_Type=paste(typeIA$Infinium_Design_Type,"A",sep="")
typeIB=typeI[,c(idname,"AddressB","Infinium_Design_Type","Color")]
typeIB$Infinium_Design_Type=paste(typeIB$Infinium_Design_Type,"B",sep="")
typeII=typeII[,c(idname,"AddressA","Infinium_Design_Type")]
typeII$Color_Channel=""
names(typeIA)=c("IlmnID","Address","Infinium_Design_Type","Color_Channel")
names(typeIB)=c("IlmnID","Address","Infinium_Design_Type","Color_Channel")
names(typeII)=c("IlmnID","Address","Infinium_Design_Type","Color_Channel")

assay.anno=rbind(typeIA,typeIB,typeII)
names(assay.anno)[which(names(assay.anno)=="IlmnID")]="Name"

assay.anno=merge(assay.anno,anno,by="Address",all.x=TRUE)

#order variables
var1=c("Name","Address","Infinium_Design_Type","Color_Channel","chr","pos","Relation_to_Island")
var2=names(assay.anno)[!(names(assay.anno) %in% c(var1))]
assay.anno=assay.anno[,c(var1,var2)]

#check duplicate entries
assay.anno=assay.anno[!duplicated(assay.anno$Address),]
ictrl.anno=ictrl.anno[!duplicated(ictrl.anno$Address),]
assay.anno=assay.anno[!(assay.anno$Address %in% ictrl.anno$Address),]

return(list(assay=assay.anno,ictrl=ictrl.anno))
}

normalize.quantiles2 <-function(mat){
    if(any(is.na(mat))){
	    xm=is.na(mat)
	    mat[is.na(mat)]=mean(mat,na.rm=TRUE)
            ref=apply(apply(mat, 2, sort),1,mean)
            mat=apply(mat,2,function(x)ref[rank(x)])
            mat[xm]=NA
            mat}else{
            ref=apply(apply(mat, 2, sort),1,mean)
            apply(mat,2,function(x)ref[rank(x)])
    }
}

normalize.quantiles.use.target2 <-function(mat,target){
    if(any(is.na(mat))){
            xm=is.na(mat)
            mat[is.na(mat)]=mean(mat,na.rm=TRUE)
            ref=sort(target)
            mat=apply(mat,2,function(x)ref[rank(x)])
            mat[xm]=NA
            mat}else{
            ref=sort(target)
            apply(mat,2,function(x)ref[rank(x)])
    }
}


normalize.q <- function(x)
{
    if(sum(is.na(x))>0)
    {stop("Multi-array quantile normalization does not allow missing value")}
    rname <- rownames(x)
    cname <- colnames(x)
    x <- normalize.quantiles2(x)
    rownames(x) <- rname
    colnames(x) <- cname
    x
}

norm.quantile <- function(mdat,method="quantile1")
{
#   if(!require("preprocessCore")){stop("Can not load preprocessCore package")}
    if(!is(mdat, "methDataSet") & !is(mdat, "MethylSet")){
      stop("object needs to be of class 'methDataSet' or 'MethylSet'")}
    cat("Analysis is running, please wait...!","\n")

    if(is(mdat,"MethylSet")){
      anno <- getAnnotation(mdat)
    }else if(is(mdat,"methDataSet")){
      anno=rowData(mdat)
      names(anno)[which(names(anno)=="Infinium_Design_Type")]="Type"
      anno$Type[anno$Type %in% "snpI"]="I"
      anno$Type[anno$Type %in% "snpII"]="II"
    }

    if (method == "quantile1")
    {
    mdat_I <- mdat[anno$Type == "I",]
    mdat_II <- mdat[anno$Type == "II",]
    assays(mdat_I)$Meth<-normalize.q(assays(mdat_I)$Meth)
    assays(mdat_I)$Unmeth<-normalize.q(assays(mdat_I)$Unmeth)
    assays(mdat_II)$Meth<-normalize.q(assays(mdat_II)$Meth)
    assays(mdat_II)$Unmeth<-normalize.q(assays(mdat_II)$Unmeth)
    methData <- rbind(assays(mdat_I)$Meth,assays(mdat_II)$Meth)
    unmethData <- rbind(assays(mdat_I)$Unmeth,assays(mdat_II)$Unmeth)
    CpGID <- rownames(mdat)
    SampleID <- colnames(mdat)
    methData <- methData[CpGID,]
    unmethData <- unmethData[CpGID,]
    assays(mdat)$Meth<-methData
    assays(mdat)$Unmeth<-unmethData
    }
    else if(method == "quantile2")
    {
    mdat_I <- mdat[anno$Type == "I",]
    mdat_II <- mdat[anno$Type == "II",]
    mat <- rbind(assays(mdat_I)$Meth,assays(mdat_I)$Unmeth)
    mat  <-  normalize.q(mat)
    assays(mdat_I)$Meth <- mat[1:(nrow(mat)/2),]
    assays(mdat_I)$Unmeth <- mat[((nrow(mat)/2)+1):nrow(mat),]
    mat <- rbind(assays(mdat_II)$Meth,assays(mdat_II)$Unmeth)
    mat  <-  normalize.q(mat)
    assays(mdat_II)$Meth <- mat[1:(nrow(mat)/2),]
    assays(mdat_II)$Unmeth <- mat[((nrow(mat)/2)+1):nrow(mat),]
    methData <- rbind(assays(mdat_I)$Meth,assays(mdat_II)$Meth)
    unmethData <- rbind(assays(mdat_I)$Unmeth,assays(mdat_II)$Unmeth)
    CpGID <- rownames(mdat)
    SampleID <- colnames(mdat)
    methData <- methData[CpGID,]
    unmethData <- unmethData[CpGID,]
    assays(mdat)$Meth<-methData
    assays(mdat)$Unmeth<-unmethData

    }
    else if(method == "quantile3")
    {
    ##quantile normalization of combined intensity values
    ##similar with lumi lumimethyN
    mat <- rbind(assays(mdat)$Meth,assays(mdat)$Unmeth)
    mat  <-  normalize.q(mat)
    methData <- mat[1:(nrow(mat)/2),]
    unmethData <- mat[((nrow(mat)/2)+1):nrow(mat),]
    assays(mdat)$Meth<-methData
    assays(mdat)$Unmeth<-unmethData
    }
    if(is(mdat, "methDataSet")){
      annotation=metadata(mdat)$preprocessMethod
      metadata(mdat)$preprocessMethod=c(annotation,paste("Normalization: ", method,sep=""))
    }else if(is(mdat, "MethylSet")){
      annotation=mdat@preprocessMethod
      mdat@preprocessMethod <- c(annotation,paste("Normalization: ", method,sep=""))
    }
    mdat
}

---
title: "ENmix User's Guide"
author: 
  - name: Zongli Xu
    affiliation: Biostatistics & Computational Biology Branch, National Institute of Environmental Health Sciences
  - name: Liang Niu
    affiliation: Division of Biostatistics and Bioinformatics, University of Cincinnati
  - name: Jack Taylor
    affiliation: Epidemiology Branch, National Institute of Environmental Health Sciences
package: ENmix
abstract: >
  A brief introduction of ENmix R package for DNA methylation data analysis.
vignette: >
  %\VignetteIndexEntry{ENmix User's Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    toc_float: true
---


# Introduction

The `r Biocpkg("ENmix")` package provides a set of quality control,
preprocessing/correction and data analysis tools for Illumina Methylation Beadchips. 
It includes functions to read in raw idat data, 
background correction, dye bias correction, probe-type bias adjustment, 
 along with a number of additional tools.
These functions can be used to remove unwanted experimental noise and thus to
improve accuracy and reproducibility of methylation measures. 
`r Biocpkg("ENmix")` functions 
are flexible and transparent. Users have option to choose a single pipeline 
command to finish all data pre-processing steps (including quality control, 
background correction, 
dye-bias adjustment, between-array normalization and probe-type bias correction) or 
to use individual functions sequentially to perform data pre-processing in a more 
customized manner. In addition the `r Biocpkg("ENmix")` package has selectable 
complementary functions for efficient data visualization (such as QC plots, data 
distribution plot, manhattan plot and Q-Q plot), quality control (identifing and filtering 
low quality data points, samples, probes, and outliers, along with 
imputation of missing values), identification of probes with multimodal 
distributions due to SNPs or other factors, exploration of data variance
 structure using principal component regression analysis plot, preparation
of experimental factors related surrogate control variables 
to be adjusted in downstream 
statistical analysis, an efficient algorithm oxBS-MLE to estimate
5-methylcytosine and 5-hydroxymethylcytosine level; estimation of celltype
proporitons; methlation age calculation and differentially methylated
region (DMR) analysis.

Most `r Biocpkg("ENmix")` package can also support the data structure used by 
several other related R packages, such as `r Biocpkg("minfi")`,
`r Biocpkg("wateRmelon")` and `r Biocpkg("ChAMP")`,
providing straightforward integration of 
ENmix-corrected datasets for subsequent data analysis. 

`r Biocpkg("ENmix")` readidat function does not depend on array annotation R packages.
It can directly read in Illuminal manifest file, which makes it easier to work with
newer array, such as MethylationEPICv2.0 and  mouse Beadchip.

The software is designed to support large scale data analysis, and provides
 multi-processor parallel computing options for most functions. 


# List of functions

<b>Data acquisition</b>

<ul>
<li>`readidat()`:  Read idat files into R</li>
<li>`readmanifest()`:  Read array manifest file into R</li>
</ul>

<b>Quality control</b>

<ul>
<li>`QCinfo()`:      Extract and visualize QC information</li>
<li>`plotCtrl()`:    Generate internal control plots</li>
<li>`getCGinfo()`:   Extract CpG probe annotation information</li>
<li>`calcdetP()`:    Compute detection P values</li>
<li>`qcfilter()`:  Remove low quality values, samples or CpGs; remove outlier samples and perform imputation</li>
<li>`nmode()`:   Identify "gap" probes, i.e. those with multimodal distribution from underlying caused by underlying SNPs</li>
<li>`dupicc()`:     Calculate Introclass correlation coefficient (ICC) using data for duplicates</li>
<li>`freqpoly()`:  Frequency polygon plot for single variable</li>
<li>`multifreqpoly()`:  Frequency polygon plot for multiple variables</li>
</ul>

<b>Preprocessing</b>

<ul>
<li>`mpreprocess()`:      Preprocessing pipeline</li>
<li>`preprocessENmix()`:  ENmix background correction and dye bias correction</li>
<li>`relic()`:            RELIC dye bias correction</li>
<li>`norm.quantile()`:    Quantile normalization</li>
<li>`rcp()`:              RCP probe design type bias correction</li>
</ul>

<b>Differential methylated region (DMR) analysis</b>

<ul>
<li>`ipdmr()`:    ipDMR differentially methylated region analysis</li>
<li>`combp()`:    Combp differentially methylated region analysis</li>
</ul>

<b>Other functions</b>

<ul>
<li>`oxBS.MLE()`:  MLE estimates of 5-methylcytosine (5mC) and 5-hydroxymethylcytosine (5hmC)</li>
<li>`estimateCellProp()`:    Estimate white blood cell type proportions</li>
<li>`methyAge()`:     Calculate methylation age</li>
<li>`methscore()`:    calculate various methylation predictors, including DNA methylation age, exposures and plasma protein levels.</li>
<li>`predSex()`:      Estimate sample sex</li>
<li>`ctrlsva()`:      Derive surrogate variables to control for experimental confounding using non-negative internal control probes</li>
<li>`pcrplot()`:      Principal component regression plot </li>
<li>`mhtplot()`:      P value manhattan plot </li>
<li>`p.qqplot()`:    P value Q-Q plot  </li>
<li>`B2M()`:          Convert Beta value to M value </li>
<li>`M2B()`:          Convert M value to Beta value  </li>
</ul>

# ENmix classes

ENmix organizes data with two different classes.

`rgDataSet` contains raw data (including internal control probes) from IDAT file, CpG annotation from Illumina manifest file and/or sample inforamtion (plate, array, and phenotypes) provided by users. Array intensity data is organized by probe (not CpG locus) at red and green channel.  

`methDataSet` contains methylated and unmethylated intensity values (organized by CpG), CpG annotation from Illumina manifest file and/or sample inforamtion (plate, array, and phenotypes) provided by users.

![The ENmix class hierarchy](./fig/ENmix_class.png)


# Example Analysis

The following examples are brief demonstrations on how to perform DNA methylation analysis using ENmix functions.


## Example 1: using pipeline

The pipeline function `mpreprocess()` can be used to perform quality 
control and data preprocessing.

```{r pipeline, eval=FALSE}
rgSet <- readidat(datapath)
beta=mpreprocess(rgSet)
```
These can achieve the following tasks:

<ul>
  <li>Read in raw idat files</li>
  <li>Data preprocessing</li>
    <ul>
      <li>background correction</li>
      <li>dye-bias correction</li>
      <li>between-array normalization</li>
      <li>probe-type bias adjustment</li>
    </ul>
  <li>Quality control</li>
    <ul>
      <li> Identify and exclude low quality probes</li>
      <li> Identify and exclude low quality samples</li>
      <li> Identify and exclude outlier samples</li>
      <li> Remove low quality and outlier data points</li>
      <li> Imputation to replace missing values</li>
    </ul>
  <li>Visualization</li>
    <ul>
      <li>Data distribution plots before and after low quality sample removal</li>
      <li>Plots of quality controls measures (detection P values, bisulfite control intensities) to guide selection of thresholds</li>
    </ul>
</ul>

The following is an executable example.

```{r example1, eval=FALSE}
library(ENmix)
#read in data
require(minfiData)
#read in IDAT files
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
#data pre-processing
beta=mpreprocess(rgSet,nCores=6)
#quality control, data pre-processing and imputation
beta=mpreprocess(rgSet,nCores=6,qc=TRUE,fqcfilter=TRUE,
     rmcr=TRUE,impute=TRUE)

#CpG information (from Illumina manifest file) is self-contained in rgDataSet or methDataSet
cginfo=getCGinfo(rgSet)
#It has the same infomation if extracted from methDataSet
meth=getmeth(rgSet)
cginfo1=rowData(meth)
#Probe information for internal controls
ictrl=getCGinfo(rgSet,type="ctrl")
```

## Example 2: using individual function

The example code below is basically doing the same thing as in example 1, but 
has more customized options. 

```{r example2, eval=FALSE}
library(ENmix)
#read in data
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
#QC info
qc<-QCinfo(rgSet)
#background correction and dye bias correction
#if qc info object is provided, the low quality or outlier samples
# and probes will be excluded before background correction
mdat<-preprocessENmix(rgSet, bgParaEst="oob", dyeCorr="RELIC", 
                     QCinfo=qc, nCores=6)
#between-array normalization
mdat<-norm.quantile(mdat, method="quantile1")
#probe-type bias adjustment
beta<-rcp(mdat,qcscore=qc)
#set low quality and outlier data points as NA (missing value)
#remove samples and probes with too many (eg. >5%) missing data
#perform imputation to replace missing values
beta <- qcfilter(beta,qcscore=qc,rmcr=TRUE,impute=TRUE)
#remove suffix and combine duplicated CpGs with rm.cgsuffix(), for EPIC v2
#beta2=rm.cgsuffix(beta)
```

## Example 3: A more comprehensive example

This example is a brief demonstration of major functions in `r Biocpkg("ENmix")`. 
Getting detailed information about data quality, SNP-like/gap probes, excluding 
user-specified probes and/or samples, principal component regression analysis,
low quality or outlier sample or probe removal, sex prediction, 
cell type proportion estimation, methylation age estimation, differentially 
methylated region analysis and surrogate variables for batch effect, etc. 

```{r example3, eval=FALSE}
library(ENmix)
#read in data
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)

#attach some phenotype info for later use
sheet <- read.metharray.sheet(file.path(find.package("minfiData"),
         "extdata"),pattern = "csv$")
rownames(sheet)=paste(sheet$Slide,sheet$Array,sep="_")
colData(rgSet)=as(sheet[colnames(rgSet),],"DataFrame")

#generate internal control plots to inspect quality of experiments
plotCtrl(rgSet)

#generate quality control (QC) information and plots, 
#identify outlier samples in data distribution
#if set detPtype="negative", recommend to set depPthre to <= 10E-6
#if set detPtype="oob", depPthre of 0.01 or 0.05 may be sufficient
#see Heiss et al. PMID: 30678737 for details
qc<-QCinfo(rgSet,detPtype="negative",detPthre=0.000001)

###data preprocessing
#background correction and dye bias correction
mdat<-preprocessENmix(rgSet, bgParaEst="oob", dyeCorr="RELIC",
                      QCinfo=qc, nCores=6)
#between-array normalization
mdat<-norm.quantile(mdat, method="quantile1")
#attach phenotype data for later use
colData(mdat)=as(sheet[colnames(mdat),],"DataFrame")
#probe-type bias adjustment
beta<-rcp(mdat,qcscore=qc)

#Search for multimodal CpGs, so called gap probes
#beta matrix before qcfilter() step should be used for this purpose
nmode<-nmode(beta, minN = 3, modedist=0.2, nCores = 6)

#filter low quality and outlier data points for each probe
#rows and columns with too many missing values will be removed if specify
#Imputation will be performed to fill missing data if specify.
beta <- qcfilter(beta,qcscore=qc,rmcr=TRUE, rthre=0.05,
                   cthre=0.05,impute=TRUE)

#If for epigenetic mutation analysis, outliers should be kept 
beta <- qcfilter(beta,qcscore=qc,rmoutlier=FALSE,rmcr=TRUE, rthre=0.05,
                   cthre=0.05,impute=FALSE)

#remove suffix and combine duplicated CpGs with rm.cgsuffix(), for EPIC v2
#beta2=rm.cgsuffix(beta)

##Principal component regression analysis plot to check data variance structure
#phenotypes to be studied in the plot
cov<-data.frame(group=colData(mdat)$Sample_Group,
    slide=factor(colData(mdat)$Slide))
#missing data is not allowed in the analysis!
pcrplot(beta, cov, npc=6)

#Non-negative control surrogate variables, which can be used in
# downstream association analysis to control for batch effects. 
csva<-ctrlsva(rgSet)

#estimate cell type proportions
#rgDataset or methDataSet can also be used for the estimation
celltype=estimateCellProp(userdata=beta,
         refdata="FlowSorted.Blood.450k",
         nonnegative = TRUE,normalize=TRUE)

#predic sex based on rgDataSet or methDataset
sex=predSex(rgSet)
sex=predSex(mdat)

#Methylation predictors (>150), including methylation age calculation
pheno=data.frame(SampleID=colnames(beta),Age=c(23,45,52,36,58,43),Female=c(1,0,1,1,0,1))
mscore=methscore(datMeth=beta,datPheno=pheno)

#ICC analysis can be performed to study measurement reliability if
# have duplicates, see manual
#dupicc()

#After association analysis, the P values can be used for DMR analysis
#simulate a small example file in BED format
dat=simubed()
#using ipdmr method, low seed value (0.01 or 0.05) should be used in real study.
ipdmr(data=dat,seed=0.1)
#using comb-p method
combp(data=dat,seed=0.1)
```

# Setting up the data

The first step is to import array raw data files (*.idat)
 to create an object of `rgDataSet`, which is similar to 
`RGChannelSetExtended` in minfi package, but with probe annotation 
integrated, and it has smaller data size (about 1/3 smaller). Alternatively, User can also 
use minfi package to read in idat files and create 
`RGChannelSetExtended`. Most functions in ENmix support 
`RGChannelSetExtended` and `RGChannelSet`.

Some array types and corresponding manifestfiles can be guessed by the program based on
the number of probes per array. However, we recommend users provide the correct manifest file (.csv format)
directly for their types of array, which can be easily downloaded from Illumina website, see below for some examples.
This option also allows the function to read in data from newer array, such as mouse array. 

```{r readdata, eval=FALSE}
library(ENmix)
rgSet <- readidat(path = "directory path for idat files", 
                 recursive = TRUE)

#Download manifestfile for HumanMethylation450 beadchip
system("wget https://webdata.illumina.com/downloads/productfiles/humanmethylation450/humanmethylation450_15017482_v1-2.csv")
mf="HumanMethylation450_15017482_v1-2.csv"
rgSet <- readidat(path = "path to idat files",manifestfile=mf,recursive = TRUE)
```

When methylation IDAT raw data files are not available, such as for some
 publicly available datasets, users can use methylated (M) and unmethylated
 (U) intensity data to create an object of `MethylSet`. `MethylSet`
is also supported by most functions in ENmix.

```{r readdata2, eval=FALSE}
M<-matrix_for_methylated_intensity
U<-matrix_for_unmethylated_intensity
pheno<-as.data.frame(cbind(colnames(M), colnames(M)))
names(pheno)<-c("Basename","filenames")
rownames(pheno)<-pheno$Basename
pheno<-AnnotatedDataFrame(data=pheno)
anno<-c("IlluminaHumanMethylation450k", "ilmn12.hg19")
names(anno)<-c("array", "annotation")
mdat<-MethylSet(Meth = M, Unmeth = U, annotation=anno, 
phenoData=pheno)
```

# Quality Control
## Internal control probes

Both Illumina 450k and EPIC array incorporate 15 different types of internal control
probes (total of 848 probes for 450K and 635 probes for EPIC). The control 
plots from ENmix function `plotCtrl()` are similar to the control plots generated  
by Illumina GenomeStudio software. See Illumina Infinium HD Methylation Assay for 
detailed description on how to interpret these control figures. 

```{r ctrlplot, eval=FALSE}
plotCtrl(rgSet)
```
 
Below is a list of internal control types and number of probes for each type. 

<table style="width:70%">
<tr style="border-top:2pt solid black;">
<td><b>Control types</b></td><td><b>450K</b></td><td><b>EPIC v1</b></td><td><b>EPIC v2</b></td>
</tr>
<tr style="border-top:1pt solid black;"><td><b>Sample-Independent Controls</b></td></tr>
<tr>
<td>STAINING</td><td>4</td><td>6</td><td>4</td></tr>
<tr><td>EXTENSION </td><td> 4 </td><td> 4</td><td>4</td></tr>
<tr><td>HYBRIDIZATION</td><td> 3 </td><td> 3</td><td>3</td></tr>
<tr><td>TARGET REMOVAL</td><td> 2 </td><td> 2</td><td>2</td></tr>
<tr><td>RESTORATION </td><td> 1 </td><td> 1</td><td>1</td></tr>
<tr style="border-top:1pt solid black;">
<td><b>Sample-Dependent Controls</b></td></tr>
<tr><td>BISULFITE CONVERSION I </td><td> 12 </td><td> 10</td><td>10</td></tr>
<tr><td>BISULFITE CONVERSION II </td><td> 4 </td><td> 4</td><td>4</td></tr>
<tr><td>SPECIFICITY I </td><td> 12 </td><td> 12</td><td>12</td></tr>
<tr><td>SPECIFICITY II </td><td> 3 </td><td> 3</td><td>3</td></tr>
<tr><td>NON-POLYMORPHIC </td><td> 4 </td><td> 9</td><td>9</td></tr>
<tr><td>NORM_A </td><td> 32 </td><td> 27</td><td>27</td></tr>
<tr><td>NORM_C </td><td> 61 </td><td> 58</td><td>58</td></tr>
<tr><td>NORM_G </td><td> 32 </td><td> 27</td><td>27</td></tr>
<tr><td>NORM_T </td><td> 61 </td><td> 58</td><td>58</td></tr>
<tr style="border-bottom:2pt solid black;"><td>NEGATIVE </td><td> 613 </td><td> 411</td><td>411</td></tr>
</table>



![Bisulfite conversion controls for type I probes](./fig/BISULFITE_CONVERSION_I.jpg)

![Negative control probes](./fig/NEGATIVE.jpg)


![NORM ACGT control probes](./fig/NORM_ACGT.jpg)


These controls can be plotted in user specified order to check 
how experimental factors affect methylation measures, such as batch,
 plate, array or array location.

```{r ctrlplot2, eval=FALSE}

path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
sheet <- read.metharray.sheet(file.path(find.package("minfiData"),
         "extdata"), pattern = "csv$") 
rownames(sheet)=paste(sheet$Slide,sheet$Array,sep="_")
colData(rgSet)=as(sheet[colnames(rgSet),],"DataFrame")
#control plots
IDorder=rownames(colData(rgSet))[order(colData(rgSet)$Slide,
        colData(rgSet)$Array)]
plotCtrl(rgSet,IDorder)
```

## Data distribution plots

Methylation intensity or beta value distribution plots are 
very useful for data summary, visual inspection and identification of 
outlier samples. Density plots are routinely generated using R function
`multidensity()`. However, the function is computationally intensive,
 and can take several hours to produce density plots for a large methylation 
dataset. 

Frequency polygon plot is an alternative figure for inspection of data
distribution. Similar to histogram, it can accurately reflect data distribution and
easy to understand. The function is also much faster, and only takes a few
minutes to produce a distribution plot for >1000 samples. 

```{r distplot, eval=FALSE}

mraw <- getmeth(rgSet)
#total intensity plot is userful for data quality inspection
#and identification of outlier samples
multifreqpoly(assays(mraw)$Meth+assays(mraw)$Unmeth,
             xlab="Total intensity")
#Compare frequency polygon plot and density plot
beta<-getB(mraw)
anno=rowData(mraw)
beta1=beta[anno$Infinium_Design_Type=="I",]
beta2=beta[anno$Infinium_Design_Type=="II",]
library(geneplotter)
jpeg("dist.jpg",height=900,width=600)
par(mfrow=c(3,2))
multidensity(beta,main="Multidensity")
multifreqpoly(beta,main="Multifreqpoly",xlab="Beta value")
multidensity(beta1,main="Multidensity: Infinium I")
multifreqpoly(beta1,main="Multifreqpoly: Infinium I",
xlab="Beta value")
multidensity(beta2,main="Multidensity: Infinium II")
multifreqpoly(beta2,main="Multifreqpoly: Infinium II",
xlab="Beta value")
dev.off()
```

See the following figures generated from the above code. When 
type I and type II probes are plotted separately bottom 4 panels, 
the difference in modes between type I and II probes can be appreciated. 
But when all probes are plotted together (top panel), the
multidensity plot obscures these differences, while they remain readily 
apparent in the multifreqpoly plot. In addition, the multidensity plots 
appear to suggest that probes range in value
from <0 to >1, whereas multifreqpoly correctly show the range from 0 to 1.

![Methylation beta value distribution plots for all probes (top 2 panels)
and for type I (middle panels) and II (bottom panels) probes separately. The
smoothing function in multidensity plots (panels on left) results in misleading
range and mode information which are more accurately depicted in the
multifreqpoly plots (panels on right)](./fig/dist.jpg)


## QC information, outlier samples, low quality samples and probes

Data quality measures, including detection P values, average intensities for 
bisulfite conversion probes and number of 
beads for each methylation read 
can be extracted or estimated using the 
function `QCinfo()` from an object of `rgDataSet` or
 `RGChannelSetExtended`. Based on default or user specified 
quality score thresholds, the `QCinfo()` will identify 
a list of low quality samples and CpG probes as wells as outlier samples based on 
total intensity or beta value distribution. These samples and probes
should be excluded before 
further analysis. Data quality score figures "qc_sample.jpg" and "qc_CpG.jpg" from
`QCinfo()` can be used to guide the selection of quality score 
thresholds. Low quality samples and probes 
are listed as "badsample" and "badCpG" in QCinfo object, and also marked 
in red in figure "freqpolygon\_beta\_beforeQC.jpg". 

```{r filter, eval=FALSE}
qc<-QCinfo(rgSet)
#QCinfo returns "detP","nbead","bisul","badsample","badCpG","outlier_sample"
#Samples with low quality data
qc$badsample
#CpGs with low quality data
qc$badCpG
#outlier samples
qc$outlier_sample

#Low quality samples and probes should be excluded before data preprocesssing 
#by specifying `QCinfo` in `preprocessENmix()`
mdat<-preprocessENmix(rgSet, QCinfo=qc, nCores=6)
```

![freqpolygon_beta_beforeQC.jpg](./fig/freqpolygon_beta_beforeQC.jpg)

![qc_sample.jpg](./fig/qc_sample.jpg)

![qc_CpG.jpg](./fig/qc_CpG.jpg)



## Filtering outliers, low quality data points, missing values and imputation

After excluding low quality samples and probes, as well as outlier samples.
we can still have many low quality and outlier data points for many CpGs.
These small percentage of data points can have big impact on downstream 
association statistical tests for individual CpGs. 
Function `qcfilter()` can be used to filter out these
data points and replace them with missing value NA. Outliers are defined as 
values smaller than 3 times IQR from the lower quartile or larger than 
3 times IQR from the upper quartile. Some statistical methods do not allow 
missing values,(e.g. principal component analysis), so 
 argument \textit{impute=TRUE} in the function can be 
specified to impute missing data using k-nearest neighbors method. 

```{r rmoutlier, eval=FALSE}
#filter outlier values only
b1=qcfilter(beta)
#filter low quality and outlier values
b2=qcfilter(beta,qcscore=qc)
#filter low quality and outlier values, remove rows 
#and columns with too many missing values
b3=qcfilter(beta,qcscore=qc,rmcr=TRUE)
#filter low quality and outlier values, remove rows and
#columns with too many (rthre=0.05,cthre=0.05, 5% in default) missing values,
# and then do imputation
b3=qcfilter(beta,qcscore=qc,rmcr=TRUE,rthre=0.05,
              cthre=0.05,impute=TRUE)
```

# Background correction and dye-bias correction

Function `preprocessENmix()` incorporates a model-based background
 correction method `ENmix`, which models methylation signal intensities 
with a flexible exponential-normal mixture distribution, together
 with a truncated normal distribution to model background noise. 
Argument "dyeCorr"
can be used to specify a method for dye-bias correction, the default is RELIC. 

Refer the following papers for the detailed description of related methods:

Zongli Xu, et. al. ENmix: a novel background correction method for Illumina
HumanMethylation450 BeadChip, Nucleic Acids Research, 2015

Xu Z, Langie SA, De Boever P, Taylor JA, Niu L.
RELIC: a novel dye-bias correction method for Illumina Methylation BeadChip.
BMC Genomics. 2017

If argument QCinfo is specified, the low quality samples and probes identified 
by function `QCinfo()` will be excluded before ENmix background correction.
Using argument `exSample` and `exCpG`, User can also specify a list 
of samples or probes to be excluded before background correction. 
 
```{r preprocessENmix, eval=FALSE}
qc=QCinfo(rgSet)
mdat<-preprocessENmix(rgSet, bgParaEst="oob", dyeCorr="RELIC", 
QCinfo=qc, exCpG=NULL, nCores=6)
```

# Inter-array normalization

Function `norm.quantile()` can be used to perform quantile 
normalization on methylation intensity values.

```{r normalize.quantile, eval=FALSE}
mdat<-norm.quantile(mdat, method="quantile1")
```

# Probe-type bias adjustment

The majority of probes on Illumina 450K and EPIC BeadChips are type II probes.
 Although type II probes facilitate increased array genome coverage, they were
 shown to have decreased dynamic range and reproducibility compared to type I
 probes. Taking advantage of the high spatial correlation of DNA methylation 
levels along the human genome, The RCP 
(Regression on Correlated Probes) method utilizes nearby (<25 bp) type I and 
II probe pairs to derive the quantitative relationship between probe types and
 then recalibrates type II probe measurements using type I probes as referents. 

```{r rcp, eval=FALSE}
beta<-rcp(mdat)
```

Refer the following publication for the detailed description of the method:

Niu L, Xu Z, Taylor JA.
RCP: a novel probe design bias correction method for Illumina Methylation
BeadChip. Bioinformatics, 2016

# Batch effect correction

Function `ctrlsva()` can be used to estimate surrogate variables for batch 
effects and unknown experimental confounders using intensity data for non-negative 
internal control probes. These surrogate variables can then be modeled as covariables in 
downstream association analysis to adjust for experimental variation.

```{r ctrlsva, eval=FALSE}
sva<-ctrlsva(rgSet)
```

# Principal component regression analysis plot

First, principal component analysis will be performed on the standardized beta value 
matrix (standardized probe by probe), and then the specified number of top 
principal components (that explain most data variation) will be used to perform
linear regression with each specified variables, such as batch or phenotype
variables. Regression P values will be 
plotted to explore methylation data variance structure and to identify possible
 confounding variables to guide association statistical analysis. Principal 
components are ordered according to the percentage of variance they explained 
from large to small. PCA will not allow for missing values, so specify impute=true 
in preprocessing (`qcfilter()`) if there is missing data.

```{r pcrplot, eval=FALSE}
require(minfiData)
mdat <- preprocessRaw(RGsetEx)
beta=getBeta(mdat, "Illumina")
group=pData(mdat)$Sample_Group
slide=factor(pData(mdat)$Slide)
cov=data.frame(group,slide)
pcrplot(beta,cov,npc=6)
```

![Example principal component regression p value plot of raw data
generated using 450K methylation data from a published study (not the data in above code)](./fig/pcr_diag.jpg)


# Multimodal CpGs or gap probes

Function `nmode()` uses an empirical approach to identify
CpGs whose methylation values are multimodal distributed 
(SNP-like or gap probes). When measured in a population of people
 the majority of CpGs on the Illumina HumanMethylation450 BeadChip have
 unimodal distributions of DNA methylation values with relatively small
 between-person variation. However, some CpGs may have multimodal
 distributions of methylation values with sizeable differences between
 modes and large between-person variation.

These multimodal distributed
data are usually caused by SNP effect in the probe region, 
problematic probe design or
other unknown artifacts rather than by actual difference in methylation 
level and thus
should be excluded from DNA methylation analysis. Researchers
 have often excluded CpGs based simply on SNP annotation information
(e.g. reported SNPs in or near probe sequences). However,
because SNP frequency and annotation always depends on population origin, we find
 that this approach alone may exclude many well-distributed (unimodal)
 CpGs, while still failing to identify other problematic multi-modal CpGs. We developed
 an empirical approach to identify CpGs that are obviously not uni-modally
 distributed,
 so that researchers can make more informed decisions about whether to
exclude them in their particular study populations and analyses.

See online supplementary materials of the following paper for  
an evaluation of the method using published EWAS data. 

Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor
ENmix: a novel background correction method for Illumina
HumanMethylation450 BeadChip, Nucleic Acids Research, 2015


```{r nmode, eval=FALSE}
nmode<- nmode(beta, minN = 3, modedist=0.2, nCores = 5)
```

# Cell type proportion estimation

Whole blood samples are often used in EWAS, however, whole blood has 
mixed cell types and methylation level can be widely different between 
different cell types. Function `estimateCellProp()` can be used 
to estimate cell types based on reference dataset. Users should select 
a reference that is most resemble to their own study samples. Currently 
there are 6 publicly deposited reference datasets: 
"FlowSorted.Blood.450k",
     "FlowSorted.DLPFC.450k","FlowSorted.CordBlood.450k",
     "FlowSorted.CordBloodCombined.450k",
     "FlowSorted.CordBloodNorway.450k" or "FlowSorted.Blood.EPIC". 

```{r celltype, eval=FALSE}
require(minfidata)
path <- file.path(find.package("minfiData"),"extdata")
#based on rgDataset
rgSet <- readidat(path = path,recursive = TRUE)
celltype=estimateCellProp(userdata=rgSet,
         refdata="FlowSorted.Blood.450k",
         nonnegative = TRUE,normalize=TRUE)
```

The input data can be in one of the follow format: `rgDataSet`,
`methDataSet`, `MethylSet`, 

`RGChannelSet` 
or methylation beta value matrix.


# Methylation predictors

Methylation age can reflect a person's biological age, which may be more
related to the person's health status than chronological age. Several 
types of methylation age can be estimated using 
`methyAge()`. The more comprehensive function is `methsocre()`, which calculate more than 150 methylation predictors, including all predictors from `methyAge()`

```{r mage, eval=FALSE}
require(minfidata)
path <- file.path(find.package("minfiData"),"extdata")
#based on rgDataset
rgSet <- readidat(path = path,recursive = TRUE)
meth=getmeth(rgSet)
beta=getB(meth)
mage=methyAge(beta)

pheno=data.frame(SampleID=colnames(beta),Age=c(23,45,52,36,58,43),Female=c(1,0,1,1,0,1))
mscore=methscore(datMeth=beta,datPheno=pheno)
```

# Differentially methylated regions (DMRs)

EWAS is usually done probe by probe. However, in the human genome, nearby 
CpGs often have similar methylation status, so combined statistics from nearby 
CpGs can potentially increase association signal. Functions `ipdmr()` and 
`combp()` can utilize P values from EWAS to identify 
differentially methylated regions (DMRs). The argument "seed" in these two functions
indicate the FDR threshold for initial selection of DMR regions, but we
note that same value of seed do not appear to have the same effect in these 
two functions. It is often more stringent in ipdmr than combp. For more information see:

Xu Z, Xie C, Taylor JA, Niu L. ipDMR: identification of differentially methylated regions with interval P-values. Bioinformatics. 2021 May 5;37(5):711-713. doi: 10.1093/bioinformatics/btaa732. PMID: 32805005; PMC8248314.

There is 
no true DMR in the following example data, we set less stringent seed threshold is only 
for demostration purpose to show what the output looks like when DMRs were
detected.

```{r dmr, eval=FALSE}
dat=simubed()
names(dat)
ipdmr(data=dat,seed=0.1)
combp(data=dat,seed=0.1)
```

# Intraclass correlation coefficient (ICC) reliability measures

Several studies have showen that a large percentage of CpGs on Illumina arrays
have poor reliability, i.e. they have low correlation between replicate measures 
in a same set of samples. Measurement of the reliability of individual CpGs 
can be assessed by calculating intraclass
 correlation coefficients (ICC) using methylation from duplicate samples. Function 
`dupicc()` can be used for this purpose. 

```{r icc, eval=FALSE}
require(minfiData)
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mdat=getmeth(rgSet)
beta=getB(mdat,"Illumina")
#assuming list in id1 are corresponding duplicates of id2
dupidx=data.frame(id1=c("5723646052_R02C02","5723646052_R04C01",
                       "5723646052_R05C02"),
                  id2=c("5723646053_R04C02","5723646053_R05C02",
                       "5723646053_R06C02"))
iccresu<-dupicc(dat=beta,dupid=dupidx)
```

# 5-methylcytosine (5mC) and 5-hydroxymethylcytosine (5hmC)

Studies showed that 5-Methylcytosine and 5-hydroxymethylcytosine can have very different
function. However, traditional bisulfite (BS) treatment commonly used in genome-wide 
methylation studies can not distinguish 5mC from 5hmC. Sequencing or array data from 
paired DNA samples (bisulfite and oxidative bisulfite conversion) are needed for the 
separate estimation. Bisulfite converted DNA can be used to estimate the proportion 
of 5mC+5hmC and oxidative bisulfite (oxBS) converted DNA can be used to estimate
 the proportion of 5mC. Function `oxBS.MLE()` can be used to generate the 
maximum Likelihood Estimate (MLE) of 5mC and 5hmC using sequencing or array data from
 paired experiments (Xu et al. Bioinformatics 2016).

```{r oxbs, eval=FALSE}
load(system.file("oxBS.MLE.RData",package="ENmix"))
resu<-oxBS.MLE(beta.BS,beta.oxBS,N.BS,N.oxBS)
dim(resu[["5mC"]])
dim(resu[["5hmC"]])
```

# Compatibility with other related R packages

Most functions in the `r Biocpkg("ENmix")` support the data structure provided 
by `r Biocpkg("minfi")` package.
The same data structures were also used by several other R packages,
 such as `r Biocpkg("ChAMP")` and `r Biocpkg("wateRmelon")`, so the output from ENmix
 functions can be utilized in 
these packages for further analysis. Here are some examples:

Example 1: mixed use of minfi and ENmix functions

```{r ENmixAndminfi, eval=FALSE}
library(ENmix)
#minfi functions to read in data
sheet <- read.metharray.sheet(file.path(find.package("minfiData"),
 "extdata"), pattern = "csv$")
rgSet <- read.metharray.exp(targets = sheet, extended = TRUE)
#ENmix function for control plot
plotCtrl(rgSet)
#minfi functions to extract methylation and annotation data
mraw <- preprocessRaw(rgSet)
beta<-getB(mraw, "Illumina")
anno=getAnnotation(rgSet)
#ENmix function for fast and accurate distribution plot
multifreqpoly(beta,main="Data distribution")
multifreqpoly(beta[anno$Type=="I",],main="Data distribution, type I")
multifreqpoly(beta[anno$Type=="II",],main="Data distribution, type II")
#ENmix background correction
mset<-preprocessENmix(rgSet, bgParaEst="oob", dyeCorr="RELIC", nCores=6)
#minfi functions for further preprocessing and analysis
gmSet <- preprocessQuantile(mset)
bumps <- bumphunter(gmSet, design = model.matrix(~ gmSet$status), B = 0,
type = "Beta", cutoff = 0.25)
```


# References
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor, ENmix: a novel background
 correction method for Illumina HumanMethylation450 BeadChip. Nucleic Acids 
Research, 2015 

Liang Niu, Zongli Xu and Jack A. Taylor, RCP: a novel probe design bias 
correction method for Illumina Methylation BeadChip. Bioinformatics 2016 

Zongli Xu, Jack A. Taylor, Yuet-Kin Leung, Shuk-Mei Ho and Liang Niu,
oxBS-MLE: An efficient method to estimate 5-methylcytosine and 
5-hydroxymethylcytosine in paired bisulfite and oxidative bisulfite 
treated DNA. Bioinformatics 2016

Zongli Xu, Sabine A. S. Langie, Patrick De Boever, Jack A. Taylor1 and 
Liang Niu, RELIC: a novel dye-bias correction method for Illumina 
Methylation BeadChip. BMC Genomics 2017

Zongli Xu, Jack A Taylor. Reliability of DNA methylation measures using
Illumina methylation BeadChip. Epigenetics 2020

Zongli Xu, Changchun Xie, Jack A. Taylor, Liang Niu, ipDMR: Identification
of differentially methyl-ated regions with interval p-values. Bioinfomatics 2020

Illumina Inc., Infinium HD Assay Methylation Protocol Guide,  Illumina, Inc.
 San Diego, CA. 

Aryee MJ, Jaffe AE, Corrada-Bravo H, Ladd-Acosta C, Feinberg AP, Hansen KD 
and Irizarry RA (2014). Minfi: A flexible and comprehensive Bioconductor
 package for the analysis of Infinium DNA Methylation microarrays. 
Bioinformatics, 30(10), pp. 13631369.

Horvath S. DNA methylation age of human tissues and cell types.
Genome biology 2013 14:R115.

Hannum G, Guinney J, Zhao L, Zhang L, Hughes G, Sadda S, et al.
Genome-wide methylation profiles reveal quantitative views
of human aging rates. Molecular cell 2013 49:359-367.

Levine ME, Lu AT, Quach A, Chen BH, Assimes TL, Bandinelli S, et al.
An epigenetic biomarker of aging for lifespan and healthspan.
Aging (Albany NY) 2018 10:573-591.

Pedersen BS1, Schwartz DA, Yang IV, Kechris KJ. Comb-p: software for combining,
analyzing, grouping and correcting spatially correlated P-values. Bioinfomatics 2012

EA Houseman, WP Accomando, DC Koestler, BC Christensen, CJ Marsit, HH
Nelson, JK Wiencke and KT Kelsey.
DNA methylation arrays as surrogate measures of cell mixture distribution.
BMC bioinformatics (2012) 13:86.

Heiss JA, Allan C Just. Improved filtering of DNA methylation microarray 
data by detection p values and its impact on downstream analyses.
Clinical Epigenetics, 2019

Xu Z, Niu L, Taylor JA.
The ENmix DNA methylation analysis pipeline for Illumina BeadChip and 
comparisons to 7 other preprocessing pipelines.
Clinical Epigenetics, 2022

Daniel W Belsky, Avshalom Caspi, David L Corcoran,et al. DunedinPACE, 
a DNA methylation biomarker of the pace of aging, eLIfe, 2022


```{r sessionInfo, echo=FALSE}
sessionInfo()
```

\name{mhtplot}
\alias{mhtplot}
\title{
   P value manhattan plot
}
\description{
P value manhattan plot
}
\usage{
        mhtplot(probe=NULL,chr=NULL, pos=NULL, p=NULL,color="bg",sigthre=NULL,
     sigthre2=NULL,threlty=c(2,1),markprobe=NULL,markcolor="red", outf="mht",
     outfmt="jpg",reducesize=0,...)
}
\arguments{
  \item{probe}{probe name}
  \item{chr}{Chromosome, 1,2,...,22,X,Y}
  \item{pos}{Chromosome positions}
  \item{p}{P values}
  \item{color}{Color scheme of manhattan plot, "bg" indicate "black and gray"}
  \item{sigthre}{P value of significant threshold line}
  \item{sigthre2}{P value of second significant threshold line}
  \item{threlty}{Threshold line type, default is c(2,1)}
  \item{markprobe}{A list of CpGs to be marked out}
  \item{markcolor}{Color code for marked probe, "red" in default}
  \item{outf}{figure file name, default "mht"}
  \item{outfmt}{Output figure file format, "jpg" or "eps"}
  \item{reducesize}{A positive interger, larger the value, smaller the eps file size.
      Smaller file size is achieved by skipping some densely packed data points}
  \item{...}{Arguments pass to plot}
}
\author{
Zongli Xu
}

\examples{
\donttest{

dat=simubed()

thre1=1E-100
dat$fdr=p.adjust(mrgd$P, method="BH")
if(sum(dat$fdr<0.05)>0){thre1=max(dat$p[dat$fdr<0.05])}

thre2=1E-7

mprobe=dat$probe[dat$p<=thre1]
mhtplot(probe=dat$probe,chr=dat$chr,pos=dat$start,p=dat$p,sigthre=thre1,sigthre2=thre2,
        markprobe=mprobe,outf="mht_try",outfmt="jpg")
}
}

\name{oxBS.MLE}
\alias{oxBS.MLE}
\title{
oxBS-MLE.
}
\description{
Maximum Likelihood Estimate (MLE) of 5-methylcytosine (5mC) and
 5-hydroxymethylcytosine (5hmC) using sequencing/array data from
 paired bisulfite and oxidative bisulfite
 treated DNA experiments.
}
\usage{
oxBS.MLE(beta.BS,beta.oxBS,N.BS,N.oxBS)
}
\arguments{
  \item{beta.BS}{A matrix of methylation beta values (proportion of
  methylated sites estimated as methylated intensity over total 
  intensity) obtained from bisulfite (BS) experiments}
  \item{beta.oxBS}{A matrix of methylation beta values obtained from 
  oxidative bisulfite (oxBS) experiments}
  \item{N.BS}{A matrix of total signals (sum of methylated and unmethylated
              intensity values) from BS experiments}
  \item{N.oxBS}{A matrix of total signals from oxBS experiments}
}
\value{
The output is a list with two elements:
\code{5mC}:  a matrix of estimated 5mC levels.
\code{5hmC}: a matrix for estimated 5hmC levels.
}
\details{
For all the inputs (\code{beta.BS}, \code{beta.oxBS}, \code{N.BS} and
 \code{N.oxBS}), the rows should be corresponding to CpG loci and the
 columns should be corresponding to samples.
The row/column names in all four matrices should be the same. 
For a specific CpG of a sample, if any one of the four 
 values (\code{beta.BS}, \code{beta.oxBS}, \code{N.BS} and \code{N.oxBS}) 
is \code{NA}, or \code{N.BS} is zero, or \code{N.oxBS} is zero, the MLE of both 5mC 
and 5hmC levels will be set as \code{NA}.
}
\references{
Zongli Xu, Jack A. Taylor, Yuet-Kin Leung, Shuk-Mei Ho and Liang Niu,
\emph{oxBS-MLE: an efficient method to estimate 5-methylcytosine and 
5-hydroxymethylcytosine in paired bisulfite and oxidative bisulfite 
treated DNA}, Bioinformatics. 2016
}
\author{
Liang Niu and Zongli Xu
}
\examples{
\donttest{

# load example data
load(system.file("oxBS.MLE.RData",package="ENmix"))
# run oxBS.MLE
resu<-oxBS.MLE(beta.BS,beta.oxBS,N.BS,N.oxBS)
dim(resu[["5mC"]])
dim(resu[["5hmC"]])
}
}

\name{methyAge}
\alias{methyAge}
\title{
Methylation Age estimator
}
\description{
To calculate Methylation Age using Hovath, Hannum or PhenoAge methods and 
pace of aging DunedinPACE.
}
\usage{
methyAge(beta,fastImputation=FALSE,normalize=TRUE,nCores=2)
}
\arguments{
  \item{beta}{Methylation beta value matrix with CpG names(row names) and
       sample ids(column names).}
  \item{fastImputation}{If "TRUE" reference methylation values will used
      for imputation, if "FALSE", KNN nearest neighbor method will be used.}
  \item{normalize}{TRUE or FALSE, if TRUE, Hovath modified BMIQ method will
       be used to perform normalization.}
  \item{nCores}{Number of cores will be used for normalization}
}
\value{
  A data frame with rows for sample and columns for types of methylation age.
}

\references{
Horvath S. \emph{DNA methylation age of human tissues and cell types}. Genome biology 2013 14:R115.

Hannum G, Guinney J, Zhao L, Zhang L, Hughes G, Sadda S, et al. \emph{Genome-wide methylation profiles reveal quantitative views of human aging rates}. Molecular cell 2013 49:359-367.

Levine ME, Lu AT, Quach A, Chen BH, Assimes TL, Bandinelli S, et al.  \emph{An epigenetic biomarker of aging for lifespan and healthspan}. Aging (Albany NY) 2018 10:573-591.

Daniel W Belsky, Avshalom Caspi, David L Corcoran,et al. \emph{DunedinPACE, a DNA methylation biomarker of the pace of aging}. eLife, 2022
}

\author{
Zongli Xu
}

\examples{
\donttest{

require(minfiData)
path <- file.path(find.package("minfiData"),"extdata")
#based on rgDataset
rgSet <- readidat(path = path,recursive = TRUE)
meth=getmeth(rgSet)
beta=getB(meth)
mage=methyAge(beta)
}
}

\docType{class}
\name{methDataSet-class}
\alias{methDataSet-class}
\alias{methDataSet}
%methods
\alias{show, methDataSet-method}

\title{Class \code{"methDataSet"}}
\description{A class for storing Illumina methylation array 
methylated and unmethylated intensity data, and CpG annotation 
infomation.
}
\usage{
methDataSet(Meth = new("matrix"), Unmeth = new("matrix"),
    rowData=new("DataFrame"),...)
}
\arguments{
\item{Meth}{A matrix of methylated intensity values with row for CpGs
     and column for samples}
\item{Unmeth}{A matrix of unmethylated intensity values with row for CpGs
     and column for samples}
\item{rowData}{A dataframe contains CpG annotation information}
\item{...}{Other arguments for class SummarizedExperiment}
}

\value{
  An object of class \code{methDataSet}
}
\details{
CpG annotation information is incorporated in the object, and can 
be extracted using command \code{rowData}
}
\examples{
showClass("methDataSet")
}

\name{estimateCellProp}
\alias{estimateCellProp}
\title{
Cell type proportion estimator
}
\description{
To estimates relative proportion of underlying cell types in a sample based
 on reference methylation data of pure cell types.  
}
\usage{
estimateCellProp(userdata,refdata="FlowSorted.Blood.450k",
                cellTypes=NULL,nonnegative = TRUE,nProbes=50,
                normalize=TRUE,refplot=FALSE)
}
\arguments{
  \item{userdata}{The input can be \code{rgDataSet},\code{methDataSet},
        \code{MethylSet},\code{RGChannelSet} or methylation beta value matrix.}
  \item{refdata}{Reference data set will used. Current option: "FlowSorted.Blood.450k", 
     "FlowSorted.DLPFC.450k","FlowSorted.CordBlood.450k",
     "FlowSorted.CordBloodCombined.450k",
     "FlowSorted.CordBloodNorway.450k" or "FlowSorted.Blood.EPIC".}
  \item{cellTypes}{Specific set of cell type data in reference data will be used for 
              deconvolution. if "NULL" all cell types data will be used. see details 
              for possible cell types}
  \item{normalize}{TRUE or FALSE, if TRUE, quantile normalization on methylated
                   and unmethylated intensities will be performed.}
  \item{nonnegative}{TRUE or FALSE. If TRUE, the estimated proportions will be
                    constrained to nonnegative values}
  \item{nProbes}{Number of best probes for each cell types will be used for the
                 estimation.}
  \item{refplot}{TRUE or FALSE. IF TRUE, refdata distribution and heatmap will
                 be plotted for inspection of reference dataset. }
}
\details{

This function use the method of Houseman et al (2012) to estimate cell type
  proportions based on reference DNA methylation data. 

The following reference datasets can be used to assist the estimation. User 
should select a reference most resemble to user's data in tissue, age, and 
 array type. 

FlowSorted.Blood.450k: consisting of 450K methylation data for 60 blood samples
 from 6 male adults. Six samples for each of the cell types: Bcell CD4T CD8T Eos 
 Gran  Mono   Neu    NK  PBMC   WBC;  See Reinius et al. 2012 for details.

FlowSorted.CordBlood.450k: consisting of 450k methylation data for 104 cord blood
 samples from 17 male and female individuals. Cell type (# samples) are:
 Bcell(15) CD4T(15)  CD8T(14) Gran(12) Mono(15) NK(14) nRBC(4) WholeBlood(15).
 See Bakulski et al. Epigenetics 2016 for details.

FlowSorted.CordBloodNorway.450k: consisting of 450K methylation data for 77 cord
 blood samples from 11 individuals (6 girls and 5 boys). 11 samples for each
 of the cell types: Bcell  CD4T  CD8T  Gran  Mono    NK   WBC. See P Yousefi
 et al Environ. Mol. Mutagen 2015 for details.

FlowSorted.Blood.EPIC: consisting of EPIC methylation data for 37 magnetically sorted
 blood cell references from 12 individuals. See LA Salas et al. 2018 for details. 

FlowSorted.DLPFC.450k: consisting of 450K methylation data for 58 brain tissue
 samples from 29 individuals. 15 females and 14 males, 6 Africans and 23
 Caucasians, age range from 13 to 79. 29 samples for each of the cell types:
 NeuN_neg and NeuN_pos. See Guintivano et al. 2013 for details.

FlowSorted.CordBloodCombined.450k: consisting of 289 combined umbilical cord blood 
 cells samples assayed by Bakulski et al, Gervin et al., de Goede et al., and 
 Lin et al. see https://github.com/immunomethylomics/
 FlowSorted.CordBloodCombined.450k. for details.
}
\value{
  The output is a data frame composed of the estimates of cell type proportions with 
  columns indicate cell types and rows indicate samples.
}

\references{
  EA Houseman, WP Accomando, DC Koestler, BC Christensen, CJ Marsit, HH
  Nelson, JK Wiencke and KT Kelsey.
  \emph{DNA methylation arrays as surrogate measures of cell mixture distribution.}
  BMC bioinformatics (2012) 13:86.
}

\author{
Zongli Xu
}

\examples{
\donttest{
require(minfiData)
path <- file.path(find.package("minfiData"),"extdata")
#based on rgDataset
rgSet <- readidat(path = path,recursive = TRUE)
celltype=estimateCellProp(userdata=rgSet,refdata="FlowSorted.Blood.450k",
         nonnegative = TRUE,normalize=TRUE)
#using methDataSet
qc=QCinfo(rgSet)
mdat<-preprocessENmix(rgSet, bgParaEst="oob", dyeCorr="RELIC",
                      QCinfo=qc,  nCores=6)
celltype=estimateCellProp(userdata=mdat,refdata="FlowSorted.Blood.450k",
        nonnegative = TRUE,normalize=TRUE)
mdat<-norm.quantile(mdat, method="quantile1")
#using beta value
beta<-rcp(mdat,qcscore=qc)
celltype=estimateCellProp(userdata=beta,refdata="FlowSorted.Blood.450k",
         nonnegative = TRUE)
}
}

\name{M2B}
\alias{M2B}
\title{
Converting methylation M value to beta value.
}
\description{
Converting methylation M value to methylation beta value.
}
\usage{
M2B(x)
}
\arguments{
  \item{x}{An numeric matrix}
}
\details{
Methylation beta value is calculated as beta=M/(M+U+a). M is methylated intensity,
U is unmethylated intensity, and a is a constant offset (by default , a=100).
M value is calculated as M=log2((M+a)/(U+a)). M or U is usually greater than 1000,
so a is negligible for most probes. if a=0, then beta=2^M/(2^M+1).
}

\value{
A matrix of methylation Beta values.
}
\author{
Zongli Xu
}

\examples{
\donttest{

if (require(minfiData)){
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mdat=getmeth(rgSet)
beta=getB(mdat,"Illumina")
m=B2M(beta)
b=M2B(m)
}}
}

\name{ctrlsva}
\alias{ctrlsva}
\title{
Non-negative internal control surrogate variables
}
\description{
Surrogate variables derived based on intensity data for non-negative internal control 
probes.
}
\usage{
ctrlsva(rgSet,percvar=0.95,npc=1,flag=1)
}
\arguments{
  \item{rgSet}{An object of class \code{rgDataSet} or \code{RGChannelSet}.}
  \item{percvar}{Minimum percentage of data variations can be explained by 
surrogate variables, range from 0 to 1,default is 0.95}
  \item{npc}{Number of surrogate variables, default is 1}
  \item{flag}{1: select number of surrogate variables based on argument
 \code{percvar}; 2: select number of surrogate variables based on argument \code{npc}}
}

\value{
The function will return an numerical matrix with columns indicating surrogate variables and rows
 corresponding to samples.  These variables can be used in association analysis to adjust for
experimental batch effects.
}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor,
\emph{ENmix: a novel background correction method for Illumina HumanMethylation450
 BeadChip}. Nucleic Acids Research 2015.
}
\author{
Zongli Xu
}

\examples{
\donttest{
if (require(minfiData)) {
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
sva<-ctrlsva(rgSet)
}}
}

\name{methscore}
\alias{methscore}
\title{
DNA Methylation predictors
}
\description{
To calculate various methylation predictors, including DNA methylation age, plasma protein levels, exposures etc.
}
\usage{
methscore(datMeth,datPheno=NULL,fastImputation=FALSE,normalize=TRUE,
	  GrimAgeComponent=3,AgeResid=FALSE)
}
\arguments{
  \item{datMeth}{Methylation beta value matrix with CpG names(row names) and sample ids(column names).}
  \item{datPheno}{Phenotype data, must include columns SampleID, Age (year) and Female (0:male,1:female).
       Age and Female information are required to calculate DNAmFitAge and PCClocks}
  \item{fastImputation}{Logical. If "TRUE" mean methylation values will used
      for imputation, if "FALSE", KNN nearest neighbor method will be used.}
  \item{normalize}{Logical. If TRUE, user data will be normalized to a reference data using a modified RCP method}
  \item{GrimAgeComponent}{Which version of GrimAgeComponent will be used to calculate bAge and DNAmFitAge.
	  1: GrimAgeV1; 2: GrimAgeV2 and 3: PCGrimAge.} 
  \item{AgeResid}{Logical. If TRUE, age residuals (age acceleration) variables will be calculated.}
}
\details{
        The original publications (see references) provided reference methylation
       	mean values for most methylation predictors, and thus these values were
       	used directly for normalization and imputations. For the following 
	predictors without reference value in their original publications,

        PEDBE,EpiToc,EpiToc2,Zhang10CpG,Horvath2,MiAge,DNAmTL,PEDBE,GACPC,
	GARPC,GARRPC,Bohlin and Knight, we used Sister Study data
       	(~5000 samples) to derive a set of reference methylatin values.
       	If UserRef was provided, it will be used to replace the Sister
       	Study reference values for these predictors. If set
       	ForceUserRef=TRUE, UserRef will be
        used for all predictors in normalization and CpG imputation.
}
\value{
  A data frame with rows for samples and columns for methylation predictors. 
  Output file "summary_CpG_count.csv" has information about CpG counts and reference for each predictor.
}

\references{
Belsky et al. \emph{DunedinPACE, a DNA methylation biomarker of the pace of aging}. Elife. 2022 Jan 14;11:e73420. doi: 10.7554/eLife.73420

Bernabeu et al. \emph{Refining epigenetic prediction of chronological and biological age}. Genome Med. 2023 Feb 28;15(1):12. doi: 10.1186/s13073-023-01161-y

Bohlin et al. \emph{Prediction of gestational age based on genome-wide differentially methylated regions}.Genome Biol. 2016 Oct 7;17(1):207. doi: 10.1186/s13059-016-1063-4.

Hannum et al. \emph{Genome-wide methylation profiles reveal quantitative views of human aging rates}.Mol Cell. 2013 Jan 24;49(2):359-367. doi: 10.1016/j.molcel.2012.10.016. Epub 2012 Nov 21.

Higgins-Chen et al. \emph{A computational solution for bolstering reliability of epigenetic clocks: Implications for clinical trials and longitudinal tracking}.	Nat Aging. 2022 Jul; 2(7): 644–661.

Horvath. \emph{DNA methylation age of human tissues and cell types}. Genome Biol. 2013;14(10):R115. doi: 10.1186/gb-2013-14-10-r115.

Horvath et al. \emph{Epigenetic clock for skin and blood cells applied to Hutchinson Gilford Progeria Syndrome and ex vivo studies}.Aging (Albany NY). 2018 Jul 26;10(7):1758-1775. doi: 10.18632/aging.101508.

Knight et al. \emph{An epigenetic clock for gestational age at birth based on blood methylation data}. Genome Biol. 2016 Oct 7;17(1):206. doi: 10.1186/s13059-016-1068-z.

Lee et al. \emph{Placental epigenetic clocks: estimating gestational age using placental DNA methylation levels}.Aging (Albany NY). 2019 Jun 30; 11(12): 4238–4253.

Levine et al. \emph{An epigenetic biomarker of aging for lifespan and healthspan}. Aging (Albany NY). 2018 Apr; 10(4): 573–591.

Li et al. \emph{Derivation and validation of an epigenetic frailty risk score in population-based cohorts of older adults}. Nat Commun. 2022 Sep 7;13(1):5269. doi: 10.1038/s41467-022-32893-x.

Lu et al. \emph{DNA methylation-based estimator of telomere length}.Aging (Albany NY). 2019 Aug 31; 11(16): 5895–5923. 

McCartney et al. \emph{Epigenetic prediction of complex traits and death}.Genome Biol. 2018 Sep 27;19(1):136. doi: 10.1186/s13059-018-1514-1.

McGreevy et al. \emph{DNAmFitAge: biological age indicator incorporating physical fitness}. Aging(Albany NY). 2023 Feb 22;15(10):3904-3938. doi: 10.18632/aging.204538.

McEwen et al. \emph{The PedBE clock accurately estimates DNA methylation age in pediatric buccal cells}.Proc Natl Acad Sci U S A. 2020 Sep 22; 117(38): 23329–23335.

Teschendorff et al. \emph{A comparison of epigenetic mitotic-like clocks for cancer risk prediction}.Genome Med. 2020 Jun 24;12(1):56. doi: 10.1186/s13073-020-00752-3.

Yang et al. \emph{Correlation of an epigenetic mitotic clock with cancer risk}. Genome Biol. 2016 Oct 3;17(1):205. doi: 10.1186/s13059-016-1064-3.

Youn et al. \emph{The MiAge Calculator: a DNA methylation-based mitotic age calculator of human tissue types}.Epigenetics. 2018; 13(2): 192–206.

Zhang et al. \emph{Improved precision of epigenetic clock estimates across tissues and its implication for biological ageing}.Genome Med. 2019 Aug 23;11(1):54. doi: 10.1186/s13073-019-0667-1.

Zhang et al. \emph{DNA methylation signatures in peripheral blood strongly predict all-cause mortality}.Nat Commun. 2017 Mar 17;8:14617. doi: 10.1038/ncomms14617.

Lu et al. \emph{DNA methylation GrimAge strongly predicts lifespan and healthspan}. Aging. 2019 Jan 21;11(2):303-327.  doi: 10.18632/aging.101684.

Lu et al. \emph{DNA methylation GrimAge version 2}. Aging (Albany NY). 2022 Dec 14;14(23):9484-9549. doi: 10.18632/aging.204434. Epub 2022 Dec 14.

Ying et al. \emph{Causality-enriched epigenetic age uncouples damage and adaptation}. Nat Aging. 2024 Feb;4(2):231-246. doi: 10.1038/s43587-023-00557-0. Epub 2024 Jan 19.
}

\author{
Zongli Xu
}

\examples{
\donttest{
require(minfiData)
path <- file.path(find.package("minfiData"),"extdata")
#based on rgDataset
rgSet <- readidat(path = path,recursive = TRUE)
meth=getmeth(rgSet)
beta=getB(meth)
pheno=data.frame(SampleID=colnames(beta),Age=c(23,45,52,36,58,43),Female=c(1,0,1,1,0,1))
mage=methscore(datMeth=beta,datPheno=pheno)
}
}

\name{relic}
\alias{relic}
\title{REgression on Logarithm of Internal Control probes (RELIC)}
\description{REgression on Logarithm of Internal Control probes (RELIC) 
correct for dye bias on whole array by utilizing the intensity values
 of paired internal control probes that monitor the two color channels. 
}
\usage{
relic (mdat,at_red,cg_grn)
}
\arguments{
  \item{mdat}{An object of class \code{methDataSet} or \code{MethylSet}.}
  \item{at_red}{an intensity matrix for Illumina control probes "NORM_A" 
and "NORM_T"}
  \item{cg_grn}{an intensity matrix for Illumina control probes "NORM_C" 
and "NORM_G"}
}
\details{
The Illumina MethylationEPIC BeadChip contains 85 pairs of internal 
normalization control probes (name with prefix NORM_A, NORM_T, NORM_G
 or NORM_C), while its predecessor, Illumina HumanMethyl-ation450
 BeadChip contains 93 pairs. RELIC first performs a regression on
 the logarithms of the intensity values of the normalization control
 probes to derive a quantitative relationship between red and green 
channels, and then uses the relationship to correct for dye-bias on
 intensity values for whole array. 
}
\value{
An object of class  \code{methDataSet} or \code{MethylSet} depends on
 input class.
}
\references{
Zongli Xu, Sabine A. S. Langie, Patrick De Boever, Jack A. Taylor and 
Liang Niu, RELIC: a novel dye-bias correction method for Illumina 
Methylation BeadChip, BMC Genomics. 2017}
\author{Zongli Xu and Liang Niu}
\seealso{
  Package \code{preprocessENmix}
}
\examples{
\donttest{
if (require(minfiData)) {

##background correction and dye bias correction
#rgDataSet as input
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mdat <- preprocessENmix(rgSet,bgParaEst="oob",nCores=6,dyeCorr ="RELIC")

#RGChannelSet as input
mdat=preprocessENmix(RGsetEx,bgParaEst="oob",nCores=6,dyeCorr ="RELIC")

##dye bias correction only
#methDataSet as input
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
ctrls <- getCGinfo(rgSet,type="ctrl")
ctrls <- ctrls[ctrls$Address \%in\% rownames(rgSet),]

ctrl_r <- assays(rgSet)$Red[ctrls$Address,]
ctrl_g <- assays(rgSet)$Green[ctrls$Address,]
CG.controls <- ctrls$Type \%in\% c("NORM_C", "NORM_G")
AT.controls <- ctrls$Type \%in\% c("NORM_A", "NORM_T")
cg_grn=ctrl_g[CG.controls,]
at_red=ctrl_r[AT.controls,]
rownames(cg_grn) = ctrls$ExtendedType[CG.controls]
rownames(at_red) = ctrls$ExtendedType[AT.controls]
mdat=getmeth(rgSet)
mdat <- relic(mdat,at_red,cg_grn)

#MethylSet as input
ctrls <- getProbeInfo(RGsetEx,type="Control")
ctrls <- ctrls[ctrls$Address \%in\% featureNames(RGsetEx),]
ctrl_r <- getRed(RGsetEx)[ctrls$Address,]
ctrl_g <- getGreen(RGsetEx)[ctrls$Address,]
CG.controls <- ctrls$Type \%in\% c("NORM_C","NORM_G")
AT.controls <- ctrls$Type \%in\% c("NORM_A","NORM_T")
cg_grn <- ctrl_g[CG.controls,]
at_red <- ctrl_r[AT.controls,]
rownames(cg_grn) = ctrls$ExtendedType[CG.controls]
rownames(at_red) = ctrls$ExtendedType[AT.controls]
mdat <- preprocessRaw(RGsetEx)
mdat <- relic(mdat,at_red,cg_grn)

}}
}

\name{combp}
\alias{combp}
\title{
Identification of differentially methylated regions
}
\description{
To identify differentially methylated regions using a 
modified comb-p method
}
\usage{
       combp(data,dist.cutoff=1000,bin.size=310,seed=0.01,
             region_plot=TRUE,mht_plot=TRUE,nCores=10,verbose=TRUE)
}
\arguments{
  \item{data}{A data frame with colname name "chr","start",
   "end","p" and "probe", indicating chromosome (1,2,3,...,X,Y),  
   chromosome start and end position, P value and probe names}
  \item{dist.cutoff}{Maximum distance in base pair to combine adjacent DMRs}
  \item{bin.size}{bin size for autocorrelation calculation}
  \item{seed}{FDR significance threshold for initial selection of DMR region}
  \item{region_plot}{If TRUE, regional plots will be generated}
  \item{mht_plot}{If TRUE, mahattan plot will be generated}
  \item{nCores}{Number of computer cores will be used in calculation}
  \item{verbose}{If TRUE, detailed running information will be printed}
}
\details{
The input should be a data frame with column names "chr","start", "end","p",
and "probe", indicating chromosome number, start position, end position, 
P value and probe name. The function use 
a modified comb-p method to identify differentially methylated regions. DMR 
results will be stored in a file with name resu_combp.csv. If plot options 
were selected, two figure files will be generated: mht.jpg and region_plot.pdf.
}
\references{
Pedersen BS1, Schwartz DA, Yang IV, Kechris KJ. Comb-p: software for combining, 
analyzing, grouping and correcting spatially correlated P-values. Bioinfomatics 2012

Zongli Xu, Changchun Xie, Jack A. Taylor, Liang Niu, ipDMR: Identification 
of differentially methyl-ated regions with interval p-values, Bioinformatics 2020

}
\author{
Liang Niu, Zongli Xu
}

\examples{
\donttest{
dat=simubed()
names(dat)
#seed=0.1 is only for demonstration purpose, it should be smaller than 0.05 or 0.01 in actual study.
combp(data=dat,seed=0.1)
}
}

\name{qcfilter}
\alias{qcfilter}
\title{
Filtering out low quality and outlier data 
}
\description{
Outlier was defined as values smaller than 3
 times IQR from the lower quartile or greater than 3 times IQR from the upper
 quartile. If data quality information were provided, low quality data points
 will be set as missing data first before looking for outliers. All outliers
and low quality data will be set as miss in output matrix. If set imput=TRUE, 
imputation will be performed using k-nearest neighbors method to impute all 
missing values. 
}
\usage{
qcfilter(mat,qcscore=NULL,rmoutlier=TRUE,byrow=TRUE,detPthre=0.000001,nbthre=3,
     rmcr=FALSE,rthre=0.05,cthre=0.05,impute=FALSE,imputebyrow=TRUE,fastimpute=FALSE,...)
}
\arguments{
  \item{mat}{An numeric matirx containing methylation beta values}
  \item{qcscore}{If the data quality infomation (the output from function QCinfo)
 were provied, low quality data points as defined by detection p value threshold 
(detPthre) and number of bead threshold (nbthre) will be set as missing value.} 
  \item{rmoutlier}{if TRUE, outliers data points will be set as missing data NA.}
  \item{byrow}{TRUE: Looking for outliers row by row, or FALSE: column by column.}
  \item{detPthre}{Detection P value threshold to define low qualitye data points, 
detPthre=0.000001 in default.}
  \item{nbthre}{Number of beads threshold to define low qualitye data points, nbthre=3
 in default.}
  \item{rmcr}{TRUE: exclude rows and columns with too many missing values as 
defined by rthre and cthre. FALSE is in default}
  \item{rthre}{Minimum of percentage of missing values for a row to be excluded}
  \item{cthre}{Minimum of percentage of missing values for a column to be excluded}
  \item{impute}{If TRUE, k-nearest neighbors methods will used for imputation.}
  \item{imputebyrow}{TRUE: impute missing values using similar values in row, 
or FALSE: in column}
  \item{fastimpute}{If TRUE, probe median will be used for fast imputation.}
  \item{...}{Arguments to be passed to the function impute.knn in R package "impute"}
}
\value{
The output is an numeric matrix.
}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor,
\emph{ENmix: a novel background correction method for Illumina HumanMethylation450
 BeadChip}. Nucleic Acids Research 2015.
}
\author{
Zongli Xu
}

\examples{
\donttest{

if (require(minfiData)) {
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
qc=QCinfo(rgSet)
mdat=preprocessENmix(rgSet,QCinfo=qc,nCores=6)
mdat=norm.quantile(mdat,method="quantile1")
beta=rcp(mdat)
#filter out outliers data points only
b1=qcfilter(beta)
#filter out low quality and outlier data points
b2=qcfilter(beta,qcscore=qc)
#filter out low quality and outlier values, remove rows and columns with
# too many missing values
b3=qcfilter(beta,qcscore=qc,rmcr=TRUE)
#filter out low quality and outlier values, remove rows and columns with
# too many missing values, and then do imputation
b3=qcfilter(beta,qcscore=qc,rmcr=TRUE,impute=TRUE)
}}}

\name{repicc}
\alias{repicc}
\title{
Calculating intraclass correlation coefficient using replicate samples
}
\description{
The function can be used to calculate ICC for each CpG probe using 
balanced or unbalanced replicate samples. 
}
\usage{
repicc(dat,repid,mvalue=FALSE,nCores=2,qcflag=FALSE,qc=NULL,
       detPthre=0.05,nbthre=3)
}
\arguments{
  \item{dat}{Methylation beta value matrix}
  \item{repid}{A data frame with two variables, id and idx. id should be 
              the same with column name of "dat", idx is a variable
              to show the relationship between samples with same value for   
              samples from same individual.}
  \item{mvalue}{If TRUE, the beta value will be converted to M value for 
             calculation of ICC}
  \item{nCores}{Number of cores will be used for calculation of ICC}
  \item{qcflag}{Whether to perform QC before calculation of ICC}
  \item{qc}{QC object from function QCinfo}
  \item{detPthre}{If qcflag=TRUE, the methylation values with detection P value
       higher than the threshold will be removed before calculation}
  \item{nbthre}{If qcflag=TRUE, the methylation values with number of bead smaller}
}
\value{
   A data frame containing ICC for each probe
}
\references{
Zongli Xu, Jack A Taylor. \emph{Reliability of DNA methylation measures using 
Illumina methylation BeadChip}. Epigenetics 2020
}
\author{
Zongli Xu
}

\examples{
\donttest{
if (require(minfiData)){
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mdat=getmeth(rgSet)
beta=getB(mdat,"Illumina")
repid=data.frame(id=c("5723646052_R02C02","5723646052_R04C01","5723646052_R05C02",
"5723646053_R04C02","5723646053_R05C02","5723646053_R06C02"),idx=c(1,1,2,2,2,2))
iccresu<-repicc(dat=beta,repid=repid)
}}
}

\name{simubed}
\alias{simubed}
\title{
Simulation of bed format example file.
}
\description{
Simulation of bed format example file.
}
\usage{
simubed(nprobe=1000)
}
\arguments{
  \item{nprobe}{Number of probes on each chromosome, default is 1000}
}
\details{
Simulation of bed format example file. 
}
\value{
A data frame
}
\author{
Zongli Xu
}
\examples{
\donttest{

simubed(nprobe=1000)
}}

\name{calcdetP}
\alias{calcdetP}
\title{To calculate detection P values}
\description{
Calculation of detection P values based on negtive internal control probes or out of the
 band (oob) probes
}
\usage{
calcdetP(rgSet,detPtype = "negative")
}
\arguments{
  \item{rgSet}{An object of class \code{rgDataSet}}
  \item{detPtype}{Calculation of detection P values based on negtive internal control
 ("negative") probes or out of the band ("oob") probes}
}
\value{
An numerical matrix of detection P values, with row for CpGs and column for samples
}
\references{
Wanding Zhou et al.
SeSAMe: reducing artifactual detection of DNA  methylation by Infinium BeadChips
 in genomic deletions, Nucleic Acids Research, 2018
}

\author{
Zongli Xu
}
\examples{
\donttest{
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
detp=calcdetP(rgSet,detPtype = "negative")
detp2=calcdetP(rgSet,detPtype = "oob")
}
}

\name{norm.quantile}
\alias{norm.quantile}
\title{
Quantile normalization.
}
\description{
Quantile normalization of methylation intensity data across samples for Illumina 
Infinium HumanMethylation 450 and MethylationEPIC BeadChip.
}
\usage{
norm.quantile(mdat, method = "quantile1")
}
\arguments{
  \item{mdat}{An object of class \code{methDataSet} or \code{MethylSet}.}
  \item{method}{Quantile normalization method: "quantile1", "quantile2", or "quantile3".}
}
\details{
By default,  method = "quantile1", which will separately quantile normalize Methylated or
 Unmethylated intensities for Infinium I or II probes. The "quantile2" will 
quantile normalize combined Methylated or Unmethylated intensities for
 Infinium I or II probes. The "quantile3" will quantile normalize combined
 Methylated or Unmethylated intensities for Infinium I and II probes together.
}
\value{
The output is an an object of class \code{methDataSet} or \code{MethylSet}.
}
\references{
Pidsley, R., CC, Y.W., Volta, M., Lunnon, K., Mill, J. and Schalkwyk, L.C.
 (2013) A data-driven approach to preprocessing Illumina 450K methylation
 array data. BMC genomics, 14, 293.
}

\author{
Zongli Xu
}

\examples{
\donttest{

#for methDataSet
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mdat<-preprocessENmix(rgSet, bgParaEst="oob",nCores=6)
mdatq<-norm.quantile(mdat, method="quantile1")

#for MethySet
if (require(minfiData)) {
mdat=preprocessENmix(RGsetEx,bgParaEst="oob",nCores=6)
mdatq=norm.quantile(mdat,method="quantile1")
}
}
}

\name{pcrplot}
\alias{pcrplot}
\title{
Principal component regression plot
}
\description{
First, principal component analysis will be performed in the standadized input data
matrix (standadized for each row/CpG), and then the specified number of top
principal components (which explain most data variation) will be used to perform
linear regression with each specified variable. Regression P values will be
plotted for exploration of methylation data variance structure or identification 
of possible confounding variables in association analysis.
}
\usage{
pcrplot(beta, cov,npc=50)
}
\arguments{
  \item{beta}{A methylation beta value matrix with rows for probes and columns for
 samples. The input matrix should not contain any missing value.}
  \item{cov}{A data frame of covariates. Categorical variables should be converted
 to factors. The number of rows should equal to the number of columns in beta matrix}
  \item{npc}{The number of top ranked principal components to be plotted}
}
\value{
A jpeg figure "svdscreeplot.jpg" to show the variations explained by each principal
 component.

A jpeg figure "pcr_diag.jpg" to show association strength between principal
 components and covariates with cell colors indicating different levels of
 association P values.
}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor,
\emph{ENmix: a novel background correction method for Illumina HumanMethylation450 
BeadChip}. Nucleic Acids Research 2015
}
\author{
Zongli Xu
}
\examples{
\donttest{

if (require(minfiData)) {
mdat <- preprocessRaw(RGsetEx)
beta=getBeta(mdat, "Illumina")
group=pData(mdat)$Sample_Group
slide=factor(pData(mdat)$Slide)
cov=data.frame(group,slide)
pcrplot(beta,cov,npc=6)
}}
}

\name{readidat}
\alias{readidat}
\title{
Parsing IDAT files for Illumina methylation arrays .
}
\description{
Read in IDAT files and create a rgDataSet with probe annotation
}
\usage{
readidat(path = NULL,manifestfile=NULL,recursive = TRUE, verbose = FALSE, force=FALSE)
}
\arguments{
  \item{path}{Directory where idat files are located}
  \item{manifestfile}{Array manifestfile, which can be downloaded 
    from Illumina website. Bioconductor manifest package will be used if not provided}
  \item{recursive}{if TRUE, idat files in the subdirectories will also be read in}
  \item{verbose}{if TRUE, detailed running info will be printed on screen}
  \item{force}{if TRUE, arrays with different sizes will be merged together}
}
\value{
An object of class \code{rgDataSet}, 
}
\details{
Some array types and corresponding manifestfiles can be guessed by the program based on 
the number of probes per array. However, we recommand to provide correct manifest file
directly, which can be easily downloaded from Illumina website, see below for some examples.

Probe annotation info can be extracted using command \code{rowData}
}
\author{
Zongli Xu
}
\examples{
\donttest{

#Illumina methylation array manifestfile

#Infinium Mouse Methylation Manifest File (CSV)
system("wget https://support.illumina.com/content/dam/illumina-support/documents/
       downloads/productfiles/mouse-methylation/infinium-mouse-methylation-
       manifest-file-csv.zip")
system("unzip infinium-mouse-methylation-manifest-file-csv.zip")
mf="infinium-mouse-methylation-manifest-file.csv"

#for MethylationEPIC v2.0
#download InfiniumMethylationEPICv2.0ProductFiles 
#mf= path to file "EPIC-8v2-0_A1.csv" or file "EPIC-8v2-0_A2.csv"
#manifest=readmanifest(mf)

#for MethylationEPIC v1.0 B5
system("wget https://webdata.illumina.com/downloads/productfiles/
       methylationEPIC/infinium-methylationepic-v-1-0-b5-
       manifest-file-csv.zip")
system("unzip infinium-methylationepic-v-1-0-b5-manifest-file-csv.zip")
mf="infinium-methylationepic-v-1-0-b5-manifest-file.csv"
manifest=readmanifest(mf)

#for MethylationEPIC v1.0 B4
system("wget https://webdata.illumina.com/downloads/productfiles/
       methylationEPIC/infinium-methylationepic-v-1-0-b4-
       manifest-file-csv.zip")
system("unzip infinium-methylationepic-v-1-0-b4-manifest-file-csv.zip")
mf="MethylationEPIC_v-1-0_B4.csv"
manifest=readmanifest(mf)

#for HumanMethylation450
system("wget https://webdata.illumina.com/downloads/productfiles/
       humanmethylation450/humanmethylation450_15017482_v1-2.csv")
mf="HumanMethylation450_15017482_v1-2.csv"

mf="HumanMethylation450_15017482_v1-2.csv"
if(require(minfiData)){
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,manifestfile=mf,recursive = TRUE)

#without providing manifest file, corresponding R manifestfile package will be used
rgSet <- readidat(path = path,recursive = TRUE)
}}
}

\name{B2M}
\alias{B2M}
\title{
Converting methylation beta value to M value.
}
\description{
Convert methylation beta value to M value.
}
\usage{
B2M(x)
}
\arguments{
  \item{x}{An numeric matrix with values between 0 and 1}
}
\details{
Methylation beta value is calculated as beta=M/(M+U+a). M is methylated intensity,
U is unmethylated intensity, and a is a constant offset (by default , a=100).
M value is calculated as M=log2((M+a)/(U+a)). M or U is usually greater than 1000,
so a is negligible for most probes. if a=0, then M=log2(beta)/(1-beta).
}
\value{
A matrix of M values
}
\author{
Zongli Xu
}

\examples{
\donttest{
if (require(minfiData)){
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mdat=getmeth(rgSet)
beta=getB(mdat,"Illumina")
m=B2M(beta)
}}
}

\name{nmode}
\alias{nmode}
\title{
Estimating number of mode for each row of data
}
\description{
Due to SNPs in CpG probe region or other unknow factors, methylation beta 
values for some CpGs have multimodal distribution. This function is to
 identify this type of probes (so called gap probes) with obovious 
multimoal distribution. 
}
\usage{
nmode(x, minN = 3, modedist=0.2, nCores = 1)
}
\arguments{
  \item{x}{A methylation beta value matrix with row for probes and column 
for samples.}
  \item{minN}{Minimum number of data points at each cluster}
  \item{modedist}{Minimum distance between adjacent modes}
  \item{nCores}{Number of cores used for computation}
}
\details{
This function uses an empirical approach to estimate number of modes in 
methylation beta value for each CpG probe. By default, the function 
requires the distance between modes have to be greater than 0.2 in 
methylation beta value, and each mode clusters should has at least 3 
data points or 5\% of data points whichever is greater. 
}
\value{
A vector of integers indicating number of modes. Gap probes are probes 
with number of mode greater than 1.
}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor,
\emph{ENmix: a novel background correction method for Illumina 
HumanMethylation450 BeadChip}. Nucleic Acids Research 2015 
}
\author{
Zongli Xu
}
\examples{
\donttest{

if (require(minfiData)) {
mdat <- preprocessRaw(RGsetEx)
beta=getBeta(mdat, "Illumina")
nmode=nmode(beta, minN = 3,modedist=0.2, nCores = 5)

path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mdat <- getmeth(rgSet)
beta=getB(mdat)
nmode=nmode(beta, minN = 3,modedist=0.2, nCores = 5)
}}}

\name{getmeth}
\alias{getmeth}
\title{
Create a methDataSet
}
\description{
To create a methDataSet based on a rgDataset
}
\usage{
getmeth(rgSet)
}
\arguments{
  \item{rgSet}{An object of class \code{rgDataSet}}
}
\value{
An object of class \code{methDataSet}
}

\details{
CpG annotation information is incorporated in the output methDataSet 
object, and can be extracted using command \code{rowData().}
}
\author{
Zongli Xu
}
\examples{
\donttest{

require(minfiData)
path <- file.path(find.package("minfiData"),"extdata")
#based on rgDataset
rgSet <- readidat(path = path,recursive = TRUE)
meth=getmeth(rgSet)
meth
cginfo=rowData(meth)
}
}

\name{p.qqplot}
\alias{p.qqplot}
\title{
P value Q-Q plot
}
\description{
P value Q-Q plot with optional confidence interval
}
\usage{
       p.qqplot(pvalues,outf="qq",outfmt="jpg",draw.conf=TRUE,
          conf.col="lightgray",conf.alpha=.95,pch=20,col="black",reducesize=0,...)
}
\arguments{
  \item{pvalues}{An numeric vector of P values}
  \item{outf}{figure file name, default "qq"}
  \item{outfmt}{Output figure file format, "jpg" or "eps"}
  \item{draw.conf}{Whether to draw confidence interval of expected P values under NULL hypothesis}
  \item{conf.col}{Color code of confidence interval}
  \item{conf.alpha}{Confidence interval range, 0.95 in default}
  \item{pch}{Point type code}
  \item{col}{Point color code}
  \item{reducesize}{A positive interger, larger the value, smaller the eps file size.
      Smaller file size is achieved by skipping some densely packed data points}
  \item{...}{Arguments pass to plot}
}
\details{P value Q-Q plot with optional confidence interval
}
\author{
Zongli Xu
}

\examples{
\donttest{

dat=simubed()
p.qqplot(pvalues=dat$p,draw.conf=TRUE,outf="qq_try",outfmt="jpg")
}
}

\name{dupicc}
\alias{dupicc}
\title{
Evaluation of measurement reliability using duplicate samples
}
\description{
The function can be used to evaluate duplicate samples by calculating: 
1) centered/un-centered Pearson's correlation coefficient between duplicates; 
2) abosolute difference between duplicates;
3) ICC for each CpG probes using oneway or twoway model.  
}
\usage{
dupicc(dat,dupid,mvalue=FALSE,center=TRUE,nCores=2,qcflag=FALSE,qc=NULL,
       detPthre=0.05,nbthre=3,skipicc=FALSE,corfig=FALSE,model="oneway")
}
\arguments{
  \item{dat}{Methylation beta value matrix}
  \item{dupid}{A data frame with two variables, id1 and id2. The two ids 
              in each row indicate a duplicate pair. These ids should be
              the same with column names of the input methylation matrix}
  \item{mvalue}{If TRUE, the beta value will be converted to M value for 
             calculation of ICC}
  \item{center}{If TRUE, the methylation beta values will be centered 
             for each CpG before calculation of ICC or correlation}
  \item{nCores}{Number of cores will be used for calculation of ICC}
  \item{qcflag}{Whether to perform QC before calculation of ICC}
  \item{qc}{QC object from function QCinfo}
  \item{detPthre}{If qcflag=TRUE, the methylation values with detection P value
       higher than the threshold will be removed before calculation}
  \item{nbthre}{If qcflag=TRUE, the methylation values with number of bead smaller 
       than the threshold will be removed}
  \item{skipicc}{If TRUE, ICC calculation will be skipped}
  \item{corfig}{If TRUE, a figure will be generated to demonstrate correlations 
         within duplicates or within non-duplicates}
  \item{model}{Using "oneway" or "twoway" model to calculate ICC} 
}
\value{
icc:  a data frame containing ICC and P values for each probe

dupcor: a data frame containing Pearson's correlation and 
        averaged absolute difference between duplicates. 

}
\references{
Zongli Xu, Jack A Taylor. \emph{Reliability of DNA methylation measures using 
Illumina methylation BeadChip}. Epigenetics 2020
}
\author{
Zongli Xu
}

\examples{
\donttest{
if (require(minfiData)){
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mdat=getmeth(rgSet)
beta=getB(mdat,"Illumina")
dupidx=data.frame(id1=c("5723646052_R02C02","5723646052_R04C01","5723646052_R05C02"),
id2=c("5723646053_R04C02","5723646053_R05C02","5723646053_R06C02"))
iccresu<-dupicc(dat=beta,dupid=dupidx)
}}
}

\name{rcp2}
\alias{rcp2}
\title{
Modified RCP method
}
\description{
Modified RCP method to normalize user's data to a list of reference values
}
\usage{
        rcp2(datMeth, reference, quantile.grid=seq(0.001,0.999,by=0.001))
}
\arguments{
  \item{datMeth}{A matrix with row for probes and column for samples}
  \item{reference}{A data frame with two columns, "cg" for CpG names and "meth_mean" for reference values}
  \item{quantile.grid}{Quantile grid used in linear regression}
}
\details{
The function will normalize user data to a reference value distribution based on 
common set of probe between user data and reference data.
}
\value{
A matrix with same dimension of user data
}
\references{
Liang Niu, Zongli Xu and Jack A. Taylor
\emph{RCP: a novel probe design bias correction method for Illumina Methylation 
BeadChip, Bioinformatics 2016}
}
\author{
Zongli Xu
}
\examples{
\donttest{

require(minfiData)
path <- file.path(find.package("minfiData"),"extdata")
#based on rgDataset
rgSet <- readidat(path = path,recursive = TRUE)
meth=getmeth(rgSet)
beta=getB(meth)
reference=data.frame(cg=rownames(beta),goldstandard=beta[,3])
reference=reference[sample(1:nrow(reference),2000),]
beta2=rcp2(beta,reference)
}}

\name{preprocessENmix}
\alias{preprocessENmix}
\title{
The ENmix background correction
}
\description{
The ENmix background correction for HumanMethylation 450 and MethylationEPIC  BeadChip.
ENmix models methylation signal intensities with a flexible exponential-normal mixture
 distribution, and models background noise with a truncated normal distribution. ENmix
 will split BeadChip intensity data into 6 parts and separately model methylated and
 unmethylated intensities, 2 different color channels and 2 different probe types.
}
\usage{
preprocessENmix(rgSet, bgParaEst = "oob", dyeCorr="RELIC", QCinfo=NULL, exQCsample=TRUE,
                    exQCcpg=TRUE, exSample=NULL, exCpG=NULL, nCores = 2)
}
\arguments{
  \item{rgSet}{An object of class \code{rgDataSet}, \code{methDataSet}, 
     \code{RGChannelSetExtended}, \code{RGChannelSet} or \code{MethylSet}.}
  \item{bgParaEst}{Method to estimate background normal distribution parameters.
 Options are: "oob","est", or "neg".}
  \item{dyeCorr}{Dye bias correction method, "mean": correction based on averaged red/green
 ratio, or "RELIC": correction with RELIC method (default method), or "none": no dye bias correction.}
  \item{QCinfo}{If QCinfo object from function QCinfo() is provided, low quality 
samples (if exQCsample=TRUE) and CpGs (if exQCcpg=TRUE) will be excluded before
 background correction.}
  \item{exQCsample}{If TRUE, low quality samples listed in QCinfo will be excluded.}
  \item{exQCcpg}{If TRUE, low quality CpGs listed in QCinfo will be excluded.}
  \item{exSample}{User specified samples to be excluded before background correction}
  \item{exCpG}{User specified probes to be excluded before background correction}
  \item{nCores}{Number of cores will be used for computation}
}
\details{
By default, ENmix will use out-of-band Infinium I intensities ("oob") to estimate
 normal distribution parameters for modeling background noise. Option "est" will use
 combined methylated and unmethylated intensities to estimate background distribution
 parameters separately for each color channel and each probe type. Option "neg" will
 use 600 chip internal controls probes to estimate background distribution parameters.
 If rgSet if a \code{MethylSet}, then only option "est" can be selected. 
}
\value{
An object of class same with input data}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor, ENmix: a novel background
 correction method for Illumina HumanMethylation450 BeadChip. Nucleic Acids
 Research 2015.

Zongli Xu, Sabine A. S. Langie, Patrick De Boever, Jack A. Taylor1 and Liang 
Niu, RELIC: a novel dye-bias correction method for Illumina Methylation BeadChip, 
BMC Genomics. 2017 
}
\author{
Zongli Xu and Liang Niu
}
\examples{
\donttest{

if (require(minfiData)) {
#rgDataSet as input
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
#quality control information
qc<-QCinfo(rgSet)
#further excluding samples which are not in the qc$badsample list
ex_id=c("5723646053_R04C02")
#further excluding cpgs which are not in the qc$badCpG list
ex_cg=c("cg00000622", "cg00001245", "cg00001261")
mdat=preprocessENmix(rgSet,QCinfo=qc,exSample=ex_id,exCpG=ex_cg,nCores=6)


#RGChannelSet as input
mdat=preprocessENmix(RGsetEx,nCores=6)
}}
}

\name{rm.cgsuffix}
\alias{rm.cgsuffix}
\title{
Remove suffix from CpG names and combine duplicated CpGs
}
\description{
Remove suffix from CpG names for EPIC v2 BeadChips and combined values for duplicated CpGs
}
\usage{
rm.cgsuffix(datMeth)
}
\arguments{
  \item{datMeth}{A methlation data matrix with row names for CpG id}
}
\details{
Remove suffix from CpG names for EPIC v2 BeadChips and combined values for duplicated CpGs
}
\value{
A matrix with number of rows equal or less than input data.
}
\author{
Zongli Xu
}

\examples{
\donttest{

#beta matrix with row for CpGs and column for samples
beta2=rm.cgsuffix(beta)
}}

\name{getB}
\alias{getB}
\title{
Extract methylation Beta values.
}
\description{
Extract Methylation Beta value, Beta = Meth / (Meth + Unmeth + offset)
}
\usage{
getB(mdat,type="Illumina",offset=100)
}
\arguments{
  \item{mdat}{An object of class \code{methDataSet} or  \code{MethylSet}.}
  \item{type}{type="Illumina" sets offset=100 as per Genome Studio software.}
  \item{offset}{Regularization factor in calculating beta ratio, 100 in default}
}
\value{
Methylation Beta value = Meth / (Meth + Unmeth + offset). Meth is methylated intensity matrix,
Unmeth is unmethylated intensity matrix.
}
\author{
Zongli Xu
}

\examples{
\donttest{
if (require(minfiData)){
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mdat=getmeth(rgSet)
beta=getB(mdat,"Illumina")
}}
}

\name{getCGinfo}
\alias{getCGinfo}
\title{
CpG probe annotation inforamtion
}
\description{
Extract CpG probe annotation inforamtion from an rgDataSet
}
\usage{
getCGinfo(rgSet, type="IandII")
}
\arguments{
  \item{rgSet}{An object of class \code{rgDataSet}}
  \item{type}{One of the following options "I","II","IandII","ctrl", 
    indicating type I, type II type I & II or control probes type}
}
\value{
An object of data frame class}
\author{
Zongli Xu
}
\examples{
\donttest{

require(minfiData)
path <- file.path(find.package("minfiData"),"extdata")
#based on rgDataset
rgSet <- readidat(path = path,recursive = TRUE)
cginfo=getCGinfo(rgSet,type="IandII")
}
}

\name{QCinfo}
\alias{QCinfo}
\title{
Extract QC information
}
\description{
Extract information for data quanlity control: detection P values, number of
 beads and averaged bisulfite conversion intensity. 
The function can also identify low quality samples and probes, as well as outlier
 samples in total intensity or beta value distribution.
}
\usage{
QCinfo(rgSet, detPthre=0.000001, detPtype="negative", nbthre=3, samplethre=0.05,
       CpGthre=0.05, bisulthre=NULL, outlier=TRUE, distplot=TRUE)
}
\arguments{
  \item{rgSet}{An object of class \code{rgDataSet}, or \code{RGChannelSetExtended}} 
  \item{detPthre}{Detection P value threshold to identify low quality data point}
  \item{detPtype}{Calculate detection P values based on negtive internal control
 ("negative") probes or out of the band ("oob") probes}
  \item{nbthre}{Number of bead threshold to identify data point of low quality}
  \item{samplethre}{Threshold to identify samples with low data quality, the percentage of
 low quality methylation data points across probes for each sample}
  \item{CpGthre}{Threshold to identify probes with low data quality, percentage of low 
quality methylation data points across samples for each probe}
  \item{bisulthre}{Threshold of bisulfite intensity for identification of low 
quality samples. By default, Mean - 3 x SD of sample bisufite control 
intensities will be used as a threshold.}
  \item{outlier}{If TRUE, outlier samples in total intensity or beta value 
distribution will be idenfied and classified as bad samples.}
  \item{distplot}{TRUE or FALSE, whether to produce beta value distribution 
plots before and after QC.}
}
\value{
detP:  a matrix of detection P values

nbead: a matrix for number of beads

bisul:  a vector of averaged intensities for bisulfite conversion controls per sample

badsample: a list of low quality or outlier samples

badCpG:  a list of low quality CpGs

outlier_sample:  a list of outlier samples in methylation beta value or totol 
                intensity distribution.

Figure "qc_sample.jpg": scatter plot of Percent of low quality data per sample 
vs. Average bisulfite conversion intensity

Figure "qc_CpG.jpg": histogram for Percent of low quality data per CpG.

Figure "freqpolygon_beta_beforeQC.jpg": distribution plot of input data, 
   samples colored in red are "bad" samples, list in badsample, including samples 
   with low data quality and outlier in methylaiton beta value or total intensity.

Figure "freqpolygon_beta_afterQC.jpg": distribution plot input data after filtering
   "bad" samples.
}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor,
\emph{ENmix: a novel background correction method for Illumina HumanMethylation450
 BeadChip}. Nucleic Acids Research 2015.
}
\author{
Zongli Xu
}

\examples{
\donttest{

if (require(minfiData)) {
#rgDataSet as input
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
qc=QCinfo(rgSet)

#RGChannelSetExtended as input
sheet <- read.metharray.sheet(file.path(find.package("minfiData"),"extdata"),
 pattern = "csv$")
rgSet <- read.metharray.exp(targets = sheet,extended = TRUE)
qc<-QCinfo(rgSet)
}}
}

\name{mpreprocess}
\alias{mpreprocess}
\title{
methylation data QC and preprocessing pipeline for Illuminal BeadChips
}
\description{
The pipeline performs background correction, dye bias correction, inter-array
 normalization and probe type bias correction for HumanMethylation 450 and
 MethylationEPIC BeadChip data. It removes or mitigates background noise and  
systematic experimental bias,It also perform quality controls, identifing and
 excluding low quality samples and probes, removing low quality and outlier values, 
and performing imputation.
}
\usage{
    mpreprocess(rgSet,nCores=2,bgParaEst="oob",dyeCorr="RELIC",
                qc=TRUE,qnorm=TRUE,qmethod="quantile1",
                fqcfilter=FALSE,rmcr=FALSE,impute=FALSE)
}
\arguments{
  \item{rgSet}{An object of class \code{rgDataSet},\code{methDataSet}, 
\code{RGChannelSetExtended}, \code{RGChannelSet} or \code{MethylSet}.}
  \item{nCores}{Number of cores will be used for computation}
  \item{bgParaEst}{Method to estimate background normal distribution parameters.
  Possible options: "oob","est", or "neg".}
  \item{dyeCorr}{Dye bias correction, "mean": correction based on averaged 
red/green ratio; or "RELIC": correction with RELIC method; or "none": no dye
 bias correction. The default is RELIC}
  \item{qc}{If TRUE, QC will be performed. Low quality samples and CpGs will
 be excluded before background correction.}
  \item{qnorm}{If TRUE, inter-array quantile normalization will be performed.}
  \item{qmethod}{Quantile normalization method. This should be one of the
 following strings: "quantile1", "quantile2", or "quantile3". See details
 in function norm.quantile.}
  \item{fqcfilter}{If TRUE, outlier and low quality values will be filtered out.}
  \item{rmcr}{TRUE: excluded rows and columns with more than 5\% of missing
 values. FALSE is in default}
  \item{impute}{Whether to impute missing values. If TRUE, k-nearest neighbor's
 methods will be used for imputation. FALSE is in default.}
}
\details{
Fuction mpreprocess is a pipeline that perform methylaiton data preprocessing
 and quality controls using functions: preprocessENmix, norm.quantile, rcp,
 QCinfo and qcfilter. More customized preprocessing steps can be achieved
 using the individual functions, see user's guide.
}
\value{
A methylation beta value matrix with rows for CpGs and columns for samples.
}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor, ENmix: a novel background
 correction method for Illumina HumanMethylation450 BeadChip. Nucleic Acids
 Research 2015.

Zongli Xu, Sabine A. S. Langie, Patrick De Boever, Jack A. Taylor1 and Liang
 Niu, RELIC: a novel dye-bias correction method for Illumina Methylation 
BeadChip, BMC Genomics, 2017

Liang Niu, Zongli Xu and Jack A. Taylor: RCP: a novel probe design bias
 correction method for Illumina Methylation BeadChip, Bioinformatics 2016
}
\author{
Zongli Xu
}
\seealso{
  Package \code{minfi} for classes \code{\linkS4class{RGChannelSet}} 
and \code{\linkS4class{MethylSet}}
}
\examples{
\donttest{

if (require(minfiData)) {
#rgDataSet as input
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
beta=mpreprocess(rgSet,nCores=6,qc=TRUE,fqcfilter=TRUE,rmcr=TRUE,impute=TRUE)

#methDataSet as input
mdat=getmeth(rgSet)
beta=mpreprocess(mdat,nCores=6)

#RGChannelSet as input
beta=mpreprocess(RGsetEx,nCores=6)

#RGChannelSetExtended as input
sheet <- read.metharray.sheet(file.path(find.package("minfiData"),"extdata"),
 pattern = "csv$")
rgSet <- read.metharray.exp(targets = sheet,extended = TRUE)
beta=mpreprocess(rgSet,nCores=6,qc=TRUE,fqcfilter=TRUE,rmcr=TRUE,impute=TRUE)

#MethylSet as input
mdat=preprocessRaw(rgSet)
beta=mpreprocess(mdat,nCores=6)

}}
}

\name{ipdmr}
\alias{ipdmr}
\title{
   Differentially methylated regions
}
\description{
   To identify differentially methylated regions using an interval P value method
}
\usage{
       ipdmr(data,include.all.sig.sites=TRUE,dist.cutoff=1000,bin.size=50,
             seed=0.05,region_plot=TRUE,mht_plot=TRUE,verbose=TRUE)
}
\arguments{
  \item{data}{A data frame with colname name "chr","start",
   "end","p" and "probe", indicating chromosome (1,2,3,...,X,Y), chromosome 
   start and end position, P value and probe names}
  \item{include.all.sig.sites}{Whether to use CpG singletons in calculation of FDR}
  \item{dist.cutoff}{Maximum distance in base pair to combine adjacent DMRs, and 
         the maximum distance between CpGs where auto-correlation will be calculated}
  \item{bin.size}{bin size for autocorrelation calculation}
  \item{seed}{FDR threshold for initial selection of DMR regions}
  \item{region_plot}{If TRUE, regional plots will be produced for each DMR}
  \item{mht_plot}{If TRUE, a p-value mahattan plot with marked DMRs will be produced}
  \item{verbose}{Whether to output detailed information}
}
\details{
  The input should be a data frame with column names "chr","start", "end","p" and "probe",
indicating chromosome, start and end position, P value and probe name. The function will use a 
novel interval p value method to identify differentially methylated regions. DMR results
 will be stored in a file with name resu_ipdmr.csv. If plot options were selected,
 two figure files will be generated: mht.jpg and region_plot.pdf.
}
\references{
Zongli Xu, Changchun Xie, Jack A. Taylor, Liang Niu, ipDMR: Identification
of differentially methyl-ated regions with interval p-values, Bioinfomatics 2020
}
\author{
Liang Niu, Zongli Xu
}

\examples{
\donttest{
dat=simubed()
names(dat)
#seed=0.1 is only for demonstration purpose, it should be smaller than 0.05 or 0.01 in actual study.
ipdmr(data=dat,seed=0.1) #seed=0.1
}
}

\docType{class}
\name{rgDataSet-class}
\alias{rgDataSet-class}
\alias{rgDataSet}
%methods
\alias{show, rgDataSet-method}

\title{Class \code{"rgDataSet"}}
\description{A class for storing Illumina methylation array 
raw intensity data of two color channels, and probe annotation 
infomation.}
\usage{
rgDataSet(Red = new("matrix"), Green = new("matrix"),
    NBeads = new("matrix"),rowData=new("DataFrame"),ictrl= new("DataFrame"),...)
}

\arguments{
\item{Red}{A matrix of Red channel intensity values with row for methylation probes
     and column for samples}
\item{Green}{A matrix of Green channel intensity values with row for methylation probes
     and column for samples}
\item{NBeads}{A matrix contains the number of beads used to generate intensity values
      on the Red and Green channels.}
\item{rowData}{A dataframe contains probe annotation information}
\item{ictrl}{A dataframe contais detailed information for Illumina internal control probes}
\item{...}{other arguments for class SummarizedExperiment}
}
\value{
  An object of class \code{rgDataSet}
}

\examples{
showClass("rgDataSet")
}

\name{multifreqpoly}
\alias{multifreqpoly}
\title{
Multiple frequency polygon plot
}
\description{
Produce Frequency polygon plot for each column of a numeric data matrix. 
Similar to multidensity function, the plot can be used to inspect data distribution
but with much faster speed.
}
\usage{
multifreqpoly(mat, nbreaks=100, col=1:ncol(mat), xlab="", 
              ylab="Frequency",legend = list(x = "top", fill=col,
              legend = if(is.null(colnames(mat))) paste(1:ncol(mat)) 
              else colnames(mat)),append=FALSE,...)
}
\arguments{
  \item{mat}{A numeric matrix}
  \item{nbreaks}{The number of bins for frequency counting}
  \item{col}{Line plot color code, the length should be equal to the number of
 columns in mat}
  \item{xlab}{x-axis lable}
  \item{ylab}{y-axis lable}
  \item{legend}{A list of arguments that get passed to the function "legend"}
  \item{append}{TRUE or FALSE, whether to create a new figure or append to the current figure.}
  \item{...}{Further arguments that get passed to the function "plot"}
}
\value{
Frequency polygon plots. 
}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor,
\emph{ENmix: a novel background correction method for Illumina HumanMethylation450
 BeadChip}. Nucleic Acids Research 2015.
}
\author{
Zongli Xu
}

\examples{
\donttest{

x=matrix(rnorm(10000),nrow=2000,ncol=5)
multifreqpoly(x,nbreaks=15,legend=list(x="topright",fill=1:ncol(x),legend=paste("V",1:5,sep="")))

if (require(minfiData)) {
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
mraw <- getmeth(rgSet)
beta<-getB(mraw)
jpeg("dist_raw.jpg")
multifreqpoly(beta,col=1:ncol(beta))
dev.off()
}
}}

\name{readmanifest}
\alias{readmanifest}
\title{
Parsing Illumina methylation arrays manifest file.
}
\description{
Parsing Illumina methylation arrays manifest file.
}
\usage{
readmanifest(file)
}
\arguments{
  \item{file}{Illumina methylation array manifest file, downloaded from Illuminal website}
}
\value{
An object of dataframe caintaining probe annotation information
}
\author{
Zongli Xu
}
\examples{
\donttest{

#for MethylationEPIC v1.0 B5
system("wget https://webdata.illumina.com/downloads/productfiles/methylationEPIC/
       infinium-methylationepic-v-1-0-b5-manifest-file-csv.zip")
system("unzip infinium-methylationepic-v-1-0-b5-manifest-file-csv.zip")
mf="infinium-methylationepic-v-1-0-b5-manifest-file.csv"
manifest=readmanifest(mf)

#for MethylationEPIC v1.0 B4
system("wget https://webdata.illumina.com/downloads/productfiles/methylationEPIC/
       infinium-methylationepic-v-1-0-b4-manifest-file-csv.zip")
system("unzip infinium-methylationepic-v-1-0-b4-manifest-file-csv.zip")
mf="MethylationEPIC_v-1-0_B4.csv"
manifest=readmanifest(mf)

#for HumanMethylation450
system("wget https://webdata.illumina.com/downloads/productfiles/
       humanmethylation450/humanmethylation450_15017482_v1-2.csv")
mf="HumanMethylation450_15017482_v1-2.csv"
manifest=readmanifest(mf)

}
}

\name{predSex}
\alias{predSex}
\title{
  Estimating sample sex
}
\description{
  Estimating sample sex based on methylation data
}
\usage{
predSex(mdat, cutoff = 2)
}
\arguments{
  \item{mdat}{An object of class \code{MethDataSet} or \code{rgDataSet}.}
  \item{cutoff}{The difference in log2 total intensity between X and Y 
chromosomes}
}
\details{
  Estimation of sex is based on the difference of log2 median total 
intensity measures on the   X and Y chromosomes. 
}
\author{
Zongli Xu
}
\examples{
\donttest{

if (require(minfiData)) {
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
sex=predSex(rgSet)
}}}

\name{freqpoly}
\alias{freqpoly}
\title{
Frequency polygon plot 
}
\description{
Similar to histogram, frequency polygon plot can be used to inspect data distribution.
}
\usage{
freqpoly(mat, nbreaks=15, col="black", xlab="", ylab="Frequency", 
         type="l",append=FALSE,...)
}
\arguments{
  \item{mat}{A numeric vector}
  \item{nbreaks}{Number of bins for frequency counting}
  \item{col}{color code}
  \item{xlab}{x-axis lable}
  \item{ylab}{y-axis lable}
  \item{type}{character indicating the type of plotting; actually any of the 'type's
        as in 'plot.default'.}
  \item{append}{TRUE or FALSE, whether to create a new figure or append to the current figure.}
  \item{...}{Further arguments that get passed to the function "plot"}
}
\value{
Frequency polygon plot.
}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor,
\emph{ENmix: a novel background correction method for Illumina HumanMethylation450
      BeadChip}. Nucleic Acids Research 2015.
}
\author{
Zongli Xu
}

\examples{
\donttest{
freqpoly(rnorm(1000))
}}

\name{plotCtrl}
\alias{plotCtrl}
\title{
Internal control plot
}
\description{
The function will generate a series of internal control plots that are similar to 
the plots from Illumina GenomeStudio software. Users should refer to GenomeStudio
online guide to interpret these figures. These figures can
 be used to check data quality and experimental procedures. 
}
\usage{
plotCtrl(rgSet,IDorder=NULL)
}
\arguments{
  \item{rgSet}{An object of class \code{rgDataSet} or \code{RGChannelSet}.}
  \item{IDorder}{A list of sample ids in the order specified by user. The list can
 be a subset of sample ids in input dataset. If an id list is provided, all
 plots will be produced in the order of the list.}
}
\value{
A set of internal control figures in jpeg format.
}
\references{
Zongli Xu, Liang Niu, Leping Li and Jack A. Taylor,
\emph{ENmix: a novel background correction method for Illumina HumanMethylation450
 BeadChip}. Nucleic Acids Research 2015.
}
\author{
Zongli Xu
}

\examples{
\donttest{

if (require(minfiData)) {
#rgDataSet as input
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
plotCtrl(rgSet)

#RGChannelSet as input
pinfo=pData(RGsetEx)
IDorder=rownames(pinfo)[order(pinfo$Slide,pinfo$Array)]
plotCtrl(RGsetEx,IDorder)
}}
}

\name{rcp}
\alias{rcp}
\title{
Illumina methylation array probe type bias correction
}
\description{
Probe design type bias correction using Regression on Correlated Probes (RCP) method
}
\usage{
        rcp(mdat, dist=25, quantile.grid=seq(0.001,0.999,by=0.001), qcscore = NULL,
 nbthre=3, detPthre=0.000001)
}
\arguments{
  \item{mdat}{An object of class \code{methDataSet} or \code{MethylSet}.}
  \item{dist}{Maximum distance in base pair between type I and type II probe pairs 
for regression calibration}
  \item{quantile.grid}{Quantile grid used in linear regression}
  \item{qcscore}{Data quality infomation object, the output from function QCinfo.  
       If the object is provied, low quality data points as defined by detection 
       p value threshold (detPthre) or number of bead threshold (nbthre) will 
       be set as missing values.}
  \item{detPthre}{Detection P value threshold to define low qualitye data points}
  \item{nbthre}{Number of beads threshold to define low qualitye data points, 
       nbthre=3 in default.}
}
\details{
The function will first identify type I and type II probe pairs within a specified
 distance, and then perform linear regression calibration between the probe types.
 With the estimates the function will then adjust type II data using type I data 
 as references. 
}
\value{
A beta value matrix 
}
\references{
Liang Niu, Zongli Xu and Jack A. Taylor
\emph{RCP: a novel probe design bias correction method for Illumina Methylation 
BeadChip, Bioinformatics 2016}
}
\author{
Liang Niu, Zongli Xu
}
\examples{
\donttest{

if (require(minfiData)) {
#methDataSet as input
path <- file.path(find.package("minfiData"),"extdata")
rgSet <- readidat(path = path,recursive = TRUE)
qc=QCinfo(rgSet)
mdat=preprocessENmix(rgSet,QCinfo=qc,nCores=6)
mdat=norm.quantile(mdat,method="quantile1")
beta=rcp(mdat)

#methylset as input
sheet <- read.metharray.sheet(file.path(find.package("minfiData"),"extdata"),
 pattern = "csv$")
rgSet <- read.metharray.exp(targets = sheet,extended = TRUE)
qc=QCinfo(rgSet)
mdat=preprocessENmix(rgSet,QCinfo=qc,nCores=6)
mdat=norm.quantile(mdat,method="quantile1")
beta=rcp(mdat)
}}
}
